#!/bin/bash

# csh and tcsh do not like this file
# ash produces incorrect quoting for " \"\\\"FOO\\\"\" "

[ "`/bin/ls -la /bin/sh |grep '\-> ash'`" != "" ] && \
	echo -e "\nash shell is not supported\n" && exit

# clean up before generating new Makefiles
[ -e "Makefile" ] && make distclean

##############################################################################

[ -z "$CC" ] && CC="cc"
[ -z "$LD" ] && LD="cc"
[ -z "$CFLAGS" ] && CFLAGS="-O2 -Wall" || CFLAGS="$CFLAGS -Wall"
[ "`gcc -v 2>&1 |grep \"gcc version 3\"`" != "" ] && CFLAGS="$CFLAGS -W -Wno-unused-parameter"
[ -z "$LFLAGS" ] && LFLAGS=""

DYNAMIC="-Wl,--export-dynamic"

# where to find libraries
LIBPATH="/lib /usr/lib /usr/local/lib"
# where to find term* headers
TERMPATH="/usr/include"

PROG="cria"

[ -z "$OWNER" ] && OWNER="root"
[ -z "$GROUP" ] && GROUP="root"

[ -z "$XMOD" ] && XMOD="0755"
[ -z "$RMOD" ] && RMOD="0644"

LANG="english"
COLOR="default"

CFGFILES="config plugins servers"
DOCFILES="BUGS CHANGELOG COLORS COPYRIGHT CREDITS FORMAT_STRINGS HELP LICENSE TIME TODO"

# install directories
[ -z "$PREFIX" ] && PREFIX="/usr/local"
BIN="bin"
LIB="lib/cria"
SYSCONF="etc/cria"
CONF=".cria"
DOC="share/doc/cria"

##############################################################################
# Platform configurations

OPTIONS=""
PLUGOPTIONS=""
TCSLIBS="ncurses curses"
# most platforms include libdl in libc
SLIBS=""
LDCONFIGARG="-r"
TCHEADERS="termcap.h term.h"
HEADERS="termios.h"

echo
PLATFORM=`uname`
CYGWIN=`echo $PLATFORM |grep CYGWIN`
ARCH=`uname -m`
SHERRORMSG="\n\tPlease change the first line in configure.sh from #!/bin/sh to #!/bin/bash and try again.\n"

# this file does not work with sh
[ "`grep "^#\!/bin/sh$" configure.sh`" != "" ] && echo -e "$SHERRORMSG" && exit

if [ "$PLATFORM" = "Linux" ]; then
	echo -e "\t*** Generating Makefiles for Linux"
	SLIBS="dl"
	LDCONFIGARG="-v"
	if [ "$ARCH" = "x86_64" ]; then
		OPTIONS="$OPTIONS -D__64BIT__"
		PLUGOPTIONS="-fPIC"
	fi
	if [ "$ARCH" = "alpha" ]; then
		OPTIONS="$OPTIONS -D__64BIT__"
	fi
	if [ "$ARCH" = "sparc64" ]; then
		OPTIONS="$OPTIONS"
	fi
	if [ "$ARCH" = "ppc" ]; then
		OPTIONS="$OPTIONS"
	fi
elif [ "$PLATFORM" = "OpenBSD" ]; then
	echo -e "\t*** Generating Makefiles for OpenBSD"
	OPTIONS="$OPTIONS"
elif [ "$PLATFORM" = "FreeBSD" ]; then
	echo -e "\t*** Generating Makefiles for FreeBSD"
	OPTIONS="$OPTIONS"
elif [ "$PLATFORM" = "Darwin" ]; then
	echo -e "\t*** Generating Makefiles for Darwin"
	TERMPATH="/sw/include /usr/include"
	CFLAGS="$CFLAGS -I/sw/include"
	OPTIONS="$OPTIONS"
	# no plugins
	COREPLUGS=1
elif [ "$PLATFORM" = "SunOS" ]; then
	echo -e "\t*** Generating Makefiles for SunOS"
	SLIBS="resolv socket nsl"
	TCSLIBS="termcap $TCSLIBS"
	OPTIONS="$OPTIONS -D__curses_h__"
	# solaris' cc sucks, use gcc
	CC="gcc"
	LD="gcc"
	# no plugins
	COREPLUGS=1

	[ "`grep "^#\!/bin/sh$" configure.sh`" != "" ] && echo -e "$SHERRORMSG" && exit
elif [ "$CYGWIN" != "" ]; then
	echo -e "\t*** Generating Makefiles for CYGWIN"
	OPTIONS="$OPTIONS -D__CHAR_BZERO__"
	# cygwin doesn't have cc symlinks
	CC="gcc"
	LD="gcc"
	# no plugins
	COREPLUGS=1

	[ "`grep "^#\!/bin/sh$" configure.sh`" != "" ] && echo -e "$SHERRORMSG" && exit
else
	echo -e "\t*** Unknown platform ***"
	OPTIONS="$OPTIONS"
	# assume no plugins
	COREPLUGS=1
fi


##############################################################################
# Parse command line options

TEST="NO"
INSTALLHOME="NO"
while [ $1 ]; do
	if [ "$1" = "--help" ]; then
		echo
		echo -e "\t--home\t\t\tInstall to home directory"
		echo -e "\t--prefix     [path]\tSet prefix path [default: $PREFIX]"
		echo -e "\t--bindir     [path]\tSet binary path [default: \$PREFIX/$BIN]"
		echo -e "\t--libdir     [path]\tSet library path [default: \$PREFIX/$LIB]"
		echo -e "\t--sysconfdir [path]\tSet system config path [default: \$PREFIX/$SYSCONF]"
		echo -e "\t--confdir    [path]\tSet user config path [default: $CONF]"
		echo -e "\t--docdir     [path]\tSet help files path [default: \$PREFIX/$DOC]"
		echo
		echo -e "\t--lang       [lang]\tSet language: english, german, spanish or finnish [default: english]"
		echo -e "\t--color      [theme]\tSet color theme: default, air, ink or kunphuzil [default: default]"
		echo
		echo -e "\t--coreplugs\t\tCompile plugins into client"
		echo
		echo -e "\t--termcap\t\tAdd termcap library to detection list"
		echo
		echo -e "\t--quiet\t\t\tEnable fewer configure.sh messages"
		echo -e "\t--efence\t\tLink to Electric Fence, for debugging"
		echo -e "\t--debug\t\t\tEnable debugging and --test option"
		echo -e "\t--test\t\t\tSetup cria to operate from source directory"
		echo
		exit
	elif [ "$1" = "--home" ]; then
		if [ "$TEST" = "YES" ]; then
			echo "### --home option ignored: can't be used with --test or --debug"
		else
			INSTALLHOME="YES"
			BINDIR="$HOME/bin"
			LIBDIR="$HOME/$CONF/lib"
			SYSCONFDIR="$HOME/$CONF/etc"
			DOCDIR="$HOME/$CONF/doc"
			OWNER="$USER"
			GROUP="$USER"
			XMOD="0700"
			RMOD="0600"
		fi
		shift
	elif [ "$1" = "--prefix" ]; then
		if [ "$TEST" = "YES" -o "$INSTALLHOME" = "YES" ]; then
			echo "### --prefix option ignored: can't use with --test, --debug or --home"
		else
			PREFIX="$2"
		fi
		shift
		shift
	elif [ "$1" = "--bindir" ]; then
		if [ "$TEST" = "YES" -o "$INSTALLHOME" = "YES" ]; then
			echo "### --bindir option ignored: can't use with --test, --debug or --home"
		else
			BINDIR="$2"
		fi
		shift
		shift
	elif [ "$1" = "--libdir" ]; then
		if [ "$TEST" = "YES" -o "$INSTALLHOME" = "YES" ]; then
			echo "### --libdir option ignored: can't use with --test, --debug or --home"
		else
			LIBDIR="$2"
		fi
		shift
		shift
	elif [ "$1" = "--sysconfdir" ]; then
		if [ "$TEST" = "YES" -o "$INSTALLHOME" = "YES" ]; then
			echo "### --sysconfdir option ignored: can't use with --test, --debug or --home"
		else
			SYSCONFDIR="$2"
		fi
		shift
		shift
	elif [ "$1" = "--confdir" ]; then
		CONFDIR="$2"
		shift
		shift
	elif [ "$1" = "--docdir" ]; then
		if [ "$TEST" = "YES" -o "$INSTALLHOME" = "YES" ]; then
			echo "### --docdir option ignored: can't use with --test, --debug or --home"
		else
			DOCDIR="$2"
		fi
		shift
		shift
	elif [ "$1" = "--lang" ]; then
		LANG="$2"
		shift
		shift
	elif [ "$1" = "--color" ]; then
		COLOR="$2"
		shift
		shift
	elif [ "$1" = "--coreplugs" ]; then
		COREPLUGS=1
		shift
	elif [ "$1" = "--termcap" ]; then
		if [ "$CYGWIN" != "" ]; then
			echo "### --termcap option ignored: Cygwin termcap is busted"
		else
			TCSLIBS="termcap $TCSLIBS"
		fi
		shift
	elif [ "$1" = "--quiet" ]; then
		QUIET=1
		shift
	elif [ "$1" = "--efence" ]; then
		SLIBS="$SLIBS efence"
		shift
	elif [ "$1" = "--debug" ]; then
		if [ "$INSTALLHOME" = "YES" ]; then
			echo "### --home option ignored: overridden by --debug"
			INSTALLHOME="NO"
		fi
		OPTIONS="$OPTIONS -D__DEBUG__"
		TEST="YES"
		BINDIR="`pwd`/src"
		LIBDIR="`pwd`/plugins"
		SYSCONFDIR="`pwd`/src/.cria"
		DOCDIR="`pwd`"
		CFLAGS="$CFLAGS -g"
		shift
	elif [ "$1" = "--test" ]; then
		if [ "$INSTALLHOME" = "YES" ]; then
			echo "### --home option ignored: overridden by --test"
			INSTALLHOME="NO"
		fi
		TEST="YES"
		BINDIR="`pwd`/src"
		LIBDIR="`pwd`/plugins"
		SYSCONFDIR="`pwd`/src/.cria"
		DOCDIR="`pwd`"
		shift
	else
		echo "Invalid option ($1)"
		exit
	fi
done

[ "$PREFIX" = "/" ] && PREFIX=""
[ "$BINDIR" = "" ] && BINDIR="$PREFIX/$BIN"
[ "$LIBDIR" = "" ] && LIBDIR="$PREFIX/$LIB"
[ "$SYSCONFDIR" = "" ] && SYSCONFDIR="$PREFIX/$SYSCONF"
[ "$CONFDIR" = "" ] && CONFDIR="$CONF"
[ "$DOCDIR" = "" ] && DOCDIR="$PREFIX/$DOC"

if [ "$COREPLUGS" = "1" ]; then
	DYNAMIC=""
	cd plugins
	COREPLUGS=""
	for file in *.c; do
		COREPLUGS="$COREPLUGS ${file%%.c}"
	done
	cd ..
fi

echo
if [ "$COREPLUGS" = "" ]; then
	echo -e "\tConfig:    plugins = YES (as libraries)"
	echo -e "\tConfig:     libdir = $LIBDIR"
else
	echo -e "\tConfig:    plugins = NO (compiled into core)"
	LIBDIR=""
fi
	echo -e "\tConfig:     bindir = $BINDIR"
	echo -e "\tConfig: sysconfdir = $SYSCONFDIR"
	echo -e "\tConfig:    confdir = $CONFDIR"
	echo -e "\tConfig:     docdir = $DOCDIR"
	echo -e "\tConfig: user.group = $OWNER.$GROUP"
	echo -e "\tConfig:    options = $OPTIONS"
	echo -e "\tConfig:   language = $LANG"
	echo -e "\tConfig:      color = $COLOR"
	echo -e "\tConfig:   --> home = $INSTALLHOME"
	echo -e "\tConfig:       test = $TEST"
echo


##############################################################################
# Search for non-standard headers and libraries

ls_lib_search() {
	for dir in $LIBPATH; do
		if [ "`/bin/ls -1 $dir 2>&1 |grep lib$1`" != "" ]; then
			echo "found"
			return
		fi
	done
}

LIBS=""
if [ -x /sbin/ldconfig ]; then
	/sbin/ldconfig $LDCONFIGARG > .ldconfig.cache 2>&1
	echo -e "\tLibrary search using [ldconfig]..."
	for file in $TCSLIBS; do
		if [ "`cat .ldconfig.cache |grep lib$file`" != "" ]; then
			echo -e "\t\tfound lib$file"
			if [ "$file" == "termcap" ]; then
				OPTIONS="$OPTIONS -D__LIBTERMCAP__"
			fi
			LIBS="$LIBS -l$file"
			break
		fi
	done
	if [ "$LIBS" = "" ]; then
		echo -e "\n\t\t'$TCSLIBS' none found, aborting\n"
		exit
	fi

	for file in $SLIBS; do
		if [ "`cat .ldconfig.cache |grep lib$file`" = "" ]; then
			echo -e "\n\t\tlib$file was not found, aborting\n"
			exit
		else
			echo -e "\t\tfound lib$file"
			LIBS="$LIBS -l$file"
		fi
	done
	rm -f .ldconfig.cache
elif [ -x /bin/ls ]; then
	echo -e "\tLibrary search using [ls]..."
	for file in $TCSLIBS; do
		if [ "`ls_lib_search $file`" != "" ]; then
			echo -e "\t\tfound lib$file"
			LIBS="$LIBS -l$file"
			break
		fi
	done
	if [ "$LIBS" = "" ]; then
		echo -e "\n\t\t'$TCSLIBS' none found in\n\t\t[$LIBPATH], aborting\n"
		exit
	fi

	for file in $SLIBS; do
		if [ "`ls_lib_search $file`" = "" ]; then
			echo -e "\n\t\tlib$file was not found in\n\t\t[$LIBPATH], aborting\n"
			exit
		else
			echo -e "\t\tfound lib$file"
			LIBS="$LIBS -l$file"
		fi
	done
else
	echo -e "\n\tCan't find LS or LDCONFIG, aborting\n"
fi

ls_term_search() {
	for dir in $TERMPATH; do
		if [ -e $dir/$file ]; then
			echo "found"
			return
		fi
	done
}

for file in $HEADERS; do
	if [ "`ls_term_search $file`" = "" ]; then
		echo -e "\n\t$file was not found in\n\t[$TERMPATH], aborting\n"
		echo -e "Install glibc-devel for Redhat or libc6-dev for Debian\n"
		exit
	else
		echo -e "\tFound: $file"
	fi
done
TCSTAT=""
for file in $TCHEADERS; do
	if [ "`ls_term_search $file`" != "" ]; then
		echo -e "\tFound: $file"
		OPTIONS="$OPTIONS -D__${file%%.h}_h__"
		TCSTAT="found"
		break
	fi
done
if [ "$TCSTAT" = "" ]; then
	echo -e "\n\t'$TCHEADERS' none found in\n\t[$TERMPATH], aborting\n"
	echo -e "Install ncurses-devel for Redhat or libncurses5-dev for Debian\n"
	exit
fi

echo


##############################################################################
# Awk/Gawk detection

if [ "`which gawk`" != "" -a "$CYGWIN" = "" ]; then
	AWK="gawk"
	echo -e "\tUsing: gawk"
elif [ "`which awk`" != "" ]; then
	AWK="awk"
	echo -e "\tUsing: awk (might not compile)"
else
	echo -e "\tCouldn't find awk or gawk"
	exit
fi

echo


##############################################################################
# Language symlinks

cd src/.cria
if [ -e _formats.$LANG ]; then
	ln -s _formats.$LANG _formats
	echo -e "\tUsing $LANG formats"
else
	ln -s _formats.english _formats
	echo -e "\tUsing english formats ($LANG formats not found)"
fi
if [ -e _commands.$LANG ]; then
	ln -s _commands.$LANG _commands
	echo -e "\tUsing $LANG commands"
else
	ln -s _commands.english _commands
	echo -e "\tUsing english commands ($LANG commands not found)"
fi
if [ -e _variables.$LANG ]; then
	ln -s _variables.$LANG _variables
	echo -e "\tUsing $LANG variables"
else
	ln -s _variables.english _variables
	echo -e "\tUsing english variables ($LANG variables not found)"
fi
if [ -e _colors.$COLOR ]; then
	ln -s _colors.$COLOR _colors
else
	ln -s _colors.default _colors
	echo -e "\tUsing 'default' color theme ('$COLOR' color theme not found)"
fi
cd ../..
echo


##############################################################################
# Final messages

if [ "$QUIET" != "1" ]; then
	echo -e "\tType 'make' to compile"
	if [ "$TEST" = "NO" ]; then
		echo -e "\tType 'make install' to install"
	fi
	echo -e "\tType 'make clean' to clean up"
	echo -e "\tType 'make distclean' to clean up and remove Makefiles"
	echo

	echo -e "\tSample configuration files are located in src/.cria/"
	echo

	if [ "$TEST" = "NO" ]; then
		echo -e "\tYou must configure with the --test option if you want"
		echo -e "\tto run cria without installing it.\n"

		echo -e "\t*******************************************"
		echo -e "\t*                                         *"
		echo -e "\t*  Did you read the above messages?       *"
		echo -e "\t*  Did you read \`./configure.sh --help\`?  *"
		echo -e "\t*                                         *"
		echo -e "\t*******************************************"
		echo
	fi
fi


##############################################################################
# create Makefile

echo -e "# Generated by configure.sh\n
all:
	(cd src ; make)" >> Makefile

[ "$COREPLUGS" = "" ] && echo -e "\t(cd plugins ; make)" >> Makefile

if [ "$COREPLUGS" = "" -a "$TEST" = "NO" ]; then
	echo -e "\t@echo -e \"\\\n*** Plugins won't work until you do \\\"make install\\\"\"" >> Makefile
fi

echo -e "
distclean:
	make clean
	rm -f Makefile .ldconfig.cache
	(cd src ; rm -f Makefile .cria/_formats .cria/_commands .cria/_variables .cria/_colors .plugin-*)" >> Makefile

[ "$COREPLUGS" = "" ] && \
	echo -e "\t(cd plugins ; rm -f Makefile common.h formats.h functions.h functions_cmds.h)" >> Makefile

echo -e "
clean:
	(cd src ; make clean)" >> Makefile

[ "$COREPLUGS" = "" ] && echo -e "\t(cd plugins ; make clean)" >> Makefile

if [ "$TEST" = "NO" ]; then
	echo -e "\ninstall:\n\t(cd src ; make install)" >> Makefile
	[ "$COREPLUGS" = "" ] && echo -e "\t(cd plugins ; make install)" >> Makefile
fi


##############################################################################
# create src/Makefile

cd src
./build_formats.sh $AWK

[ "$COREPLUGS" != "" ] && ./coreplugs.sh $COREPLUGS

echo "# Generated by configure.sh" >> Makefile

OBJS=""
for file in *.c; do
	OBJS="$OBJS ${file%%.c}.o"
done
if [ "$COREPLUGS" != "" ]; then
	rm -f .plugin-*.c
	for file in $COREPLUGS; do
		OBJS="$OBJS .plugin-${file}.o"
		ln -s ../plugins/${file}.c .plugin-${file}.c
	done
	OPTIONS="$OPTIONS -DCOREPLUGS"
fi

echo -e "
all: $PROG

clean:
	rm -f $PROG *.o core *.file cgc_*

$PROG: $OBJS
	$LD $DYNAMIC $LFLAGS -L/usr/local/lib -o $PROG $OBJS $LIBS

.c.o:
	$CC $CFLAGS -o \$@ -c \$< -DPLUGDIR=\"\\\"$LIBDIR\\\"\" -DCFGDIR=\"\\\"$SYSCONFDIR\\\"\" -DCFGFILES=\"\\\"$CFGFILES\\\"\" -DCONFDIR=\"\\\"$CONFDIR\\\"\" -DDOCDIR=\"\\\"$DOCDIR\\\"\" $OPTIONS
" >> Makefile
for file in *.c; do
	echo "${file%%.c}.o: $file common.h functions.h cgc_cmds.h cgc_formats.h Makefile" >> Makefile
done
if [ "$COREPLUGS" != "" ]; then
	for file in $COREPLUGS; do
		echo ".plugin-${file}.o: .plugin-${file}.c common.h functions.h cgc_cmds.h cgc_formats.h Makefile" >> Makefile
	done
fi

if [ "$TEST" = "NO" ]; then
	echo -e "\ninstall: $PROG\n\tstrip $PROG" >> Makefile

	echo -e "\t(umask 022 ; /bin/mkdir -m $XMOD -p $SYSCONFDIR $BINDIR $DOCDIR)" >> Makefile
	echo -e "\t(cd .cria/ ; install -o $OWNER -g $GROUP -m $RMOD $CFGFILES $SYSCONFDIR/)" >> Makefile
	echo -e "\tinstall -o $OWNER -g $GROUP -m $XMOD $PROG $BINDIR" >> Makefile
	echo -e "\t(cd .. ; install -o $OWNER -g $GROUP -m $RMOD $DOCFILES $DOCDIR/)" >> Makefile
fi
cd ..


##############################################################################
# create plugins/Makefile

cd plugins

if [ "$COREPLUGS" = "" ]; then
	ln -s ../src/common.h .
	ln -s ../src/functions.h .
	ln -s ../src/cgc_*.h .

	echo "# Generated with configure.sh" >> Makefile

	OBJS=""
	for file in *.c; do
		OBJS="$OBJS ${file%%.c}.so"
	done

	echo -e "\nall: $OBJS\n\nclean:\n\trm -f *.so cgc_*\n" >> Makefile

	for file in *.c; do
		echo "${file%%.c}.so: $file common.h functions.h cgc_cmds.h cgc_formats.h Makefile" >> Makefile
		echo -en "\t$CC $CFLAGS -fPIC -shared -o ${file%%.c}.so $file -DPLUGDIR=\"\\\"$LIBDIR\\\"\" -DCFGDIR=\"\\\"$SYSCONFDIR\\\"\"" >> Makefile
		echo -e "-DCFGFILES=\"\\\"$CFGFILES\\\"\" -DCONFDIR=\"\\\"$CONFDIR\\\"\" -DDOCDIR=\"\\\"$DOCDIR\\\"\" $OPTIONS $PLUGOPTIONS" >> Makefile
	done

	if [ "$TEST" = "NO" ]; then
		echo -e "\ninstall: $OBJS" >> Makefile
		echo -e "\t(umask 022 ; /bin/mkdir -m $XMOD -p $LIBDIR)" >> Makefile
		echo -e "\tinstall -o $OWNER -g $GROUP -m $XMOD *.so $LIBDIR/" >> Makefile
	fi
fi
