//	cria - page saver plugin (logging)
//	Copyright (c) 2003 Brand Huntsman
//


#include <time.h>			// time() and localtime()
#include "common.h"
#include "functions.h"


#define MAX_COLOR_NAME 7


//////////////////////////////////////////////////////////////////////////

// local

////////////////////////////////////////////////////////////////////////// PRIVATE


void fprintf_string( FILE *fp, char *string ){
	for(; string != NULL && *string != '\0'; string++){
		if(*string > '\0' && *string < ' '){
			switch(*string){
			case '\n':
			case '\t':
				fputc(*string, fp);
				break;
			default:
				if(var_show_unknown_markers)
					fputc('\001', fp);
			}
		} else fputc(*string, fp);
	}
}


int savepage_callback( char *data, int datalen ){
	char filename[500], *pagename = p_current->name;
	time_t ct = time(NULL);
	struct tm *tp = localtime(&ct);
	s_line *line;
	FILE *fp;

	// open file
	if(*pagename == '!') pagename++;
	snprintf(filename, 500, "%s/%s-%d%02d%02d-%02d%02d", home, pagename,
		tp->tm_year + 1900, tp->tm_mon + 1, tp->tm_mday, tp->tm_hour, tp->tm_min);

	if((fp = fopen(filename, "w")) == NULL){
		msgCurrentPage(-1, NULL, MF_ERROR, 1, "can't open log file for writing");
		return(0);
	}

	// save log to file
	for(line = p_current->buffer_head; line != NULL; line = line->next){
		char item_color[10], c, *fs, *item_list[10];
		int nr_items, x, xx, fs_len;
		s_line_item *i;

		// timestamp
		if(var_timestamps){
			tp = localtime(&(line->ct));
			fprintf(fp, "%02d:%02d ", tp->tm_hour, tp->tm_min);
		}

		// network
		if(line->network != NULL)
			fprintf(fp, "%s ", line->network->name);

		// get format string for line type
		fs = formats[line->type];

		// collect line items in array
		for(nr_items = 0, i = line->first_item; i != NULL && nr_items < 10; nr_items++, i = i->next)
			item_list[nr_items] = i->item;

		// display formatted line
		fs_len = strlen(fs);
		for(x = 0; x < fs_len; x++){
			if(fs[x] == '$'){
				c = fs[x+1];
				if(c >= '0' && c <= '9'){
					x++;
					xx = c - '0';
					if(xx < nr_items)
						fprintf_string(fp, item_list[xx]);
					else fprintf(fp, "$%c", c);
				} else fputc('$', fp);
			} else if(fs[x] > '\0' && fs[x] < ' '){
				switch(fs[x]){
				case '\n':
				case '\t':
					fputc(fs[x], fp);
					break;
				default:
					if(var_show_unknown_markers)
						fputc('\001', fp);
				}
			} else if(fs[x] == '%'){
				xx = 0;
				do{
					x++;
					c = fs[x];
					if(c != '-') item_color[xx] = c;
					xx++;
				} while(c != '-' && xx <= MAX_COLOR_NAME);
				if(c != '-'){
					fputc('%', fp);
					x -= MAX_COLOR_NAME + 1;
				} else {
					xx--;
					item_color[xx] = '\0';
					switch(xx){
					case 2:
						if(!strcmp(item_color, "bg")){
						} else if(!strcmp(item_color, "fg")){
						} else { fputc('%', fp); x -= 3; }
						break;
					case 3:
						if(!strcmp(item_color, "sfg")){
						} else if(!strcmp(item_color, "net")){
						} else { fputc('%', fp); x -= 4; }
						break;
					case 4:
						if(!strcmp(item_color, "chan")){
						} else if(!strcmp(item_color, "info")){
						} else if(!strcmp(item_color, "mode")){
						} else if(!strcmp(item_color, "host")){
						} else if(!strcmp(item_color, "nick")){
						} else if(!strcmp(item_color, "time")){
						} else { fputc('%', fp); x -= 5; }
						break;
					case 5:
						if(!strcmp(item_color, "snick")){
						} else if(!strcmp(item_color, "power")){
						} else if(!strcmp(item_color, "error")){
						} else { fputc('%', fp); x -= 6; }
						break;
					case 6:
						if(!strcmp(item_color, "hsnick")){
						} else { fputc('%', fp); x -= 7; }
						break;
					case 7:
						if(!strcmp(item_color, "bracket")){
						} else { fputc('%', fp); x -= 8; }
						break;
					}
				}
			} else fputc(fs[x], fp);
		}

		fputc('\n', fp);
	}

	// close log file
	fclose(fp);

	msgCurrentPage(-1, NULL, MF_SAVE_PAGE, 1, filename);

	return(0);
}


////////////////////////////////////////////////////////////////////////// GLOBAL


int pagesaverInit( s_plug *self ){
	if(!plugVersion(0,2,9,4)){
		msgServer(3, NULL, MF_PLUG_REQUIRES, 2, "0.2.9cvs4", "Page Saver");
		return(0);
	}

	plugHookUser(self, "savepage", 0, &savepage_callback);

	return(1);
}


void pagesaverUnload( ){
}
