//	cria - partyline plugin
//	Copyright (c) 2002 Leif Ames
//


#include "common.h"
#include "functions.h"


//////////////////////////////////////////////////////////////////////////

// local
s_plug *party_plug;

typedef struct s_chanlist {
	struct s_chanlist *next, *prev;
	s_network *network;
	char *chan;
} s_chanlist;

s_chanlist *partylist;

////////////////////////////////////////////////////////////////////////// PRIVATE


int add_chan( s_chanlist **list, s_network *network, char *chan ){
	s_chanlist *newchan = malloc(sizeof(s_chanlist));
	char msg[200];

	newchan->network = network;
	newchan->chan = strdup(chan);
	newchan->prev = *list;
	newchan->next = NULL;
	if(*list) {
		(*list)->next = newchan;
	}
	*list = newchan;
	sprintf(msg, "Added channel %s%s to partyline", network->name, chan);
	msgCurrentPage(-1, NULL, MF_DEBUG, 1, msg);
	return(0);
}


int del_chan( s_chanlist *list, s_chanlist *node ){
	if(node->next)
		node->next->prev = node->prev;
	if(node->prev)
		node->prev->next = node->next;
	if(node == list) {
		if(node->next)
			list = node->next;
		else
			list = NULL;
	}
	free(node->chan);
	free(node);
	return(0);
}


int send_messages( s_chanlist *list, s_network *network, char *chan, char *nick, char *message ){
	s_chanlist *node, *prev;

	for(node=list;node;node=prev) {
		prev=node->prev;
		if(node->network != network || stringcompare(node->chan, chan) == 0) {
			netSend(node->network, "PRIVMSG %s :<%s> %s\n", node->chan, nick, message);
			msgUnkPage(node->network, chan, -1, NULL, MF_PLINE_MESSAGE, 3,
				message, nick, n_current->nickname);
		}
	}
	return(0);
}


int check_list( s_chanlist *list, s_network *network, char *chan ){
	s_chanlist *node, *prev;
	for(node=list;node;node=prev) {
		prev=node->prev;
		if(node->network == network && stringcompare(node->chan, chan) != 0) {
			return(1);
		}
	}
	return(0);
}


////////////////////


static int privmsg_callback( s_network *network, char *from, char *data, int datalen ){
	char *msg = skipword(data) + 1;
	char *host = skipnick(from);

	if(check_list(partylist, network, data)) {
		send_messages(partylist, network, data, from, msg);
		return(0); // dont passthru
	}

	*(host-1) = '!';
	*(msg-2) = ' ';
	return(1); // passthru to normal network command
}


int addchan_callback( char *data, int datalen ){
	char *chan = skipword(data);

	if(*chan == '\0')
		add_chan(&partylist, n_current, data);
	else
		add_chan(&partylist, networkGet(data), chan);
	return(0);
}


////////////////////////////////////////////////////////////////////////// GLOBAL


int partylineInit( s_plug *self ){
	if(!plugVersion(0,2,5,7)){
		msgServer(3, NULL, MF_PLUG_REQUIRES, 2, "0.2.5cvs7", "partyline");
		return(0);
	}

	plugHookServer(self, "PRIVMSG", &privmsg_callback);
	plugHookUser(self, "addchan", 0, &addchan_callback);

	return(1);
}


void partylineUnload( ){
}
