//	cria - IRC services plugin
//	Copyright (c) 2002-2004 Brand Huntsman
//


#include "common.h"
#include "functions.h"

#include <regex.h>


//////////////////////////////////////////////////////////////////////////

// local
typedef struct services_names {
	char *network;
	char *chanserv;
	char *nickserv;
	char *memoserv;
	char *otherserv;
	struct services_names *next;
} services_names;

services_names *services_head;
int services_got_password;
char default_chanserv[] = "chanserv";
char default_nickserv[] = "nickserv";
char default_memoserv[] = "memoserv";
char default_otherserv[] = "otherserv";

////////////////////////////////////////////////////////////////////////// PRIVATE


uint findotherserv( char *name, char *otherserv ){
	char *match = findstring(otherserv, name), *end;

	if(match == NULL) return(0);

	if(match != otherserv && *(match-1) != ' ') return(0);
	end = match + strlen(name);
	if(*end == '\0' || *end == ' ') return(1);
	return(0);
}


static int privmsg_callback( s_network *network, char *from, char *data, int datalen ){ // intercept PRIVMSG
	services_names *sn;
	char *host;

	for(sn = services_head; sn != NULL; sn = sn->next)
		if(!strcmp(sn->network, network->name))
			break;

	host = skipnick(from);

	if(pageGet(network, from) == NULL){
		if(stringcompare(from, sn->chanserv)){
			message_to_server = 1;
		} else if(stringcompare(from, sn->nickserv)){
			message_to_server = 1;
		} else if(stringcompare(from, sn->memoserv)){
			message_to_server = 1;
		} else if(findotherserv(from, sn->otherserv == NULL ? "" : sn->otherserv)){
			message_to_server = 1;
		}
	}

	*(host-1) = '!';
	return(1); // passthru to normal privmsg command
}


static int notice_callback( s_network *network, char *from, char *data, int datalen ){ // intercept NOTICE
	services_names *sn;
	char *host;

	for(sn = services_head; sn != NULL; sn = sn->next)
		if(!strcmp(sn->network, network->name))
			break;

	host = skipnick(from);

	// display message on channel page
	if(stringcompare(from, sn->chanserv)){
		data = findchar(data, ':');
		if(*data == ':'){
			regmatch_t match;
			regex_t re1, re2;
			char *message;

			data++;

			regcomp(&re1, "^\\[#[^] 	]+\\] ", REG_EXTENDED);
			regcomp(&re2, " \\[#[^] 	]+\\]$", REG_EXTENDED);

			if(!regexec(&re1, (const char *)data, 0, NULL, 0)){
				// "[#channe;] message"

				data++; // skip opening bracket
				message = skipword(data); // get channel name
				message[-2] = '\0'; // remove closing bracket

				msgUnkPage(network, data, 1, NULL, MF_NOTICE, 3, message, host, from);

				return(0); // dont passthru
			} else if(!regexec(&re2, (const char *)data, 1, &match, 0)){
				// "message [#channel]"

				message = data;

				// length of channel name - two brackets and space + null terminator
				data = (char *)track_malloc(match.rm_eo - match.rm_so - 3 + 1, "notice_callback");
				snprintf(data, match.rm_eo - match.rm_so - 3 + 1, "%s", &message[match.rm_so] + 2);
				// &message[match.rm_so] + 2 = skip over space and opening bracket

				msgUnkPage(network, data, 1, NULL, MF_NOTICE, 3, message, host, from);

				track_free(data, "notice_callback");

				return(0); // dont passthru
			}

			regfree(&re1);
			regfree(&re2);
		}
	}

	if(pageGet(network, from) == NULL){
		if(stringcompare(from, sn->chanserv)){
			message_to_server = 1;
		} else if(stringcompare(from, sn->nickserv)){
			message_to_server = 1;
		} else if(stringcompare(from, sn->memoserv)){
			message_to_server = 1;
		} else if(findotherserv(from, sn->otherserv == NULL ? "" : sn->otherserv)){
			message_to_server = 1;
		}
	}

	*(host-1) = '!';
	return(1); // passthru to normal notice command
}


void id_password_callback( ){
	services_got_password = 1;
}
int id_callback( char *data, int datalen ){ // identify with nickserv
	// ID
	char password[33];
	services_names *sn;
	int recon_failed;

	if(datalen > 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/ID");
	} else {
		for(sn = services_head; sn != NULL; sn = sn->next)
			if(!strcmp(sn->network, n_current->name))
				break;

		promptSetup(C_IN_FG, C_IN_INFO, "Enter password: ", strlen("Enter password: "),
			password, 32, 1, 1, &id_password_callback);
		services_got_password = 0;
		recon_failed = 0;
		while(!services_got_password && recon_failed != -1)
			recon_failed = listenSockets(RECONNECT_DISABLED);
		if(strlen(password) > 0){
			netSend(n_current, "PRIVMSG %s :IDENTIFY %s\n", sn->nickserv, password);
		}
	}

	return(0); // dont passthru
}
int ghost_callback( char *data, int datalen ){ // ghost a nick with nickserv
	// GHOST nick
	char password[33], *end;
	services_names *sn;
	int recon_failed;

	end = skipword(data);
	if(datalen == 0 || *end != '\0'){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/GHOST nick");
	} else {
		for(sn = services_head; sn != NULL; sn = sn->next)
			if(!strcmp(sn->network, n_current->name))
				break;

		promptSetup(C_IN_FG, C_IN_INFO, "Enter password: ", strlen("Enter password: "),
			password, 32, 1, 1, &id_password_callback);
		services_got_password = 0;
		recon_failed = 0;
		while(!services_got_password && recon_failed != -1)
			recon_failed = listenSockets(RECONNECT_DISABLED);
		if(strlen(password) > 0){
			netSend(n_current, "PRIVMSG %s :GHOST %s %s\n", sn->nickserv, data, password);
		}
	}

	return(0); // dont passthru
}


int cop_callback( char *data, int datalen ){ // op nicks with chanserv
	// COP nick [nick nick ...]
	services_names *sn;

	if(datalen == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/COP nick [nick nick ...]");
	} else {
		for(sn = services_head; sn != NULL; sn = sn->next)
			if(!strcmp(sn->network, n_current->name))
				break;

		netSend(n_current, "PRIVMSG %s :OP %s %s\n", sn->chanserv, p_current->name, data);
	}

	return(0); // dont passthru
}


int opme_callback( char *data, int datalen ){ // op self with chanserv
	// OPME
	services_names *sn;

	if(datalen > 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/OPME");
	} else {
		for(sn = services_head; sn != NULL; sn = sn->next)
			if(!strcmp(sn->network, n_current->name))
				break;

		netSend(n_current, "PRIVMSG %s :OP %s %s\n", sn->chanserv, p_current->name, n_current->nickname);
	}

	return(0); // dont passthru
}


int memo_callback( char *data, int datalen ){ // send/receive memos
	// MEMO SEND nick/channel message
	// MEMO READ [number]
	// MEMO DEL [number]
	char *value, *message;
	services_names *sn;
	s_page* page;

  // TODO: if used in channel pages then access memos for the channel instead of user memos

	if(datalen == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/MEMO SEND nick/channel message");
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/MEMO READ [number]");
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/MEMO DEL [number]");
	} else {
		value = skipword(data);
		for(sn = services_head; sn != NULL; sn = sn->next)
			if(!strcmp(sn->network, n_current->name))
				break;

		if(stringcompare(data, "send")){
			message = skipword(value);
			if(strlen(value) == 0 || strlen(message) == 0){
				msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/MEMO SEND nick/channel message");
			} else {
				if((page = pageGet(n_current, sn->memoserv)) == NULL)
					msgServer(3, n_current, MFP_SENT_MEMO, 2, message, value);
				else msgPage(page, 3, n_current, MFP_SENT_MEMO, 2, message, value);

				netSend(n_current, "PRIVMSG %s :SEND %s %s\n", sn->memoserv, value, message);
			}
		} else if(stringcompare(data, "read")){
			if(strlen(value) == 0){
				netSend(n_current, "PRIVMSG %s :list\n", sn->memoserv);
			} else {
				char msg[50];
				snprintf(msg, 50, "read memo %s", value);
				if((page = pageGet(n_current, sn->memoserv)) == NULL)
					msgServer(3, n_current, MF_DEBUG, 1, msg);
				else msgPage(page, 3, n_current, MF_DEBUG, 1, msg);

				netSend(n_current, "PRIVMSG %s :read %s\n", sn->memoserv, value);
			}
		} else if(stringcompare(data, "del")){
			if(strlen(value) == 0){
				netSend(n_current, "PRIVMSG %s :list\n", sn->memoserv);
			} else {
				char msg[50];
				snprintf(msg, 50, "delete memo %s", value);
				if((page = pageGet(n_current, sn->memoserv)) == NULL)
					msgServer(3, n_current, MF_DEBUG, 1, msg);
				else msgPage(page, 3, n_current, MF_DEBUG, 1, msg);

				netSend(n_current, "PRIVMSG %s :del %s\n", sn->memoserv, value);
			}
		} else {
			msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/MEMO SEND nick/channel message");
			msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/MEMO READ [number]");
			msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/MEMO DEL [number]");
		}
	}

	return(0); // dont passthru
}


void help_network_callback( ){
	msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/NETWORK CHANSERV name");
	msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/NETWORK NICKSERV name");
	msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/NETWORK MEMOSERV name");
	msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/NETWORK OTHERSERV name [name name ...]");
}
int network_callback( char *data, int datalen ){ // intercept NETWORK command
	// patches original commands
	// NETWORK +/- network_name
	// new commands
	// NETWORK CHANSERV name
	// NETWORK NICKSERV name
	// NETWORK MEMOSERV name
	// NETWORK OTHERSERV name [name name ...]
	char *name, *nextname;
	services_names *old, *sn;
	int namelen;

	if(datalen == 0){
		help_network_callback();
		return(1); // passthru to normal network command
	}

	name = skipword(data);
	namelen = strlen(name);

	if(!strcmp(data, "+")){
		// add new network
		sn = (services_names *)track_malloc(sizeof(services_names), "services network_callback");
		sn->network = NULL;
		sn->chanserv = NULL;
		sn->nickserv = NULL;
		sn->memoserv = NULL;
		sn->otherserv = NULL;
		setMakeStr(&sn->network, name);
		setMakeStr(&sn->chanserv, default_chanserv);
		setMakeStr(&sn->nickserv, default_nickserv);
		setMakeStr(&sn->memoserv, default_memoserv);
		sn->next = services_head;
		services_head = sn;

		// fix damage to data before returning
		if(namelen) *(name-1) = ' ';

		return(1); // passthru to normal network command
	} else if(!strcmp(data, "-")){
		// remove network
		for(old = sn = services_head; sn != NULL; old = sn, sn = sn->next){
			if(!strcmp(sn->network, name)){
				if(sn == services_head){
					services_head = sn->next;
				} else {
					old->next = sn->next;
				}
				if(sn->network) track_free(sn->network, "PLUG services");
				if(sn->chanserv) track_free(sn->chanserv, "PLUG services");
				if(sn->nickserv) track_free(sn->nickserv, "PLUG services");
				if(sn->memoserv) track_free(sn->memoserv, "PLUG services");
				if(sn->otherserv) track_free(sn->otherserv, "PLUG services");
				track_free(sn, "PLUG services");

				// fix damage to data before returning
				if(namelen) *(name-1) = ' ';

				return(1); // passthru to normal network command
			}
		}

		// fix damage to data before returning
		if(namelen) *(name-1) = ' ';

		return(1); // passthru to normal network command
	}

	stringtolower(data);

	for(sn = services_head; sn != NULL; sn = sn->next)
		if(!strcmp(sn->network, n_current->name))
			break;

	if(!strcmp(data, default_chanserv)){
		nextname = skipword(name);
		if(*nextname != '\0')
			help_network_callback();
		else if(namelen){
			setMakeStr(&sn->chanserv, name);
			msgServer(3, n_current, MFP_SERVICES_SET_ITEM_VALUE, 2, sn->chanserv, "service_chanserv");
		} else
			msgServer(3, n_current, MFP_SERVICES_ITEM_VALUE, 2, sn->chanserv, "service_chanserv");
		return(0); // dont passthru
	} else if(!strcmp(data, default_nickserv)){
		nextname = skipword(name);
		if(*nextname != '\0')
			help_network_callback();
		else if(namelen){
			setMakeStr(&sn->nickserv, name);
			msgServer(3, n_current, MFP_SERVICES_SET_ITEM_VALUE, 2, sn->nickserv, "service_nickserv");
		} else
			msgServer(3, n_current, MFP_SERVICES_ITEM_VALUE, 2, sn->nickserv, "service_nickserv");
		return(0); // dont passthru
	} else if(!strcmp(data, default_memoserv)){
		nextname = skipword(name);
		if(*nextname != '\0')
			help_network_callback();
		else if(namelen){
			setMakeStr(&sn->memoserv, name);
			msgServer(3, n_current, MFP_SERVICES_SET_ITEM_VALUE, 2, sn->memoserv, "service_memoserv");
		} else
			msgServer(3, n_current, MFP_SERVICES_ITEM_VALUE, 2, sn->memoserv, "service_memoserv");
		return(0); // dont passthru
	} else if(!strcmp(data, default_otherserv)){
		if(namelen){
			setMakeStr(&sn->otherserv, name);
			msgServer(3, n_current, MFP_SERVICES_SET_ITEM_VALUE, 2, sn->otherserv, "service_otherserv");
		} else
			msgServer(3, n_current, MFP_SERVICES_ITEM_VALUE, 2, sn->otherserv, "service_otherserv");
		return(0); // dont passthru
	}

	// fix damage to data before returning
	if(namelen) *(name-1) = ' ';

	return(1); // passthru to normal network command
}


s_network *process_servers_config( char *line, s_network *n ){
	char *name;
	services_names *sn;
	s_network *n2;

	name = skipword(line);
	stringtolower(line);
	if(!strcmp(line, "/network")){
		line = name;
		name = skipword(name);
		if(!strcmp(line, "+")){
			n2 = networkGet(name);
			if(n2 != NULL)
				msgServer(3, n2, MFP_SERVICES_RELOAD, 0);
			return(n2);
		}

		if(n == NULL) return(NULL);

		for(sn = services_head; sn != NULL; sn = sn->next)
			if(!strcmp(sn->network, n->name))
				break;
		if(sn == NULL) return(n);

		stringtolower(line);
		if(!strcmp(line, default_chanserv)){
			setMakeStr(&sn->chanserv, name);
			msgServer(3, n, MFP_SERVICES_SET_ITEM_VALUE, 2, name, "service_chanserv");
		} else if(!strcmp(line, default_nickserv)){
			setMakeStr(&sn->nickserv, name);
			msgServer(3, n, MFP_SERVICES_SET_ITEM_VALUE, 2, name, "service_nickserv");
		} else if(!strcmp(line, default_memoserv)){
			setMakeStr(&sn->memoserv, name);
			msgServer(3, n, MFP_SERVICES_SET_ITEM_VALUE, 2, name, "service_memoserv");
		} else if(!strcmp(line, default_otherserv)){
			setMakeStr(&sn->otherserv, name);
			msgServer(3, n, MFP_SERVICES_SET_ITEM_VALUE, 2, name, "service_otherserv");
		}
	}
	return(n);
}


////////////////////////////////////////////////////////////////////////// GLOBAL


int servicesInit( s_plug *self ){
	char *c_ptr, c_buffer[LEN_BUFFER_MSG], path[LEN_PATHNAME];
	services_names *sn;
	s_network *n;
	FILE *fp;

	if(!plugVersion(0,2,8,4)){
		msgServer(3, NULL, MF_PLUG_REQUIRES, 2, "0.2.8cvs4", "Services");
		return(0);
	}

	services_head = NULL;

	plugHookUser(self, "id",	CMD_ONLINE, &id_callback);
	plugHookUser(self, "ghost",	CMD_ONLINE, &ghost_callback);
	plugHookUser(self, "cop",	CMD_ONLINE|CMD_CHANNEL, &cop_callback);
	plugHookUser(self, "opme",	CMD_ONLINE|CMD_CHANNEL|CMD_JOINED, &opme_callback);
	plugHookUser(self, "memo",	CMD_ONLINE, &memo_callback);
	plugHookUser(self, "network",	0, &network_callback);

	plugHookServer(self, "PRIVMSG", &privmsg_callback);
	plugHookServer(self, "NOTICE", &notice_callback);

	if(config_in_progress) return(1);

	// install default names for all loaded networks
	for(n = n_head; n != NULL; n = n->next){
		sn = (services_names *)track_malloc(sizeof(services_names), "services init");
		sn->network = NULL;
		sn->chanserv = NULL;
		sn->nickserv = NULL;
		sn->memoserv = NULL;
		sn->otherserv = NULL;
		setMakeStr(&sn->network, n->name);
		setMakeStr(&sn->chanserv, default_chanserv);
		setMakeStr(&sn->nickserv, default_nickserv);
		setMakeStr(&sn->memoserv, default_memoserv);
		sn->next = services_head;
		services_head = sn;
	}

	// load servers config file
	snprintf(path, LEN_PATHNAME, ".cria/servers");
	if((fp = fopen(path, "r")) == NULL){
		snprintf(path, LEN_PATHNAME, "%s/.cria/servers", getenv("HOME"));
		fp = fopen(path, "r");
	}

	// return if servers file not found
	if(fp == NULL){
		msgServer(3, NULL, MFP_SERVICES_DEFAULTS, 0);
		return(1);
	}

	// process servers file
	n = NULL;
	while(fgets(c_buffer, 400, fp) != NULL){
		c_ptr = strip_whitespace(c_buffer, SIDE_BOTH);
		if(*c_ptr != '#' && strlen(c_ptr) > 0){
			n = process_servers_config(c_ptr, n);
		}
	}
	fclose(fp);

	return(1);
}


void servicesUnload( ){
	services_names *temp, *sn;

	sn = services_head;
	while(sn != NULL){
		temp = sn->next;
		if(sn->network) track_free(sn->network, "PLUG services unload");
		if(sn->chanserv) track_free(sn->chanserv, "PLUG services unload");
		if(sn->nickserv) track_free(sn->nickserv, "PLUG services unload");
		if(sn->memoserv) track_free(sn->memoserv, "PLUG services unload");
		if(sn->otherserv) track_free(sn->otherserv, "PLUG services unload");
		track_free(sn, "PLUG services unload");
		sn = temp;
	}
}
