#!/bin/bash
F=".cria/_formats"
V=".cria/_variables"
C=".cria/_commands"
CO=".cria/_colors"

[ -z "$1" ] && exit
AWK=$1


# build cgc_formats.h
  # /set MF* "string" --> MF*,
echo -e "// THIS FILE IS GENERATED FROM $F.*, DO NOT EDIT\n" > cgc_formats.h
echo "enum MSG_FORMATS {" >> cgc_formats.h
cat $F |grep "\/set" |${AWK} '{print "\t"$2","}' >> cgc_formats.h
echo "MSG_FORMATS_LAST };" >> cgc_formats.h


# build cgc_formats.c
  # /set MF* "string" --> "string",
echo -e "// THIS FILE IS GENERATED FROM $F.*, DO NOT EDIT\n" > cgc_formats.c
echo "char *default_formats[] = {" >> cgc_formats.c
cat $F |grep "\/set" |${AWK} '{printf("\t%s",$3); n = 4; while(n <= NF){printf(" %s",$n); n++}; printf(",\n")}' >> cgc_formats.c
echo "};" >> cgc_formats.c


# build cgc_setformats.c
  # /set MF* "string" --> {"lowercase(*)", &formats[MF*]},
echo -e "// THIS FILE IS GENERATED FROM $F.*, DO NOT EDIT\n" > cgc_setformats.c
echo -e "#include \"common.h\"\n" >> cgc_setformats.c
echo "s_set_item setMFItems[] = {" >> cgc_setformats.c
cat $F |grep "\/set" |${AWK} '{printf("\t{\"%s\", &formats[%s]},\n", tolower(substr($2, 3)), $2)}' >> cgc_setformats.c
echo -e "\t{0,0}\n};" >> cgc_setformats.c


# build cgc_setvars.c
echo -e "// THIS FILE IS GENERATED FROM $V.*, DO NOT EDIT\n" > cgc_setvars.c
echo -e "#include \"common.h\"\n" >> cgc_setvars.c
  # -INT var name --> {"lowercase(name)", &var},
echo "s_set_item setIntItems[] = {" >> cgc_setvars.c
cat $V |grep "\-INT" |${AWK} '{printf("\t{\"%s\", &%s},\n", tolower($3), $2)}' >> cgc_setvars.c
echo -e "\t{NULL, NULL}\n};" >> cgc_setvars.c
  # -CHAR var name --> {"lowercase(name)", &var},
echo "s_set_item setCItems[] = {" >> cgc_setvars.c
cat $V |grep "\-CHAR" |${AWK} '{printf("\t{\"%s\", &%s},\n", tolower($3), $2)}' >> cgc_setvars.c
echo -e "\t{NULL, NULL}\n};" >> cgc_setvars.c
  # -STR var name --> {"lowercase(name)", &var},
echo "s_set_item setStrItems[] = {" >> cgc_setvars.c
cat $V |grep "\-STR" |${AWK} '{printf("\t{\"%s\", &%s},\n", tolower($3), $2)}' >> cgc_setvars.c
echo -e "\t{NULL, NULL}\n};" >> cgc_setvars.c
  # -COLOR enum name --> {"lowercase(name)", &colors[enum]},
echo "s_set_item setColorItems[] = {" >> cgc_setvars.c
cat $V |grep "\-COLOR" |${AWK} '{printf("\t{\"%s\", &colors[%s]},\n", tolower($3), $2)}' >> cgc_setvars.c
echo -e "\t{NULL, NULL}\n};" >> cgc_setvars.c
  # -COLORLIST 1-8 --> "1", ... "8"
echo "char color_list[8][8] = {" >> cgc_setvars.c
cat $V |grep "\-LISTOFCOLORS" |${AWK} '{printf("\t%s",$2); n = 3; while(n <= 9){printf(" %s",$n); n++};}' >> cgc_setvars.c
echo -e "\n};" >> cgc_setvars.c


# build cgc_setcmds.c
  # -CMD function name --> {"lowercase(name)", &function},
echo -e "// THIS FILE IS GENERATED FROM $C.*, DO NOT EDIT\n" > cgc_setcmds.c
echo -e "#include \"common.h\"\n#include \"cgc_cmds.h\"\n" >> cgc_setcmds.c
echo "s_user_cmd user_commands[] = {" >> cgc_setcmds.c
cat $C |grep "\-CMD" |${AWK} '{printf("\t{\"%s\", %s, &%s},\n", tolower($3), $4, $2)}' >> cgc_setcmds.c
echo -e "\t{NULL, 0, NULL}\n};" >> cgc_setcmds.c


# build cgc_cmds.h
  # -CMD function name --> void function ( char *data, int datalen );
echo -e "// THIS FILE IS GENERATED FROM $C.*, DO NOT EDIT\n" > cgc_cmds.h
cat $C |grep "\-CMD" |${AWK} '{printf("void %s ( char *data, int datalen );\n", $2)}' >> cgc_cmds.h


# build cgc_colors.c
  # /set name fg bg attr --> setRawColor(C_BG, uppercase(fg), uppercase(bg), uppercase(attr)s/,/|/ );
echo -e "// THIS FILE IS GENERATED FROM $CO.*, DO NOT EDIT\n" > cgc_colors.c
echo -e "#include \"common.h\"\n#include \"functions.h\"\n" >> cgc_colors.c
echo -e "void default_color_config( ){" >> cgc_colors.c
cat $CO |grep "\/set" |sed s/\,/\|/g |sed s/color_/c_/g \
	|${AWK} '{if($5 == "") $5 = "0"; printf("\tsetRawColor(%s, %s, %s, %s);\n", \
	toupper($2), toupper($3), toupper($4), toupper($5))}' >> cgc_colors.c
echo -e "}" >> cgc_colors.c
