//	cria - command processing
//	Copyright (c) 2001-2010 Brand Huntsman
//


#include <sys/time.h>		// gettimeofday() used by ping
#include <time.h>			// time() and localtime()
#include <sys/socket.h>		// getsockname()
#include <netinet/in.h>		// sockaddr
#include <arpa/inet.h>		// inet_ntoa()

#include "common.h"
#include "functions.h"

#ifndef __socklen_t_defined
  typedef unsigned int socklen_t;
#endif


//////////////////////////////////////////////////////////////////////////

// global
int message_to_server, display_motd;
char *msg_host;
s_alias *alias_head;
s_bind *bind_head;

// local
uint list_visible_channels;

////////////////////////////////////////////////////////////////////////// PRIVATE


int command_to_number( char *command ){
	if(command[0] >= '0' && command[0] <= '9' &&
	   command[1] >= '0' && command[1] <= '9' &&
	   command[2] >= '0' && command[2] <= '9')

		return(cria_atoui(command));
	else return(-1);
}


////////////////////////////////////////////////////////////////////////// GLOBAL


void cmdInit( ){
	// message_to_server is cleared by cmdFromServer()
	display_motd = 0;
	msg_host = NULL;
	alias_head = NULL;
	bind_head = NULL;
}


void cmdFromServer( s_network *network, char *data ){
	char m_buffer[LEN_BUFFER_MSG];
	char *command, *from, *host, *nick, *channel, *message, *d1, *d2, *d3, *d4, *d5;
	int cmd_nr, cmdlen, n;
	double new_time, old_time;
	s_page *page;
	struct tm *tp;
	time_t ct;
	struct timeval tv;
	s_plug_server *plug;
	s_plug_server_int *iplug;
	socklen_t addr_len;
	struct sockaddr_in local;

#ifdef __DEBUG__
  out_to_file("---", data);
#endif

	message_to_server = 0;

	from = data;
	data = skipword(data);
	msg_host = NULL;

	if(*from == ':'){
		from++;
		command = data;
		data = skipword(data);
		cmdlen = strlen(command);

		for(plug = plug_server_head; plug != NULL; plug = plug->next)
			if(!strcmp(plug->command, command))
				if(!(*plug->callback)(network, from, data, strlen(data)))
					return;

//------------------------------
		if(cmdlen == 3){
//------------------------------
			if((cmd_nr = command_to_number(command)) == -1){
				msgServer(3, network, MF_UNKNOWN_SERVER_COMMAND, 3, data, command, from);
				return;
			}

			for(iplug = plug_server_int_head; iplug != NULL; iplug = iplug->next)
				if(iplug->command == cmd_nr)
					if(!(*iplug->callback)(network, from, data, strlen(data)))
						return;

			switch(cmd_nr){
			case 1:
				// * 001 * :Welcome to the Internet Relay Network <nick>

				message = skipword(data);
				if(*message == ':') message++;
				msgServer(3, network, MF_INFO, 1, message);

				// grab nick and update if different
				nick = message + strlen(message);
				while(*(nick-1) != ' ') nick--;
				skipnick(nick);
				if(strcmp(nick, network->nickname)){
					setMakeStr(&network->nickname, nick);
					msgServer(-1, network, MF_NICK_CHANGE_SELF, 1, nick);
				}

				// grab real hostname for server
				setMakeStr(&network->server_name, from);
				barsShow();

				break;

			case 2:
				// * 002 * :Your host is <servername>, running version <version>

				message = skipword(data);
				if(*message == ':') message++;
				msgServer(3, network, MF_INFO, 1, message);

				break;

			case 3:
				// * 003 * :This server was created <date>

				message = skipword(data);
				if(*message == ':') message++;
				msgServer(3, network, MF_INFO, 1, message);

				break;

			case 4:
				// * 004 * <server> <version> <avail umodes> <avail chan modes>

				message = skipword(data);
				msgServer(3, network, MF_INFO, 1, message);

				d1 = skipword(message);	// <version>
				d2 = skipword(d1);		// <umodes>
				d3 = skipword(d2);		// <chan modes>

				setMakeStr(&network->server_version, d1);
				setMakeStr(&network->user_modes, d2);
				setMakeStr(&network->channel_modes, d3);

				// send user mode
				modeToString(data, network->umode);
				netSend(network, "MODE %s %s\n", network->nickname, data);

				// load network config file
				if(network->flags.x.firstconn){
					network->flags.x.firstconn = 0;
					n_current = network;
					pageShow(p_server);
					snprintf(m_buffer, LEN_BUFFER_MSG, "nc.%s", network->name);
					configLoadFile(m_buffer, 1, LOAD_NOISY);
				}

				// get local IP address
				addr_len = sizeof(struct sockaddr_in);
				getsockname(network->socket, (struct sockaddr *)&local, &addr_len);
				setMakeStr(&network->local_ip_addr, inet_ntoa(local.sin_addr));

				if(!network->flags.x.registered){
					// rejoin all channels
					for(page = p_head; page != NULL; page = page->next){
						if(page->network == network && page->name[0] == '#' && !page->flags.x.unjoined){
							// join channel
							if(page->host_key)
								netSend(network, "JOIN %s %s\n", page->name, page->host_key);
							else netSend(network, "JOIN %s\n", page->name);

							// get channel mode
							netSend(network, "MODE %s\n", page->name);
						}
					}

					network->flags.x.registered = 1;
				}

				break;

			case 5:
				// * 005 * :Try server <server name>, port <port number>
				// * 005 * <server features> :are available on this server

				message = skipword(data);
				if(*message == ':'){
					msgGotoServer(network, MF_ERROR, 1, message+1);
				} else {
					msgServer(3, network, MF_INFO, 1, message);

  // TODO: strip the " :are available on this server" before storing

					setMakeStr(&network->server_features, message);
				}

				break;

  // TODO: 200-221

  // TODO: 241-244

			case 250:
			case 251:
			case 252:
			case 253:
			case 254:
			case 255:
			case 256:
			case 257:
			case 258:
			case 259:
				// * 250 * :Highest connection count: <integer> (<integer> clients)
				//		(<integer> since server was (re)started)
				// * 251 * :There are <integer> victims and <integer>
				//		hiding on 25 servers
				// * 252 * <integer> :operator(s) online
				// * 253 * <integer> :unknown connection(s)
				// * 254 * <integer> :channels formed
				// * 255 * :I have <integer> clients and <integer> servers
				// * 256 * <server> :Administrative info
				// * 257 * :<admin loc1 info>
				// * 258 * :<admin loc2 info>
				// * 259 * :<admin email info>

				message = skipword(data);

				if(*message == ':'){
					message++;
					msgServer(3, network, MF_INFO, 1, message);
				} else {
					d1 = skipword(message) + 1;
					snprintf(m_buffer, LEN_BUFFER_MSG, "%s %s", message, d1);
					msgServer(3, network, MF_INFO, 1, m_buffer);
				}

				break;

  // TODO: 261

			case 265:
			case 266:
				// * 265 * :Current local  users: 1013  Max: 1273
				// * 266 * :Current global users: 4104  Max: 4853

				message = skipword(data) + 1;

				msgServer(3, network, MF_INFO, 1, message);

				break;

			case 300:
				// ignore, dummy reply

				break;

			case 301:
				// * 301 * <nick> :<away msg>

				nick = skipword(data);			// nick = <nick>
				message = skipword(nick) + 1;	// message = <away msg>

				msgUnkPage(network, nick, 1, NULL, MF_WHOIS_AWAY, 2, message, nick);

				break;

			case 302: // USERHOST
				// * 302 * :nick=+host
				// * 302 * :nick=-host
				// * 302 * :nick*=+host
				// * 302 * :nick*=-host

				nick = skipword(data) + 1;		// nick = nick
				host = skipuserhost(nick);		// host = host

  // TODO: the * "indicates whether the client has registered as an Operator"
  // TODO: the - means the client is away, + means the client is not away

  // TODO: display this information in the MF_USERHOST format

				if(strlen(nick) > 0){
					if((page = pageGet(network, nick)) != NULL){
						if(page->host_key == NULL){
							setMakeStr(&page->host_key, host);
							if(p_current == page) barsShow();
						}
					} else {
						msgGotoServer(network, MF_USERHOST, 2, host, nick);
					}
				}

				break;

			case 303: // ISON
				// * 303 * :<nick>		<-- user online
				// * 303 * :			<-- user not online

				nick = skipword(data) + 1;

				if(strlen(nick) > 0)
					msgServer(0, network, MF_ISON, 1, nick);

				break;

			case 305: // BACK
				// * 305 * :Does this mean you're really back?

				message = skipword(data) + 1;

				network->flags.x.away = 0;
				msgServer(-1, network, MF_DEBUG, 1, message);

				if(p_current == p_server) barsShow();

				break;

			case 306: // AWAY
				// * 306 * :We'll miss you

				message = skipword(data) + 1;

				network->flags.x.away = 1;
				msgServer(-1, network, MF_DEBUG, 1, message);

				if(p_current == p_server) barsShow();

				break;

			case 311: // WHOIS USER
				// * 311 * <nick> <user> <host> * :<realname>

				nick = skipword(data);					// nick = <nick>
				d1 = skipword(nick);					// d1 = <user>
				host = skipword(d1);					// host = <host>
				data = skipword(skipword(host)) + 1;	// data = <realname>

				msgUnkPage(network, nick, 1, NULL, MF_WHOIS_USER, 3, host, d1, nick);
				msgUnkPage(network, nick, 1, NULL, MF_WHOIS_NAME, 2, data, nick);

				break;

			case 312: // WHOIS SERVER
				// * 312 * <nick> <server> :<server info>

				nick = skipword(data);			// nick = <nick>
				host = skipword(nick);			// host = <server>
				d1 = skipword(host) + 1;		// d1 = <server info>

				msgUnkPage(network, nick, 1, NULL, MF_WHOIS_SERVER, 3, d1, host, nick);

				break;

			case 313: // WHOIS OPERATOR
				// * 313 * <nick> :is an IRC operator

				nick = skipword(data);
				skipword(nick);

				msgUnkPage(network, nick, 1, NULL, MF_WHOIS_OPERATOR, 1, nick);

				break;

			case 314: // WHO WAS USER
				// * 314 * <nick> <user> <host> * :<realname>

				nick = skipword(data);					// nick = <nick>
				d1 = skipword(nick);					// d1 = <user>
				host = skipword(d1);					// host = <host>
				data = skipword(skipword(host)) + 1;	// data = <realname>

				msgUnkPage(network, nick, 1, NULL, MF_WHOWAS_USER, 3, host, d1, nick);
				msgUnkPage(network, nick, 1, NULL, MF_WHOIS_NAME, 2, data, nick);

				break;

			case 315:
				// ignore, who end

				break;

  // TODO:		case 316: // WHOIS CHAN OP
  //				break;

			case 317: // WHOIS IDLE
				// * 317 * <nick> <time> :seconds idle

				nick = skipword(data);		// nick = <nick>
				data = skipword(nick);
				skipword(data);				// data = <time>

				{
					char days[5], hours[3], minutes[3], seconds[3];
					uint time;

					time = cria_atoui(data);

					snprintf(days, 5, "%u", time/86400);
					time = time%86400;

					snprintf(hours, 3, "%u", time/3600);
					time = time%3600;

					snprintf(minutes, 3, "%u", time/60);
					time = time%60;

					snprintf(seconds, 3, "%u", time);

					msgUnkPage(network, nick, 1, NULL, MF_WHOIS_IDLE, 5, seconds, minutes, hours, days, nick);
				}

				break;

			case 318:
				// ignore, whois end

				break;

			case 319: // WHOIS CHANNELS
				// * 319 * <nick> :<@+channels>

				nick = skipword(data);			// nick = <nick>
				message = skipword(nick) + 1;	// message = <@+channels>

				msgUnkPage(network, nick, 1, NULL, MF_WHOIS_CHANNELS, 2, message, nick);

				break;

			case 320:
				// * 320 * <nick> :is an identified user

				nick = skipword(data);			// nick = <nick>
				message = skipword(nick) + 1;	// message = <message>

				msgUnkPage(network, nick, 1, NULL, MF_WHOIS_320, 2, message, nick);

				break;

			case 321: // LIST BEGIN
				list_visible_channels = 0;

				break;

			case 322: // LIST
				// * 322 * <channel> <visible users> :<topic>

				channel = skipword(data);		// channel = <channel>
				d1 = skipword(channel);			// d1 = <visible users>
				d2 = skipword(d1) + 1;			// d2 = <topic>

				if(*channel != '*'){
					msgGotoServer(network, MF_LIST_CHANNELS, 3, d2, d1, channel);
					list_visible_channels++;
				}

				break;

			case 323: // LIST END
				snprintf(m_buffer, LEN_BUFFER_MSG, "%d", list_visible_channels);

				msgGotoServer(network, MF_TOTAL_CHANNELS, 1, m_buffer);

				break;

			case 324: // CHANNEL MODE
				// * 324 * <channel> <mode> <limit> <key>

				d1 = skipword(data);			// d1 = <channel>
				data = skipword(d1);			// data = <mode> <limit> <key>

				if((page = pageGet(network, d1))){
					msgPage(page, 3, NULL, MF_SERVER_MODE, 2, data, from);
					page->mode = 0;
					cmodeToInt(page, data);
					barsShow();
				}

				break;

  // TODO:	case 328:
				// * 328 * <channel> :<url>
				// DALNET
  //				break;

			case 329: // CHANNEL CREATION TIME
				// * 329 * <channel> <time>

				channel = skipword(data);		// channel = <channel>
				data = skipword(channel);		// data = <time>

				if((page = pageGet(network, channel))){
					ct = cria_atoui(data);
					tp = localtime(&ct);
					strftime(m_buffer, LEN_BUFFER_MSG, var_time_format, tp);

					msgPage(page, 3, NULL, MF_CHANNEL_CREATION, 1, m_buffer);
				}

				break;

			case 331:
				// * 331 * <channel> :There isn't a topic.

				channel = skipword(data);			// channel = <channel>
				message = skipword(channel) + 1;	// message = <topic>

				page = msgUnkPage(network, channel, 3, NULL, MF_TOPIC, 1, message);
				if(page) setMakeStr(&page->topic, message);
				barsShow();

				break;

			case 332: // TOPIC
				// * 332 * <channel> :<topic>

				channel = skipword(data);			// channel = <channel>
				message = skipword(channel) + 1;	// message = <topic>

				page = msgUnkPage(network, channel, 3, NULL, MF_TOPIC, 1, message);
				if(page) setMakeStr(&page->topic, message);
				barsShow();

				break;

			case 333: // TOPIC SET BY
				// * 333 * <channel> <setby_nick> <time>

				channel = skipword(data);		// channel = <channel>
				nick = skipword(channel);		// nick = <setby_nick>
				data = skipword(nick);			// data = <time>

				// irc.mozilla.org uses nick!user@host
				skipnick(nick);

				ct = cria_atoui(data);
				tp = localtime(&ct);
				strftime(m_buffer, LEN_BUFFER_MSG, var_time_format, tp);

				msgUnkPage(network, channel, 3, NULL, MF_TOPIC_SETBY, 2, m_buffer, nick);

				break;

			case 341: // INVITE USER TO CHANNEL
				// * 341 * <nick> <channel>

				nick = skipword(data);
				channel = skipword(nick);

				msgUnkPage(network, channel, -1, NULL, MF_SENT_INVITE, 1, nick);

				break;

  // TODO:		case 342:
				// <user> :Summoning user to IRC
  //				break;

			case 351: // SERVER VERSION
				// * 351 * <version>.<debuglevel> <server> :<comments>

				data = skipword(data);			// data = <version>
				host = skipword(data);			// host = <server>
				message = skipword(host) + 1;	// message = <comments>

				msgServer(0, network, MF_SERVER_VERSION_REPLY, 3, message, host, data);

				break;

			case 352: // WHO chan
				// * 352 * <channel> <user> <host> <server> <nick> HG*@+ :<hopcount> <realname>

				channel = skipword(data);	// channel = channel>
				d1 = skipword(channel);		// d1 = <user>
				d2 = skipword(d1);			// d2 = <host>
				host = skipword(d2);		// host = <server>
				nick = skipword(host);		// nick = <nick>
				d3 = skipword(nick);		// d3 = HG*@+
				d4 = skipword(d3) + 1;		// d4 = <hopcount>
				d5 = skipword(d4);			// d5 = <realname>

				if((page = pageGet(network, channel)) == NULL)
					msgServer(3, network, MF_WHO, 7,
						host, d4, d5, d2, d1, d3, nick);
				else
					msgPage(page, 3, NULL, MF_WHO, 7,
						host, d4, d5, d2, d1, d3, nick);

				break;

			case 353: // CHANNEL NAMES LIST
				// * 353 * = <channel> :nick1 @nick2 +nick3

				d1 = skipword(data);			// =
				channel = skipword(d1);			// channel = <channel>
				d1 = skipword(channel) + 1;		// d1 = nick1 @nick2 +nick3

				if(strlen(d1) > 0){
					n = 1;
					nick = d1;
					while(*nick != '\0'){
						nick++;
						if(*nick == ' ') n++;
					}
					snprintf(m_buffer, LEN_BUFFER_MSG, "%d", n);
					page = msgUnkPage(network, channel, 3, NULL, MF_NAMES, 2,
						d1, m_buffer);

					if(page){
						// build nicklist
						while(*d1 != '\0'){
							nick = skipword(d1);
							if(*d1 == '@' || *d1 == '%' || *d1 == '+'){
								if(stringcompare(d1+1, network->nickname)){
									switch(*d1){
									case '@':
										page->flags.x.chanop = 1;
										barsShow();
										break;
									case '%':
										page->flags.x.halfchanop = 1;
										barsShow();
										break;
									case '+':
										page->flags.x.voice = 1;
										barsShow();
										break;
									}
								}

  // TODO: keep track of number of chanops, halfchanops and voiced users in channel

								d1++;
							}
							nicksAddNick(page, d1);
							d1 = nick;
						}
					}
				}

				break;

  // TODO:		case 364: // LINKS
				// <mask> <server> :<hopcount> <server info>
  //				break;

			case 365:
				// ignore, end of links

				break;

			case 366:
				// ignore, names end

				break;

			case 367: // BAN LIST
				// * 367 * <channel> <mask> <by_nick!by_user@by_host> <time>
				// * 367 * <channel> %<mask> <by_nick!by_user@by_host> <time>

				channel = skipword(data);	// channel = <channel>
				d1 = skipword(channel);		// d1 = <mask>
				data = skipword(d1);
				d2 = skipword(data);		// d2 = <time>
				host = skipnick(data);		// host = <by_user@by_host>
											// nick = <by_nick>

				ct = cria_atoui(d2);
				tp = localtime(&ct);
				strftime(m_buffer, LEN_BUFFER_MSG, var_time_format, tp);

				if(*d1 == '%'){
					d1++;
					msgUnkPage(network, channel, 1, NULL, MF_LIST_QUIET, 4,
						m_buffer, host, data, d1);
					total_quiet++;
				} else {
					msgUnkPage(network, channel, 1, NULL, MF_LIST_BANS, 4,
						m_buffer, host, data, d1);
					total_bans++;
				}
				break;

			case 368:
				// * 368 * <channel> :End of channel ban list

				channel = skipword(data);		// channel = <channel>
				skipword(channel);

				if(display_ban_totals){
					snprintf(m_buffer, LEN_BUFFER_MSG, "%d", total_bans);
					msgUnkPage(network, channel, 1, NULL, MF_TOTAL_BANS, 1, m_buffer);
					if(total_quiet){
						snprintf(m_buffer, LEN_BUFFER_MSG, "%d", total_quiet);
						msgUnkPage(network, channel, 1, NULL, MF_TOTAL_QUIET, 1, m_buffer);
					}
					display_ban_totals = 0;
				}
				break;

			case 369:
				// ignore, whowas end

				break;

			case 371: // INFO
				// * 371 * :<message>

				message = skipword(data) + 1;

				msgServer(3, network, MF_INFO, 1, message);

				break;

			case 372: // MOTD
				// * 372 * :message

				if(var_motd || display_motd){
					message = skipword(data) + 1;

					msgServer(3, network, MF_MOTD, 1, message);
				}

				break;

			case 374:
				// ignore, end of info

				break;

			case 375:
				// ignore, motd begin

				break;

			case 376:
				// ignore, motd end

				display_motd = 0;
				break;

			case 379:
				// * 379 * <channel1> <channel2> :Forwarding to another channel

				channel = skipword(data);		// channel = channel1
				data = skipword(channel);		// data = channel2
				message = skipword(data);		// message = message

				pageFree(pageGet(network, channel));

				msgUnkPage(network, data, 3, NULL, MF_FORWARD_CHAN, 2, data, channel);

				break;

  // 381: handled by ircop plugin
  // 382: handled by ircop plugin

			case 391:
				// * 391 * <server> :<server time>

				d1 = skipword(skipword(data)) + 1;	// d1 = <server time>

				msgServer(3, network, MF_INFO, 1, d1);

				break;

			case 392:
				// ignore, users start

				break;

  // TODO:		case 393: // USERS
				// :%-8s<userID> %-9s<terminal> %-8s<host>
  //				break;

			case 394:
				// ignore, end of users

				break;

  // TODO:		case 395: // NO USERS
				// :Nobody logged in
  //				break;

			case 401:
				// * 401 * <nick/channel> :No such nick/channel

				data = skipword(data);			// data = <nick/channel>
				skipword(data);

				if(*data == '#'){
					if((page = pageGet(network, data)) == NULL)
						msgGotoServer(network, MF_NO_SUCH_CHAN, 1, data);
					else
						msgPage(page, 3, NULL, MF_NO_SUCH_CHAN, 1, data);
				} else {
					if((page = pageGet(network, data)) == NULL){
						msgGotoServer(network, MF_NO_SUCH_NICK, 1, data);
					} else {
						msgPage(page, 3, NULL, MF_NO_SUCH_NICK, 1, data);

						if(page->host_key != NULL){
							track_free(page->host_key, "401: host_key");
							page->host_key = NULL;
							if(p_current == page) barsShow();
						}
					}
				}

				break;

			case 402:
				// * 402 * <server name> :No such server
				// * 402 * =<pong time> :No such server
				// * 402 * <pong time> :No such server

				data = skipword(data);			// data = <server name>
				skipword(data);

				if(*data == '='){
					d1 = data + 1;				// d1 = seconds
					d2 = findchar(d1, '.');		// d2 = microseconds
					*d2 = '\0';
					d2++;

					gettimeofday(&tv, NULL);
					new_time = (double)tv.tv_sec + (double)tv.tv_usec/1000000;
					old_time = (double)(cria_atoui(d1)) + (double)cria_atoui(d2)/1000000;
					if(new_time - old_time <= LAG_PING_TIME - 1 ){
						network->lag = new_time - old_time;
						network->pong_received = 1;
					}

					barsShow();
				} else if(*data >= '0' && *data <= '9'){
					d1 = data;					// d1 = seconds
					d2 = findchar(d1, '.');		// d2 = microseconds
					*d2 = '\0';
					d2++;

					gettimeofday(&tv, NULL);
					new_time = (double)tv.tv_sec + (double)tv.tv_usec/1000000;
					old_time = (double)(cria_atoui(d1)) + (double)cria_atoui(d2)/1000000;
					snprintf(m_buffer, LEN_BUFFER_MSG, "%.3f seconds", new_time - old_time);

					msgServer(0, network, MF_SERVER_PING_REPLY, 2, from, m_buffer);
				} else {
					msgGotoServer(network, MF_NO_SUCH_SERVER, 1, data);
				}

				break;

			case 403:
				// ignore, no such channel
				// * 403 * <channel> :No such channel

				break;

			case 404:
				// * 404 * <channel> :The channel demigods have stolen your voice

				channel = skipword(data);
				skipword(channel);

				msgUnkPage(network, channel, 3, NULL, MF_CANT_TALK, 0);

				break;

			case 405:
				// * 405 * <channel name> :You have joined too many channels

				data = skipword(data);

  // TODO: finish

				msgGotoServer(network, MF_ERROR, 1, data);

				break;

			case 406: // invalid WHOWAS
				// * 406 * <nick> :There was no such nickname

				nick = skipword(data);
				skipword(nick);

				msgGotoServer(network, MF_WHOWAS_UNKNOWN, 1, nick);

				break;

  // TODO: 407-414
  //	* 409 * :No origin specified

			case 412:
				// * 412 * :No text to send

				msgCurrentPage(3, NULL, MF_EMPTY_MESSAGE, 0);

				break;

  // TODO:		case 421:
				// * 421 * <command> :Unknown command
  //				break;

			case 422:
				// * 422 * :MOTD File is missing

				if(var_motd){
					message = skipword(data) + 1;

					msgServer(3, network, MF_DEBUG, 1, message);
				}

				break;

  // TODO: 423-424, 431

			case 432:
				// * 432 <nick> :Erroneous Nickname

				skipword(data);			// data = <nick>

				msgGotoServer(network, MF_INVALID_NICK, 1, data);

				break;

			case 433:
				// * 433 * <nick> :Nickname is already in use.

				nick = skipword(data);
				skipword(nick);

				msgGotoServer(network, MF_NICKINUSE, 1, nick);

				if(network->server_name != NULL){
					netSend(network, "USERHOST %s\n", nick);
				} else {
					netSend(network, "NICK %s%c\n", nick, var_nick_char);

					snprintf(m_buffer, LEN_BUFFER_MSG, "%s%c", nick, var_nick_char);
					setMakeStr(&network->nickname, m_buffer);
				}

				break;

  // TODO: 436

			case 441:
				// * 441 * <nick> <channel> :They aren't on that channel

				nick = skipword(data);			// nick = <nick>
				channel = skipword(nick);		// channel = <channel>
				skipword(channel);

				msgUnkPage(network, channel, 3, NULL, MF_NICK_NOT_IN_CHAN, 1, nick);

				break;

			case 442:
				// * 442 * <channel> :You're not on that channel

				channel = skipword(data);
				skipword(channel);

				msgUnkPage(network, channel, -1, NULL, MF_NOT_IN_CHAN, 0);

				break;

			case 443:
				// * 443 * <nick> <channel> :is already on channel

				nick = skipword(data);		// nick = <nick>
				channel = skipword(nick);
				skipword(channel);			// channel = <channel>

				msgUnkPage(network, channel, -1, NULL, MF_NICK_IN_CHANNEL, 1, nick);

				break;

  // TODO: 444-446

			case 451:
				// * 451 * <command> :Register first.
				// * 451 <command> :You have not registered

				if(data[0] == '*' && data[1] == ' ')
					data = &data[2];
				skipword(data);

				msgGotoServer(network, MF_NOT_REGISTERED, 1, data);

				break;

  // TODO: 461-467, 471

			case 472:
				// * 472 * <char> :is unknown mode char to me

				data = skipword(data);			// data = <char>
				skipword(data);

				msgGotoServer(network, MF_UNKNOWN_MODE, 1, data);

				break;

			case 473:
				// * 473 * <channel> :You need to be invited to that channel

				channel = skipword(data);
				skipword(channel);

				msgUnkPage(network, channel, 3, NULL, MF_INVITE_ONLY, 0);

				break;

			case 474:
				// * 474 * <channel> :You're banned from that channel

				channel = skipword(data);			// channel = <channel>
				message = skipword(channel) + 1;	// message = <message>

				msgUnkPage(network, channel, 3, NULL, MF_BANNED, 0);

				break;

			case 475:
				// * 475 * <channel> :You must specify the channel key (+k)

				channel = skipword(data);
				skipword(channel);

				msgUnkPage(network, channel, 3, NULL, MF_KEY_REQUIRED, 0);

				break;

			case 479:
				// * 479 * <channel> :Illegal channel name

				channel = skipword(data);
				skipword(channel);

				msgGotoServer(network, MF_INVALID_CHANNEL, 1, channel);
				pageFree(pageGet(network, channel));

				break;

  // 481: handled by ircop plugin

			case 482:
				// * 482 * <channel> :You need to be a channel operator to do that

				channel = skipword(data);
				skipword(channel);

				msgUnkPage(network, channel, 3, NULL, MF_NOT_CHANOP, 0);

				break;

  // TODO: 483

  // 491: handled by ircop plugin

  // TODO: 501-502

			default:
				msgServer(3, network, MF_UNKNOWN_SERVER_COMMAND, 3, data, command, from);
			}

//-------------------------------------
		} else if(cmdlen == 4){
//-------------------------------------
			if(!strcmp(command, "PONG")){
				// <server> PONG <server name> :<server name>
				// <server> PONG =<time> :*
				// <server> PONG <time> :*

				skipword(data);					// data = <time>|<server name>

				if(*data == '='){
					d1 = data + 1;				// d1 = seconds
					d2 = findchar(d1, '.');
					*d2 = '\0';
					d2++;						// d2 = microseconds

					gettimeofday(&tv, NULL);
					new_time = (double)tv.tv_sec + (double)tv.tv_usec/1000000;
					old_time = (double)(cria_atoui(d1)) + (double)cria_atoui(d2)/1000000;
					if(new_time - old_time <= LAG_PING_TIME - 1 ){
						network->lag = new_time - old_time;
						network->pong_received = 1;
					}

					barsShow();
				} else if(*data >= '0' && *data <= '9'){
					d1 = data;					// d1 = seconds
					d2 = findchar(d1, '.');
					*d2 = '\0';
					d2++;						// d2 = microseconds

					gettimeofday(&tv, NULL);
					new_time = (double)tv.tv_sec + (double)tv.tv_usec/1000000;
					old_time = (double)(cria_atoui(d1)) + (double)cria_atoui(d2)/1000000;
					snprintf(m_buffer, LEN_BUFFER_MSG, "%.3f seconds", new_time - old_time);

					msgServer(0, network, MF_SERVER_PING_REPLY, 2, from, m_buffer);
				} else {
					// server doesn't return timestamp, disable lag meter
					network->lag = -1;
				}

			} else if(!strcmp(command, "PART")){
				// <nick>!<host> PART <channel> :<message>

				host = skipnick(from);			// host = <host>
												// from = <nick>
				message = skipword(data) + 1;	// message = <message>
												// data = <channel>

				if(ignoreFind(IG_PART, data, from, host, message)) return;

  // TODO: can part messages contain beeps?
				beep_once(message);

				if(!strcmp(from, network->nickname)){
					// only display message if page is still open

					if((page = pageGet(network, data)) != NULL){
						// you left channel, page still open
						if(page->flags.x.cycled){
							// cycling
							msgPage(page, 3, NULL, MF_CYCLE_CHANNEL, 0);
						} else {
							msgPage(page, 3, NULL, MF_LEAVE_CHANNEL, 0);
						}
						page->flags.x.cycled = 0;
						page->flags.x.joined = 0;

						nicksFree(page);
					}
				} else {
					// user left channel

					if(var_chop_host) chop_userhost(host);
					page = msgUnkPage(network, data, 3, NULL, MF_PART, 3,
						message, host, from);
					nicksRemoveNick(page, from);
				}

			} else if(!strcmp(command, "QUIT")){
				// <nick>!<host> QUIT :<message>

				host = skipnick(from);		// host = <host>
											// from = <nick>
				data++;						// data = <message>

  // TODO: can quit messages contain beeps?
				beep_once(data);

				for(page = p_head; page != NULL; page = page->next){
					if(page->nicklist != NULL && page->network == network){
						if(nicksRemoveNick(page, from)){
							if(!ignoreFind(IG_QUIT, page->name, from, host, data))
								msgPage(page, 3, NULL, MF_QUIT, 3,
									data, host, from);
						}
					}
				}

				if((page = pageGet(network, from)) != NULL){
					msgPage(page, 3, NULL, MF_QUIT, 3, data, host, from);
					if(page->host_key != NULL){
						track_free(page->host_key, "commandFromServer: host_key");
						page->host_key = NULL;
					}
					barsShow();
				}

			} else if(!strcmp(command, "JOIN")){
				// <nick>!<host> JOIN :<channel>
				// <nick>!<host> JOIN <channel>

				host = skipnick(from);		// host = <host>
											// from = <nick>
				if(*data == ':') data++;	// data = <channel>

				if(ignoreFind(IG_JOIN, data, from, host, "")) return;

				if(strcmp(from, network->nickname)){
					if(var_chop_host) chop_userhost(host);

					page = msgUnkPage(network, data, 2, NULL, MF_JOIN, 2, host, from);

					nicksAddNick(page, from);

					// send join to message page, if one
					if((page = pageGet(network, from)) != NULL){
						if(page->host_key == NULL){
							msgPage(page, 2, NULL, MF_JOIN, 2, host, from);

							netSend(network, "USERHOST %s\n", from);
						}
					}
				} else {
					if((page = pageGet(network, data)) != NULL){
						page->flags.x.joined = 1;
						page->flags.x.chanop = page->flags.x.halfchanop =
							page->flags.x.voice = page->flags.x.quiet = 0;
						barsShow();

						// free nicks, just in case
						nicksFree(page);
					} else {
						page = pageCreate(network, data, "");
						if(page){
							page->flags.x.joined = 1;
							pageShow(page);
						}

						// get channel mode
						netSend(network, "MODE %s\n", data);
					}
				}

			} else if(!strcmp(command, "MODE")){
				// * MODE * :<usermode>
				// * MODE <channel> <mode>
				// <nick>!<host> MODE <channel> <mode>

				d1 = skipword(data);			// d1 = <mode> or :<usermode>
												// data = <channel> or *

				if(*data == '#'){
					host = skipnick(from);		// from = nick
												// host = host

					if(strlen(host) > 0)
						page = msgUnkPage(network, data, 3, NULL, MF_MODE, 3,
							d1, host, from);
					else
						page = msgUnkPage(network, data, 3, NULL, MF_SERVER_MODE, 2,
							d1, from);
					cmodeToInt(page, d1);
					barsShow();
				} else {
					d1++;
					msgServer(3, network, MF_UMODE, 1, d1);
					network->umode = umodeToInt(network->umode, d1);
					barsShow();
				}

			} else if(!strcmp(command, "NICK")){ // NICK CHANGE
				// <old_nick>!<host> NICK :<new_nick>

				host = skipnick(from);		// host = <host>
											// from = <old_nick>
				data++;						// data = <new_nick>

				if(strcmp(network->nickname, from)){
					if(var_chop_host) chop_userhost(host);

					for(page = p_head; page != NULL; page = page->next){
						if(page->nicklist != NULL && page->network == network){
							if(nicksRemoveNick(page, from)){
								nicksAddNick(page, data);
								if(!ignoreFind(IG_NICK, page->name, from, host, ""))
									msgPage(page, 3, NULL, MF_NICK_CHANGE, 3,
										data, host, from);
							}
						}
					}

					if((page = pageGet(network, from)) != NULL){
						msgPage(page, 3, NULL, MF_NICK_CHANGE, 3, data, host, from);
						setMakeStr(&page->name, data);
						barsShow();

						if(page->host_key == NULL)
							netSend(network, "USERHOST %s\n", data);
					} else if((page = pageGet(network, data)) != NULL){
						msgPage(page, 3, NULL, MF_NICK_CHANGE, 3, data, host, from);

						if(page->host_key == NULL)
							netSend(network, "USERHOST %s\n", data);
					}
				} else {
					for(page = p_head; page != NULL; page = page->next){
						if(page->network == network && *(page->name) != '!'){
							if(page->nicklist != NULL){
								nicksRemoveNick(page, from);
								nicksAddNick(page, data);
							}
							msgPage(page, -1, NULL, MF_NICK_CHANGE_SELF, 1, data);
						}
					}
					setMakeStr(&network->nickname, data);
					barsShow();
				}

			} else if(!strcmp(command, "KICK")){
				// <nick>!<host> KICK <channel> <kicked_nick> :<message>

				channel = data;					// channel = <channel>
				nick = skipword(data);			// nick = <kicked_nick>
				message = skipword(nick) + 1;	// message = <message>
				host = skipnick(from);			// from = <nick>
												// host = <host>

				beep_once(message);

				if(strcmp(network->nickname, nick)){
					page = msgUnkPage(network, channel, 1, NULL, MF_KICK, 3,
						message, from, nick);
					nicksRemoveNick(page, nick);
				} else {
					page = msgUnkPage(network, channel, 0, NULL, MF_KICK_SELF, 2,
						message, from);
					if(page){
						page->flags.x.joined = 0;
						page->flags.x.chanop = page->flags.x.halfchanop =
							page->flags.x.voice = page->flags.x.quiet = 0;
						barsShow();

						if(var_kick_rejoin){
							// join
							if(page->host_key)
								netSend(network, "JOIN %s %s\n",
									channel, page->host_key);
							else
								netSend(network, "JOIN %s\n", channel);

							// get channel mode
							netSend(network, "MODE %s\n", data);
						}
					}
				}

			} else msgServer(3, network, MF_UNKNOWN_SERVER_COMMAND, 3, data, command, from);

//-------------------------------------
		} else if(cmdlen == 5){
//-------------------------------------
			if(!strcmp(command, "TOPIC")){
				// <nick>!<host> TOPIC <channel> :<message>

				message = skipword(data) + 1;	// message = <message>
												// data = <channel>
				host = skipnick(from);			// from = <nick>
												// host = <host>

				page = msgUnkPage(network, data, 3, NULL, MF_TOPIC_CHANGE, 3,
					message, host, from);
				if(page) setMakeStr(&page->topic, message);
				barsShow();

			} else msgServer(3, network, MF_UNKNOWN_SERVER_COMMAND, 3, data, command, from);

//-------------------------------------
		} else if(cmdlen == 6){
//-------------------------------------
			if(!strcmp(command, "NOTICE")){
				// * NOTICE * :<message>
				// <nick>!<host> NOTICE * :<message>
				// <nick>!<host> NOTICE <channel> :<message>

				message = skipword(data) + 1;	// message = <message>
				host = skipnick(from);			// from = <nick/server>
												// data = <channel> or *
												// host = <host>

				if(ignoreFind(IG_NOTICE, data, from, host, message)) return;

				beep_once(message);

				msg_host = host;
				if(*message == 01){
					command = message + 1;
					*(command + strlen(command) - 1) = '\0';
					message = skipword(command);	// message = message
													// command = ctcp command

  // TODO: can servers send ctcp notices?

  // TODO: handle ctcp notices sent everyone

					if(!strcmp(command, "ACTION")){
						// ACTION notice
						// nick!host NOTICE to :ACTION message

						if(*data == '#'){
							msgUnkPage(network, data, 1, NULL, MF_NOTICE_ACTION, 3,
								message, host, from);
						} else {
							page = msgUnkPage(network, from, 0, NULL, MF_NOTICE_ACTION,
								3, message, host, from);

							if(page)
								if(page->host_key == NULL)
									netSend(network, "USERHOST %s\n", from);
						}
					} else if(!strcmp(command, "VERSION")){
						// VERSION reply
						// nick!host NOTICE self :VERSION message

						if((page = pageGet(network, from)) == NULL){
							msgServer(0, network, MF_CTCP_REPLY, 4,
								message, "VERSION", host, from);
						} else {
							msgPage(page, 0, NULL, MF_CTCP_REPLY, 4,
								message, "VERSION", host, from);

							if(page->host_key == NULL)
								netSend(network, "USERHOST %s\n", from);
						}
					} else if(!strcmp(command, "TIME")){
						// TIME reply
						// nick!host NOTICE self :TIME message

						if((page = pageGet(network, from)) == NULL){
							msgServer(0, network, MF_CTCP_REPLY, 4,
								message, "TIME", host, from);
						} else {
							msgPage(page, 0, NULL, MF_CTCP_REPLY, 4,
								message, "TIME", host, from);

							if(page->host_key == NULL)
								netSend(network, "USERHOST %s\n", from);
						}
					} else if(!strcmp(command, "PING")){
						// PING reply
						// nick!host NOTICE * :PING <seconds> <microseconds> <channel>

						d1 = message;				// d1 = seconds
						d2 = skipword(d1);			// d2 = microseconds
						channel = skipword(d2);		// channel = channel

						gettimeofday(&tv, NULL);
						new_time = (double)tv.tv_sec + (double)tv.tv_usec/1000000;
						old_time = (double)(cria_atoui(d1)) + (double)cria_atoui(d2)/1000000;
						snprintf(m_buffer, LEN_BUFFER_MSG, "%.3f seconds",
							new_time - old_time);

						if(strlen(channel) > 0){
							if((page = pageGet(network, channel)) == NULL){
								msgServer(0, network, MF_CTCP_REPLY, 4,
									m_buffer, "PING", host, from);
							} else {
								msgPage(page, 0, NULL, MF_CTCP_REPLY, 4,
									m_buffer, "PING", host, from);
							}
						} else if((page = pageGet(network, from)) == NULL){
							msgServer(0, network, MF_CTCP_REPLY, 4,
								m_buffer, "PING", host, from);
						} else {
							msgPage(page, 0, NULL, MF_CTCP_REPLY, 4,
								m_buffer, "PING", host, from);

							if(page->host_key == NULL)
								netSend(network, "USERHOST %s\n", from);
						}
					} else {
						// UNKNOWN CTCP reply

						if(*data == '#'){
							msgUnkPage(network, data, 1, NULL, MF_UNKNOWN_CTCP_REPLY, 4,
								message, command, host, from);
						} else {
							if((page = pageGet(network, from)) == NULL){
								msgServer(0, network, MF_UNKNOWN_CTCP_REPLY, 4,
									message, command, host, from);
							} else {
								msgPage(page, 0, NULL, MF_UNKNOWN_CTCP_REPLY, 4,
									message, command, host, from);

								if(page->host_key == NULL)
									netSend(network, "USERHOST %s\n", from);
							}
						}
					}
				} else if(*data == '#'){
					// notice from user to channel

					msgUnkPage(network, data, 1, NULL, MF_NOTICE, 3, message, host, from);
				} else if(*data == '$' || strlen(data) == 0){
					// notice to all

  // TODO: datalen will never be zero because it would consume the message if nothing was there
  //		if the TO can be empty, message and data will need to be reworked when parsing the NOTICE

					msgServer(1, network, MF_NOTICE, 3, message, "", from);
				} else if(strlen(host) > 0){
					// notice from user

					if(message_to_server){
						msgServer(1, network, MF_NOTICE, 3, message, host, from);
					} else {
						page = msgUnkPage(network, from, 0, NULL, MF_NOTICE, 3,
							message, host, from);

						if(page){
							if(var_beep_notice && p_current != page) _BEEP();

							if(page->host_key == NULL)
								netSend(network, "USERHOST %s\n", from);
						}
					}
				} else {
					// notice from server

					msgServer(3, network, MF_SERVER_NOTICE, 1, message);
				}

			} else if(!strcmp(command, "INVITE")){
				// <nick>!<host> INVITE * :<channel>
				// <nick>!<host> INVITE * <channel>

				channel = skipword(data);		// channel = <channel>
				host = skipnick(from);			// host = <host>
												// from = <nick>
				if(*channel == ':') channel++;

				if(var_invite_auto_join){
					// join
					netSend(network, "JOIN %s\n", channel);

					// get channel mode
					netSend(network, "MODE %s\n", channel);

					page = msgUnkPage(network, channel, 3, NULL, MF_INVITE_AUTO, 3,
						channel, host, from);

					page->flags.x.unjoined = 0;
				} else {
					page = msgUnkPage(network, from, 0, NULL, MF_INVITE_MANUAL, 3,
						channel, host, from);
					if(page){
						setMakeStr(&page->invite_channel, channel);

						if(var_beep_invite && p_current != page) _BEEP();

						if(page->host_key == NULL)
							netSend(network, "USERHOST %s\n", from);
					}
				}

			} else msgServer(3, network, MF_UNKNOWN_SERVER_COMMAND, 3, data, command, from);

//-------------------------------------
		} else if(cmdlen == 7){
//-------------------------------------
			if(!strcmp(command, "PRIVMSG")){
				// <nick>!<host> PRIVMSG <channel> :<message>
				// <nick>!<host> PRIVMSG * :<message>

				host = skipnick(from);			// host = <host>
												// from = <nick>
				message = skipword(data) + 1;	// message = <message>
												// data = <channel> or *

				if(*data == '#'){
					if(ignoreFind(IG_PUBLIC, data, from, host, message)) return;
				} else if(*message == 01){
					message[strlen(message)-1] = '\0';
					if(ignoreFind(IG_CTCP, data, from, host, message+1)) return;
					message[strlen(message)] = 01;
				} else {
					if(ignoreFind(IG_MESSAGE, data, from, host, message)) return;
				}

				beep_once(message);

				msg_host = host;
				if(*message == 01){
					command = message + 1;
					*(command + strlen(command) - 1) = '\0';
					message = skipword(command);	// message = message
													// command = ctcp command

  // TODO: can servers send ctcp messages?

  // TODO: handle ctcp messages sent to everyone

					if(!strcmp(command, "ACTION")){
						// ACTION message

						if(*data == '#'){
							msgUnkPage(network, data, 1, NULL, MF_ACTION, 2,
								message, from);
						} else {
							page = msgUnkPage(network, from, 0, NULL, MF_ACTION, 2,
								message, from);

							if(page){
								if(var_beep_msg && p_current != page) _BEEP();

								if(page->host_key == NULL)
									netSend(network, "USERHOST %s\n", from);
							}
						}
					} else if(!strcmp(command, "PING")){
						// PING n1 n2

						netSend(network, "NOTICE %s :\001PING %s\001\n", from, message);

						if(*data == '#'){
							msgUnkPage(network, data, 1, NULL, MF_CTCP, 3,
								"PING", host, from);
						} else {
							if((page = pageGet(network, from)) == NULL){
								msgServer(1, network, MF_CTCP, 3,
									"PING", host, from);
							} else {
								msgPage(page, 1, NULL, MF_CTCP, 3,
									"PING", host , from);

								if(page->host_key == NULL)
									netSend(network, "USERHOST %s\n", from);
							}
						}
					} else if(!strcmp(command, "VERSION")){
						// VERSION

						netSend(network, "NOTICE %s :\001VERSION %s\001\n", from,
							var_version_reply);

						if(*data == '#'){
							msgUnkPage(network, data, 1, NULL, MF_CTCP, 3,
								"VERSION", host, from);
						} else {
							if((page = pageGet(network, from)) == NULL){
								msgServer(1, network, MF_CTCP, 3,
									"VERSION", host, from);
							} else {
								msgPage(page, 1, NULL, MF_CTCP, 3,
									"VERSION", host, from);

								if(page->host_key == NULL)
									netSend(network, "USERHOST %s\n", from);
							}
						}
					} else if(!strcmp(command, "TIME")){
						// TIME

						ct = time(NULL);
						tp = localtime(&ct);
						strftime(m_buffer, LEN_BUFFER_MSG, var_time_format, tp);

						netSend(network, "NOTICE %s :\001TIME %s\001\n", from, m_buffer);

						if(*data == '#'){
							msgUnkPage(network, data, 1, NULL, MF_CTCP, 3,
								"TIME", host, from);
						} else {
							if((page = pageGet(network, from)) == NULL){
								msgServer(1, network, MF_CTCP, 3,
									"TIME", host, from);
							} else {
								msgPage(page, 1, NULL, MF_CTCP, 3,
									"TIME", host, from);

								if(page->host_key == NULL)
									netSend(network, "USERHOST %s\n", from);
							}
						}
					} else {
						// UNKNOWN CTCP message

						if(*data == '#'){
							msgUnkPage(network, data, 1, NULL, MF_UNKNOWN_CTCP, 4,
								message, command, host, from);
						} else {
							if((page = pageGet(network, from)) == NULL){
								msgServer(1, network, MF_UNKNOWN_CTCP, 4,
									message, command, host, from);
							} else {
								msgPage(page, 1, NULL, MF_UNKNOWN_CTCP, 4,
									message, command, host, from);

								if(page->host_key == NULL)
									netSend(network, "USERHOST %s\n", from);
							}
						}
					}
				} else if(*data == '#'){
					// message from user to channel

					msgUnkPage(network, data, 1, NULL, MF_MESSAGE, 2,
						message, from);
				} else if(*data == '$'){
					// message to all

					msgServer(1, network, MF_MESSAGE, 2, message, from);
				} else {
					// message from user to self

					if(message_to_server){
						msgServer(1, network, MF_MESSAGE, 2, message, from);
					} else {
						page = msgUnkPage(network, from, 0, NULL, MF_MESSAGE, 2,
							message, from);

						if(page){
							if(var_beep_msg && p_current != page) _BEEP();

							if(page->host_key == NULL)
								netSend(network, "USERHOST %s\n", from);
						}
					}
				}

			} else if(!strcmp(command, "WALLOPS")){
				// * WALLOPS :<message>
				// <nick>!<host> WALLOPS :<message>

				skipnick(from);			// from = <nick/server>
										// data = :<message>

				msgServer(1, network, MF_WALLOPS, 2, data+1, from);

			} else msgServer(3, network, MF_UNKNOWN_SERVER_COMMAND, 3, data, command, from);

//-------------------------------------
//		} else if(cmdlen == 8){
//-------------------------------------
//
//-------------------------------------
		} else msgServer(3, network, MF_UNKNOWN_SERVER_COMMAND, 3, data, command, from);

	} else if(!strcmp(from, "PING")){
		netSend(network, "PONG %s\n", data);

	} else if(!strcmp(from, "NOTICE")){
		// NOTICE AUTH :<message>
		// NOTICE * :<message>

		message = skipword(data) + 1;

		msgServer(3, network, MF_SERVER_NOTICE, 1, message);

	} else if(!strcmp(from, "ERROR")){
		// ERROR :<message>

		data++;

		msgGotoServer(network, MF_ERROR, 1, data);

	} else msgServer(3, network, MF_UNKNOWN_SERVER_COMMAND, 3, data, from, "");
}


void cmdFromUser( char *data ){
	char *command, m_buffer[LEN_BUFFER_MSG], *b;
	uint x, len;
	s_plug_user *plug;
	s_alias *a;

	msg_host = NULL;

	if(*data == '/'){
		data++;

		if(var_smart_slash){
			// not a-zA-Z?
			if(!is_alphanumeric(*data) || (*data >= '0' && *data <= '9')){ data--; goto say; }

			// is filename?
			for(b = data; *b != '\0'; b++){
				if(*b == '/'){ data--; goto say; }
				if(*b == ' ') break;
			}
		}

		command = data;
		data = skipword(data);
		stringtolower(command);

		// scan aliases
		for(a = alias_head; a != NULL; a = a->next){
			if(!strcmp(a->name, command)){
				for(b = a->body, x = 0; *b != '\0'; b++, x++){
					if(*b == '$'){
						if(*(b+1) == '-'){
							m_buffer[x] = '\0';
							len = LEN_BUFFER_MSG-strlen(m_buffer);
							strncat(m_buffer, data, len);
							x += (len < strlen(data)) ? len - 1 : strlen(data) - 1;
							b++;
						} else m_buffer[x] = '$';
					} else m_buffer[x] = *b;
				}
				m_buffer[x] = '\0';
				cmdFromUser(m_buffer);
				return;
			}
		}

		// scan plugin hooks
		for(plug = plug_user_head; plug != NULL; plug = plug->next){
			if(!strcmp(plug->command, command)){
				if(plug->cmdflags & CMD_NOTREADY){
					msgCurrentPage(-1, NULL, MF_FINISH_ME, 1, command);
				} else if(plug->cmdflags & CMD_ONLINE && n_current->socket == 0){
					msgCurrentPage(-1, n_current, MF_NETWORK_OFFLINE, 0);
				} else if(plug->cmdflags & CMD_CHANNEL && *(p_current->name) != '#'){
					msgCurrentPage(-1, NULL, MF_CHANNEL_ONLY, 1, command);
				} else if(plug->cmdflags & CMD_JOINED && !p_current->flags.x.joined){
					msgCurrentPage(-1, NULL, MF_NOT_IN_CHAN, 0);
				} else if((*plug->callback)(data, strlen(data))) break;
				return;
			}
		}

		// scan built-in commands
		for(x = 0; user_commands[x].name != NULL; x++){
			if(!strcmp(user_commands[x].name, command)){
				if(user_commands[x].cmdflags & CMD_NOTREADY){
					msgCurrentPage(-1, NULL, MF_FINISH_ME, 1, command);
				} else if(user_commands[x].cmdflags & CMD_ONLINE && n_current->socket == 0){
					msgCurrentPage(-1, n_current, MF_NETWORK_OFFLINE, 0);
				} else if(user_commands[x].cmdflags & CMD_CHANNEL && *(p_current->name) != '#'){
					msgCurrentPage(-1, NULL, MF_CHANNEL_ONLY, 1, command);
				} else if(user_commands[x].cmdflags & CMD_JOINED && !p_current->flags.x.joined){
					msgCurrentPage(-1, NULL, MF_NOT_IN_CHAN, 0);
				} else (*user_commands[x].command)(data, strlen(data));
				return;
			}
		}

		msgCurrentPage(-1, NULL, MF_UNKNOWN_USER_COMMAND, 1, command);
	} else {
		say:
		if(strlen(data) == 0) return;

		if(p_current == p_server){
			// send message to server

  // TODO: check oper status and send wallops

			msgServer(-1, n_current, MF_CANT_TALK, 0);
		} else if(*(p_current->name) == '!'){
			// send message to something like !dcc

  // TODO: display error message or ignore??

		} else if(p_current->flags.x.joined){
			// send message to channel or user

			if(n_current->socket == 0){
				msgGotoServer(n_current, MF_NETWORK_OFFLINE, 0);
			} else {
				netSend(n_current, "PRIVMSG %s :%s\n", p_current->name, data);

				msgCurrentPage(-1, NULL, MF_SENT_MESSAGE, 2, data, n_current->nickname);

				if(*(p_current->name) != '#' && p_current->host_key == NULL)
					netSend(n_current, "USERHOST %s\n", p_current->name);
			}
		} else msgCurrentPage(-1, NULL, MF_NOT_IN_CHAN, 0);
	}
}


int cmdLookup( char *name ){
	s_plug_user *plug;
	s_alias *a;
	int x;

	// aliases
	for(a = alias_head; a != NULL; a = a->next)
		if(!strcmp(a->name, name))
			return(1);

	// plugin commands
	for(plug = plug_user_head; plug != NULL; plug = plug->next)
		if(!strcmp(plug->command, name))
			return(1);

	// built-in commands
	for(x = 0; user_commands[x].name != NULL; x++)
		if(!strcmp(user_commands[x].name, name))
			return(1);

	return(0);
}
