//	cria - globals
//	Copyright (c) 2001-2010 Brand Huntsman
//


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef __curses_h__
	#include <curses.h>
#endif
#ifdef __term_h__
	#include <term.h>
#endif
#ifdef __termcap_h__
	#include <termcap.h>
#endif
#include <termios.h>
#include <time.h>
#include <sys/types.h>

#include "cgc_formats.h"


//////////////////////////////////////////////////////////////////////////

#ifdef __DEBUG__
//#define DEBUG_KEYS
#endif

#define PROG_NAME		"cria"
#define PROG_URL		"http://cria.sf.net"
#define YEAR			"2001-2011"
#define VERSION_MAJOR	0
#define VERSION_MINOR	2
#define VERSION_SUB		9
#define VERSION_CVS		17

#define LAG_PING_TIME 10

#define LEN_NICKNAME 63
#define LEN_CHANNEL 63
#define LEN_COMMAND 400
#define LEN_BUFFER_NET 512
#define LEN_BUFFER_MSG 512+100
#define LEN_COLOR_NAME 100
#define LEN_FORMAT_NAME 100
#define LEN_PATHNAME 1000

#define BAR_INPUT height
#define BAR_INFO height-1
#define BAR_PAGE height-2
#define BAR_VIEW height-3

#define _BEEP() if(tc_beep){ printf("%s", tc_beep); fflush(stdout); }
#define _ATTR_BOLD() if(tc_bold) printf("%s", tc_bold);
#define _ATTR_UNDERLINE() if(tc_underline) printf("%s", tc_underline);
#define _ATTR_REVERSE() if(tc_reverse) printf("%s", tc_reverse);
#define _ATTR_OFF() if(tc_attr_off) printf("%s", tc_attr_off);

#ifdef __LIBTERMCAP__
 #define _MOVE(x,y) printf(tc_move, 0, y, x);
 #define _SCROLL() printf(tc_scroll, 0, BAR_VIEW, width+1);
 #define _SCROLL_REGION(t,b) printf(tc_region, 0, t, b);
#else
 #define _MOVE(x,y) printf("\033[%d;%dH", y, x);
 #define _SCROLL() printf("\033[%d;%dH\n", BAR_VIEW, width+1);
 #define _SCROLL_REGION(t,b) printf("\033[%d;%dr", t, b);
#endif

#define _UPDATE_TITLEBAR() if(has_window){ \
	printf("\033]2;cria: %s in %s on %s\007", (n_current->nickname != NULL ? n_current->nickname : "(no name)"), p_current->name, n_current->name); fflush(stdout); }

#ifndef uint64
	typedef unsigned long long uint64;
#endif
//#ifndef uint
//	typedef unsigned int uint;
//#endif

//////////////////////////////////////////////////////////////////////////


//
// main.c
//

extern char version_string[15], *home;
extern int width, height, view_lines, kill_client, stdin_settings;
extern int has_window, resize_window_flag;
extern struct termios shell_settings, client_settings;
extern char *tc_beep, *tc_cursor_on, *tc_cursor_off, *tc_move;
extern char *tc_bold, *tc_underline, *tc_reverse, *tc_attr_off;
extern char *tc_region, *tc_scroll;


//
// shutdown.c
//


//
// listen.c
//

#define RECONNECT_DISABLED 0
#define RECONNECT_ENABLED 1

double listen_timer, lag_timer;


//
// key.c
//

#define KDT_EOF 0
#define KDT_CHAR 1
#define KDT_ESCSEQ 2
#define KDT_ABORT 3

struct key_data {
	int type;
	int data;
};
typedef struct key_name {
	int keycode;
	const char *name;
} key_name;

extern int input_base, input_offset, insert_mode, input_prompt_changed, input_hidden;
extern int input_aborted, input_special, input_scrolled, input_pos, input_max;
extern enum e_color_item input_prompt_color;
extern char *input_prompt, *input_temp_buffer;
extern struct key_name keynames[];


//
// net.c
//

// max timeout before reconnecting to network, in seconds (3600 == 1 hour)
#define MAX_RECONNECT_TIMEOUT 3600

extern uint proxy, proxy_ip, proxy_port;


//
// cmd.c
//

#define CMD_ONLINE 1
#define CMD_CHANNEL 2
#define CMD_JOINED 4
#define CMD_NOTREADY 8

typedef struct s_user_cmd {
	char *name;
	uint cmdflags;
	void (*command)(char *data, int datalen);
} s_user_cmd;
typedef struct s_alias {
	char *name;
	char *body;
	struct s_alias *next;
} s_alias;
typedef struct s_bind {
	int key;
	char *body;
	struct s_bind *next;
} s_bind;

extern int message_to_server, display_motd;
extern char *msg_host;
extern s_alias *alias_head;
extern s_bind *bind_head;


//
// set.c
//

enum set_type { SET_NUM, SET_CHAR, SET_STR, SET_COLOR, SET_MF };

typedef struct s_set_item {
	char *name;
	void *value;
} s_set_item;


//
// msg.c
//

#define TABSPACES 2

extern int suppress_messages;


//
// nicks.c
//

typedef struct s_nicklist {
	char *nick;
	struct s_nicklist *next;
} s_nicklist;


//
// page.c
//

#define SCROLL_PAGE_DOWN 0
#define SCROLL_PAGE_UP 1

typedef struct s_line_item {
	char *item;
	struct s_line_item *next;
} s_line_item;
typedef struct s_line {
	time_t ct;
	struct s_network *network;
	int type;
	int nr_lines;				// screen lines
	struct s_line_item *first_item;
	struct s_line *next, *prev;
} s_line;
typedef struct s_page {
	char *name;
	char *host_key;
	char *topic;
	uint64 mode;
	int limit;
	union {
		struct {
			uint joined:1;
			uint cycled:1;
			uint hidden:1;
			uint locked:1;
			uint chanop:1;
			uint halfchanop:1;
			uint voice:1;
			uint quiet:1;
			uint unjoined:1;
		} x;
		uint all;
	} flags;
	int yday;				// day in the year of last message
	struct s_network *network;
	struct s_line *buffer_head, *buffer_tail;
	int nr_lines;				// message lines
	int nr_lines_below;
	int msgtype;
	int new_messages;
	struct s_line *top_line, *bottom_line;
	s_nicklist *nicklist;
	char *invite_channel;
	struct s_page *next;
} s_page;

extern s_page *p_head, *p_current, *p_server;
extern int nr_hidden_pages;


//
// history.c
//

typedef struct s_hist_line {
	char *text;
	struct s_hist_line *next, *prev;
} s_hist_line;

extern s_hist_line *h_head, *h_tail;
extern int nr_hlines;


//
// servers.c
//

#define MODE_a (1LL << 0)
#define MODE_b (1LL << 1)
#define MODE_c (1LL << 2)
#define MODE_d (1LL << 3)
#define MODE_e (1LL << 4)
#define MODE_f (1LL << 5)
#define MODE_g (1LL << 6)
#define MODE_h (1LL << 7)
#define MODE_i (1LL << 8)
#define MODE_j (1LL << 9)
#define MODE_k (1LL << 10)
#define MODE_l (1LL << 11)
#define MODE_m (1LL << 12)
#define MODE_n (1LL << 13)
#define MODE_o (1LL << 14)
#define MODE_p (1LL << 15)
#define MODE_q (1LL << 16)
#define MODE_r (1LL << 17)
#define MODE_s (1LL << 18)
#define MODE_t (1LL << 19)
#define MODE_u (1LL << 20)
#define MODE_v (1LL << 21)
#define MODE_w (1LL << 22)
#define MODE_x (1LL << 23)
#define MODE_y (1LL << 24)
#define MODE_z (1LL << 25)
#define MODE_A (1LL << 26)
#define MODE_B (1LL << 27)
#define MODE_C (1LL << 28)
#define MODE_D (1LL << 29)
#define MODE_E (1LL << 30)
#define MODE_F (1LL << 31)
#define MODE_G (1LL << 32)
#define MODE_H (1LL << 33)
#define MODE_I (1LL << 34)
#define MODE_J (1LL << 35)
#define MODE_K (1LL << 36)
#define MODE_L (1LL << 37)
#define MODE_M (1LL << 38)
#define MODE_N (1LL << 39)
#define MODE_O (1LL << 40)
#define MODE_P (1LL << 41)
#define MODE_Q (1LL << 42)
#define MODE_R (1LL << 43)
#define MODE_S (1LL << 44)
#define MODE_T (1LL << 45)
#define MODE_U (1LL << 46)
#define MODE_V (1LL << 47)
#define MODE_W (1LL << 48)
#define MODE_X (1LL << 49)
#define MODE_Y (1LL << 50)
#define MODE_Z (1LL << 51)

typedef struct s_server {
	char *name;
	uint port;
	struct s_server *next;
} s_server;
typedef struct s_network {
	char *name;
	char *requested_nickname;
	char *nickname;
	char *password;
	uint64 requested_umode;
	uint64 umode;
	int socket;
	double current_timeout;
	int last_timeout;
	char *server_name;			// not always same as s_current->name
	struct s_server *s_current;
	struct s_server *server_head;
	int nr_pages;
	int nr_hidden_pages;
	int buffer_offset;
	char buffer[LEN_BUFFER_NET];
	union {
		struct {
			uint autoconn:1;
			uint firstconn:1;
			uint away:1;
			uint registered:1;
			uint password:1;
		} x;
		uint all;
	} flags;
	char *local_ip_addr;
	int lag, pong_received;

	char *server_version;
	char *user_modes;
	char *channel_modes;
	char *server_features;

	struct s_network *next;
} s_network;

extern s_network *n_head, *n_current;


//
// bars.c
//

extern s_page *left_page, *right_page;
extern int pb_left_arrow;


//
// config.c
//

#define RESET_COLOR "\033[0m"
#define BOLD 1
#define UNDERLINE 2
#define REVERSE 4

#define LOAD_NOISY 0
#define LOAD_QUIET 1

typedef enum e_color_name {
	BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE
} e_color_name;
typedef enum e_color_item {
	C_BG, C_FG, C_SFG, C_NICK, C_SNICK, C_HSNICK, C_HOST, C_MODE, C_TIME,
	C_INFO, C_ERROR, C_POWER, C_BRACKET, C_CHANNEL, C_NETWORK,

	C_PB_BG, C_PB_FG, C_PB_SEP, C_PB_MSG0, C_PB_MSG1, C_PB_MSG2, C_PB_KEY,
	C_PB_NETWORK, C_PB_VSCROLL, C_PB_CURRENT, C_PB_HSCROLL,

	C_IB_BG, C_IB_FG, C_IB_NICK, C_IB_SNICK, C_IB_HOST, C_IB_MODE, C_IB_TIME,
	C_IB_INFO, C_IB_ERROR, C_IB_POWER, C_IB_BRACKET, C_IB_CHANNEL, C_IB_NETWORK,
	C_IB_SCROLL,

	C_IN_FG, C_IN_NICK, C_IN_TIME, C_IN_INFO, C_IN_BRACKET, C_IN_NETWORK,
	C_IN_SCROLL, C_IN_LAG,

	MAX_COLORS
} e_color_item;
typedef struct s_color {
	e_color_name fg;
	e_color_name bg;
	int attr;
} s_color;

extern s_color colors[MAX_COLORS];
extern char fgcolors[16][8], bgcolors[8][6];

extern char *var_username, *var_realname, *var_quit_msg, *var_version_reply, *var_away_msg;
extern char *var_ip_addr, *var_time_format;
extern char var_nick_char, var_voice_char, var_op_char;
extern int var_timestamps, var_buffer_max, var_history_max, var_motd, var_color;
extern int var_beep_notice, var_beep_msg, var_kick_rejoin, var_chop_host, var_public_beeps;
extern int var_invite_auto_join, var_beep_invite, var_beep_unhide, var_pb_show_networks;
extern int var_show_bold_markers, var_show_color_markers, var_show_revvid_markers;
extern int var_show_underline_markers, var_show_unknown_markers, var_smart_slash;

extern int ver_major, ver_minor, ver_sub, ver_cvs, config_in_progress;

extern char *formats[MSG_FORMATS_LAST];


//
// plug.c
//

#ifdef __OpenBSD__
	#define RTLD_NOW RTLD_LAZY
	#define RTLD_GLOBAL RTLD_LAZY
#endif

typedef struct s_plug {
	char *name;
	#ifndef COREPLUGS
		void *handle;
	#else
		struct s_coreplug *cp;
	#endif
	struct s_plug_dep *deps;
	struct s_plug *next;
} s_plug;
typedef struct s_plug_dep {
	s_plug *plug;
	struct s_plug_dep *next;
} s_plug_dep;

typedef struct s_plug_set {
	s_plug *owner;
	char *vname;
	int type;
	void *value;
	struct s_plug_set *next;
} s_plug_set;
typedef struct s_plug_socket {
	s_plug *owner;
	int socket;
	int (*callback)(int socket);
	struct s_plug_socket *next;
} s_plug_socket;
typedef struct s_plug_user {
	s_plug *owner;
	char *command;
	uint cmdflags;
	int (*callback)(char *data, int datalen);
	struct s_plug_user *next;
} s_plug_user;
typedef struct s_plug_server {
	s_plug *owner;
	char *command;
	int (*callback)(s_network *network, char *from, char *data, int datalen);
	struct s_plug_server *next;
} s_plug_server;
typedef struct s_plug_server_int {
	s_plug *owner;
	int command;
	int (*callback)(s_network *network, char *from, char *data, int datalen);
	struct s_plug_server_int *next;
} s_plug_server_int;

typedef struct s_plug_timer {
	s_plug *owner;
	uint id;
	uint seconds;
	uint current;
	int (*callback)(uint id);
	struct s_plug_timer *next;
} s_plug_timer;

extern s_plug *plug_head;
extern s_plug_set *plug_set_head;
extern s_plug_socket *plug_socket_head;
extern s_plug_user *plug_user_head;
extern s_plug_server *plug_server_head;
extern s_plug_server_int *plug_server_int_head;
extern s_plug_timer *plug_timer_head;


//
// ignore.c
//

#define IG_NOTICE   1<<0
#define IG_MESSAGE  1<<1
#define IG_PUBLIC   1<<2
#define IG_CTCP		1<<3
#define IG_JOIN		1<<4
#define IG_PART		1<<5
#define IG_QUIT		1<<6
#define IG_NICK		1<<7
typedef struct s_ignore {
	uint type; // multiple or'd types
	char *from; // mask
	char *host; // mask
	char *body; // mask

	struct s_ignore *prev, *next;
} s_ignore;

extern s_ignore *ignore_head, *ignore_tail;


//
// user.c
//

extern int total_bans, total_quiet, display_ban_totals;


//
// cgc_plugins.c
//

#ifdef COREPLUGS
	typedef struct s_coreplug {
		const char *name;
		int (*init)(s_plug *self);
		void (*unload)();
	} s_coreplug;

	extern s_coreplug coreplugs[];
#endif


//
// cgc_formats.c
//

extern char *default_formats[];


//
// cgc_setformats.c
//

extern s_set_item setMFItems[];


//
// cgc_setvars.c
//

extern s_set_item setIntItems[];
extern s_set_item setCItems[];
extern s_set_item setStrItems[];
extern s_set_item setColorItems[];
extern char color_list[8][8];


//
// cgc_setcmds.c
//

extern s_user_cmd user_commands[];


//
// cgc_colors.c
//


//
// color.c
//


//
// misc.c
//

enum misc_sides { SIDE_LEFT, SIDE_RIGHT, SIDE_BOTH };


//
// mem-tracker.c
//

