//	cria - configuration
//	Copyright (c) 2001-2004 Brand Huntsman
//


#include <sys/utsname.h>	// OS name/version and cpu
#include <unistd.h>			// usleep() to prevent render bug

#include "common.h"
#include "functions.h"


//////////////////////////////////////////////////////////////////////////

// global
s_color colors[MAX_COLORS];
char fgcolors[16][8], bgcolors[8][6];

char *var_username, *var_realname, *var_quit_msg, *var_version_reply, *var_away_msg;
char *var_ip_addr, *var_time_format;
char var_nick_char, var_voice_char, var_op_char;
int var_timestamps, var_buffer_max, var_history_max, var_motd, var_color;
int var_beep_notice, var_beep_msg, var_kick_rejoin, var_chop_host, var_public_beeps;
int var_invite_auto_join, var_beep_invite, var_beep_unhide, var_pb_show_networks;
int var_show_bold_markers, var_show_color_markers, var_show_revvid_markers;
int var_show_underline_markers, var_show_unknown_markers, var_smart_slash;

int ver_major, ver_minor, ver_sub, ver_cvs, config_in_progress;

char *formats[MSG_FORMATS_LAST];

// local

////////////////////////////////////////////////////////////////////////// PRIVATE


void default_num_config( ){
	var_nick_char = '-';
	var_voice_char = '+';
	var_op_char = '@';

	var_timestamps = 1;
	var_buffer_max = 2000;
	var_history_max = 20;

  // TODO: motd should be ON by default, OFF for testing

	var_motd = 0;

	var_color = 1;

	var_beep_notice = 0;
	var_beep_msg = 1;
	var_beep_invite = 1;
	var_public_beeps = 1;
	var_beep_unhide = 1;

	var_kick_rejoin = 1;
	var_chop_host = 1;
	var_invite_auto_join = 0;

	var_pb_show_networks = 1;

	var_show_bold_markers = 0;
	var_show_unknown_markers = 0;
	var_smart_slash = 1;
}


void default_string_config( ){
	char *username, buffer[150];
	struct utsname uts;

	if(var_username == NULL){
		username = getenv("USER");
		if(username != NULL)
			setMakeStr(&var_username, username);
		else setMakeStr(&var_username, "cria-user");

	}
	if(var_realname == NULL) setMakeStr(&var_realname, "I have no name.");
	if(var_quit_msg == NULL){
		snprintf(buffer, 150, "%s %s -- %s", PROG_NAME, version_string, PROG_URL);
		setMakeStr(&var_quit_msg, buffer);
	}
	if(var_version_reply == NULL){
		uname(&uts);
		snprintf(buffer, 150, "%s %s [%s %s %s] -- %s", PROG_NAME, version_string,
			uts.sysname, uts.release, uts.machine, PROG_URL);
		setMakeStr(&var_version_reply, buffer);
	}
	if(var_away_msg == NULL){
		setMakeStr(&var_away_msg, "I'm gone.");
	}
}


void default_formats_config(){
	int f;

	for(f = 0; f < MSG_FORMATS_LAST; f++){
		if(formats[f] == NULL)
			setMakeStr(&formats[f], default_formats[f]);
	}
}


void default_attr_config( ){
	int c;

	// clear all attributes
	for(c=0; c<MAX_COLORS; c++) colors[c].attr = 0;

	// set a couple
	colors[C_NICK].attr = BOLD;
	colors[C_HSNICK].attr = BOLD;
	colors[C_ERROR].attr = BOLD;
	colors[C_POWER].attr = BOLD;
	colors[C_CHANNEL].attr = BOLD;
	colors[C_NETWORK].attr = BOLD;

	colors[C_PB_BG].attr = REVERSE;
	colors[C_PB_FG].attr = REVERSE;
	colors[C_PB_SEP].attr = REVERSE;
	colors[C_PB_MSG0].attr = REVERSE | BOLD | UNDERLINE;
	colors[C_PB_MSG1].attr = REVERSE | BOLD;
	colors[C_PB_MSG2].attr = REVERSE;
	colors[C_PB_KEY].attr = REVERSE;
	colors[C_PB_NETWORK].attr = REVERSE;
	colors[C_PB_VSCROLL].attr = UNDERLINE;
	colors[C_PB_CURRENT].attr = BOLD|REVERSE;
	colors[C_PB_HSCROLL].attr = BOLD;

	colors[C_IB_NICK].attr = BOLD;
	colors[C_IB_ERROR].attr = BOLD;
	colors[C_IB_CHANNEL].attr = BOLD;

	colors[C_IN_NICK].attr = BOLD;
	colors[C_IN_SCROLL].attr = BOLD;
	colors[C_IN_LAG].attr = BOLD;
}


void default_color_config( );	// function located in cgc_colors.c


void clear_strings_and_formats( ){
	int f;

	var_username = NULL;
	var_realname = NULL;
	var_quit_msg = NULL;
	var_version_reply = NULL;
	var_away_msg = NULL;
	var_ip_addr = NULL;
	var_time_format = NULL;
	setMakeStr(&var_time_format, "%H:%M %a %e %b %Y %Z");

	for(f = 0; f < MSG_FORMATS_LAST; f++) formats[f] = NULL;
}


void command_wrapper( const char *data ){
	char temp[100];

	strncpy(temp, data, 100);
	cmdFromUser(temp);
}


////////////////////////////////////////////////////////////////////////// GLOBAL


int configLoadFile( char *filename, int config, int quiet ){
	char *c_ptr, c_buffer[LEN_BUFFER_MSG], path[LEN_PATHNAME];
	FILE *fp;
	int counter;

	// hack to prevent render bug
	counter = 0;

	if(config){
		snprintf(path, LEN_PATHNAME, "%s/%s", CONFDIR, filename);
		if((fp = fopen(path, "r")) == NULL){
			snprintf(path, LEN_PATHNAME, "%s/%s/%s", home, CONFDIR, filename);
			if((fp = fopen(path, "r")) == NULL){
				if(!quiet){
					snprintf(path, LEN_PATHNAME, "%s/%s", CONFDIR, filename);
					msgGotoServer(NULL, MF_LOAD_FAILED, 1, path);
				}
				return(0);
			}
		}
		msgServer(3, NULL, MF_LOADING, 1, path);
	} else {
		if(*filename == '~'){
			snprintf(path, LEN_PATHNAME, "%s%s", home, filename+1);
		} else strncpy(path, filename, LEN_PATHNAME);

		if((fp = fopen(path, "r")) == NULL){
			msgGotoServer(NULL, MF_LOAD_FAILED, 1, filename);
			return(0);
		}
		msgServer(3, NULL, MF_LOADING, 1, filename);
	}

	suppress_messages = quiet;
	while(fgets(c_buffer, 400, fp) != NULL){
		c_ptr = strip_whitespace(c_buffer, SIDE_BOTH);
		if(*c_ptr != '#' && strlen(c_ptr) > 0){
			cmdFromUser(c_ptr);
			counter++;
		}
		if(counter > 10){
			counter = 0;
			usleep(100000);		// 0.1 sec
		}
	}
	suppress_messages = 0;

	fclose(fp);
	return(1);
}


void configEnableColor( ){
	int x;

	for(x=0; x<=7; x++) snprintf(fgcolors[x], 8, "\033[0;3%dm", x);
	for(x=8; x<=15; x++) snprintf(fgcolors[x], 8, "\033[1;3%dm", x-8);
	for(x=0; x<=7; x++) snprintf(bgcolors[x], 6, "\033[4%dm", x);

	default_color_config();
	configLoadFile("colors", 1, LOAD_QUIET);
	pageShow(p_current);
}


void configDisableColor( ){
	printf("%s", RESET_COLOR);
	default_attr_config();
	configLoadFile("no_colors", 1, LOAD_QUIET);
	pageShow(p_current);
}


void configLoad( ){
	int s;

	ver_major = VERSION_MAJOR;
	ver_minor = VERSION_MINOR;
	ver_sub = VERSION_SUB;
	ver_cvs = VERSION_CVS;

	default_num_config();
	default_attr_config();
	clear_strings_and_formats();

	s = configLoadFile("formats", 1, LOAD_QUIET);
	default_formats_config();
	if(!s) msgServer(3, NULL, MF_USING_DEFAULTS, 1, CONFDIR "/formats");

	config_in_progress = 1;

	if(configLoadFile("plugins", 1, LOAD_NOISY) == 0)
		msgServer(3, NULL, MF_USING_DEFAULTS, 1, CONFDIR "/plugins");

	if(configLoadFile("servers", 1, LOAD_NOISY) == 0)
		msgServer(3, NULL, MF_USING_DEFAULTS, 1, CONFDIR "/servers");
	if(n_head == NULL){
		command_wrapper("/network + FN");
		command_wrapper("/server + irc.freenode.net");
	}

	if(configLoadFile("config", 1, LOAD_NOISY) == 0)
		msgServer(3, NULL, MF_USING_DEFAULTS, 1, CONFDIR "/config");
	default_string_config();

	if(var_color) configEnableColor();
	else configDisableColor();

	config_in_progress = 0;
}


void configSaveInfo( ){
}
void configSaveServers( ){
}
void configSaveFormats( ){
}
