//	cria - console keyboard handling
//	Copyright (c) 2001-2004 Brand Huntsman and Nathaniel Pinckney
//


#include <unistd.h>	// usleep() to slow down large pastes

#include "common.h"
#include "functions.h"


//////////////////////////////////////////////////////////////////////////

// global
int input_base, input_offset, insert_mode, input_prompt_changed, input_hidden;
int input_aborted, input_special, input_scrolled, input_pos, input_max;
e_color_item input_prompt_color;
char *input_prompt, *input_temp_buffer;
struct key_name keynames[] = {
	{0x0, "^SPACE"},
	{0x1, "^A"}, {0x2, "^B"}, {0x3, "^C"}, {0x4, "^D"}, {0x5, "^E"},
	{0x6, "^F"}, {0x7, "^G"}, {0x8, "^H"}, {0x9, "TAB"},
	// ^I is TAB
	// {0xA, "^J"} is ENTER
	{0xB, "^K"}, {0xC, "^L"}, {0xD, "^M"},
	{0xE, "^N"}, {0xF, "^O"}, {0x10, "^P"}, {0x11, "^Q"}, {0x12, "^R"},
	{0x13, "^S"}, {0x14, "^T"}, {0x15, "^U"}, {0x16, "^V"}, {0x17, "^W"},
	{0x18, "^X"}, {0x19, "^Y"}, {0x1A, "^Z"},

	{0x1C, "^\\"}, {0x1D, "^]"}, {0x1E, "^^"}, {0x1F, "^-"},

	{0x10B, "F1"}, {0x10C, "F2"}, {0x10D, "F3"}, {0x10E, "F4"}, {0x10F, "F5"},
	{0x111, "F6"}, {0x112, "F7"}, {0x113, "F8"}, {0x114, "F9"}, {0x115, "F10"},
	{0x117, "F11"}, {0x118, "F12"},

	{0x11D, "MenuKey"},

	{0xFF, NULL} // end of list
};

// local
int input_buffer_len, input_color;
char *input_buffer;
s_hist_line *input_ln;
void (*input_callback)();

////////////////////////////////////////////////////////////////////////// PRIVATE


void getkey( struct key_data *key_data ){
	int c;
	int esc;
	int a;  // accumulator

	a = 0;
	esc = 0;
	do {
		if((c = fgetc(stdin)) == -1){
			key_data->type = KDT_EOF;		// no key
			return;
		}

		if(esc == 0){					// first code in sequence
			switch((char) c){
			case 0x1b:
				esc++;
				continue;
			default:
				key_data->type = KDT_CHAR;
				key_data->data = c;
			}
		} else if(esc == 1){				// second code in ESC sequence
			switch((char) c){
			case '[':
				esc++;
				continue;
			case 0x1b: // double escape, lets abort
				key_data->type = KDT_ABORT;
				break;
			default: // he just accidently touched it.
				 // one thing we could do is ungetc,
				 // and then send this back as a regular escape.
				 // but for right now its ok.
				key_data->type = KDT_CHAR;
				key_data->data = c;
			}
		} else if(esc == 2){				// third code in ESC[ sequence
			if((char) c >= '0' && (char) c <= '9'){
				esc++;
				a = ((char) c - '0');
				continue;
			} else { // should check for legal, but i dunno what is legal.
				key_data->type = KDT_ESCSEQ;
				key_data->data = (char) c;
			}
		} else if(esc == 3){				// fourth code in ESC[%d sequence
			if((char) c >= '0' && (char) c <= '9'){
				a *= 10;
				a += ((char) c - '0');
				continue;
			} else {  // should just be ~, more of a catch all.
				key_data->type = KDT_CHAR;
				key_data->data = a + 256;
			}
		}
		break; // while break
	} while(1);
}


////////////////////////////////////////////////////////////////////////// GLOBAL


void keyInit( ){
	// input_base is set by promptSetup()
	// input_offset is set by promptSetup()
	insert_mode = 1;
	input_prompt_changed = 1;
	// input_hidden is set by promptSetup()
	// input_aborted is set by promptSetup()
	// input_special is set by promptSetup()
	// input_scrolled is set by promptSetup()
	// input_pos is set by promptSetup()
	// input_max is set by promptSetup()
	// input_prompt_color is set by promptSetup()
	input_prompt = NULL;
	input_temp_buffer = NULL;
	// keynames is a constant structure
}


void keyDisplayInput( ){
	int left, right, max, z;

	max = width - input_base;
	left = input_scrolled;
	right = input_scrolled + width - input_base;
	if(right > input_max) right = input_max;

	// display left scroll indicator
	_MOVE(input_base - 1, BAR_INPUT);
	do_color(C_IN_SCROLL);
	if(input_scrolled) putchar('<');
	else putchar(' ');

	// display line
	do_color(C_IN_FG);
	z = input_base;
	if(input_hidden) for(; left < right; left++){
		putchar('*');
		z++;
		}
	else for(; left < right; left++){
		if(input_buffer[left] == '\002'){
			do_rev_on();
			putchar('B'); // bold marker
			do_rev_off();
			do_color(C_IN_FG);
		} else if(input_buffer[left] == '\003'){
			do_rev_on();
			putchar('C'); // mirc color marker
			do_rev_off();
			do_color(C_IN_FG);
		} else if(input_buffer[left] == '\007'){
			do_rev_on();
			putchar('G'); // beep marker
			do_rev_off();
			do_color(C_IN_FG);
		} else if(input_buffer[left] == '\017'){
			do_rev_on();
			putchar('U'); // underline marker
			do_rev_off();
			do_color(C_IN_FG);
		} else if(input_buffer[left] == '\020'){
			do_rev_on();
			putchar('V'); // reverse video marker
			do_rev_off();
			do_color(C_IN_FG);
		} else putchar(input_buffer[left]);
		z++;
	}

	if(input_scrolled + width - input_base < input_max){
		// display right scroll indicator
		do_color(C_IN_SCROLL);
		putchar('>');
	} else {
		// clear to end of line
		do_color(C_IN_FG);
		for(; z < width; z++) putchar(' ');
		do_color(C_IN_SCROLL);
		putchar(' ');
	}

	// reset cursor
	_MOVE(input_base + input_offset, BAR_INPUT);
	do_color(C_IN_FG);

	fflush(stdout);
}


void promptSetup( e_color_item fgcolor, e_color_item prompt_color, char *prompt, int prompt_len,
		char *buffer, int buffer_len, int hidden, int special, void (*callback)() ){

	input_color = fgcolor;
	input_prompt_color = prompt_color;
	input_prompt = prompt;
	// +1 is needed because termcap starts lines at 1
	if(prompt != NULL) input_base = prompt_len + 1;
	input_offset = 0;

	input_buffer = buffer;
	input_buffer_len = buffer_len;

	input_max = 0; // strlen(input_buffer)
	input_pos = 0; // current char in line
	input_scrolled = 0;

	input_hidden = hidden;
	input_special = special;
	input_callback = callback;

	input_ln = NULL;
	if(input_temp_buffer){
		track_free(input_temp_buffer, "promptSetup: temp buffer");
		input_temp_buffer = NULL;
	}

	input_prompt_changed = 1;
	input_aborted = 0;

	barsShow();
}


char *keyGetBuffer( int *length ){
	if(length != NULL) *length = input_buffer_len;
	input_buffer[input_max] = '\0';
	return(input_buffer);
}


void keyBufferReplace( char *new_buffer ){
	int len;

	len = strlen(new_buffer);
	if(len < input_buffer_len){
		// strcpy is protected by the if statement above
		strcpy(input_buffer, new_buffer);
		input_max = len;
		input_pos = len;

		if(input_max > width - input_base){
			input_scrolled = input_max - width + input_base;
			input_offset = width;
			keyDisplayInput();
		} else {
			input_scrolled = 0;
			input_offset = input_max;
			barsShow();
		}
	}
}


void keyInsert( char c ){
	if(input_pos == input_max){
		input_buffer[input_pos] = c;
		input_max++;
	} else if(insert_mode == 0){
		input_buffer[input_pos] = c;
	} else {
		input_buffer[input_max] = '\0';
		stringrevcopy(&input_buffer[input_pos+1], &input_buffer[input_pos]);
		input_buffer[input_pos] = c;
		input_max++;
	}
	input_pos++;

	if(input_base + input_offset >= width){
		input_scrolled++;
		keyDisplayInput();
	} else {
		input_offset++;
		if(input_pos < input_max && insert_mode == 1)
			keyDisplayInput();
		else fflush(stdout);
	}
}


void keyBackspace( ){
	if(input_pos > 0){
		input_pos--;
		input_max--;
		if(input_offset > 0){
			input_offset--;
		} else {
			input_scrolled--;
			keyDisplayInput();
		}

		if(input_pos < input_max){
			input_buffer[input_max+1] = '\0';
			strcpy(&input_buffer[input_pos], &input_buffer[input_pos+1]);
			keyDisplayInput();
		} else {
			_MOVE(input_base + input_offset, BAR_INPUT);
			putchar(' ');
			_MOVE(input_base + input_offset, BAR_INPUT);

			fflush(stdout);
		}
	}
}


int keyGet( ){
	int x, c, counter;
	char bindbody[LEN_COMMAND];
	struct key_data key_data;
	s_bind *b;
	#ifdef DEBUG_KEYS
	char msg[50];
	#endif

	// hack to slow down large pastes
	counter = 0;

	do{
	getkey(&key_data);
	if(key_data.type == KDT_EOF) return(0);

	#ifdef DEBUG_KEYS
	snprintf(msg, 50, "key [%d]-[0x%X]", key_data.type, key_data.data);
	out_to_file("KEY: ", msg);
//msgCurrentPage(3, NULL, MF_DEBUG, 1, msg);
	#endif

	// hack to slow down large pastes
	counter++;
	if(counter > 20){
		usleep(10000);		// 0.01 sec
		counter = 0;
	}

	// reset 5 ^C counter
	if(key_data.type == KDT_CHAR && key_data.data == 0x3){
		kill_client++;
		if(kill_client == 5) shutdownClient("Killed by keyboard");
	} else {
		kill_client = 0;
	}

	switch(key_data.type) {
	case KDT_ABORT:
		// double ESC

		input_buffer[0] = '\0'; // return empty buffer
		(*input_callback)(); // just because
		input_aborted = 1;
		return(1);

	case KDT_CHAR:
		c = key_data.data;
		if(c == 127){
			// BACKSPACE

			keyBackspace();

		} else if(c == 0xa){
			// ENTER
// should be 0xd

			input_buffer[input_max] = '\0';
			(*input_callback)();
			input_prompt_changed = 1;

			return(1);

		} else if(c <= 0x1F || c >= 0x100){ // special keys
			switch(c) {
			case 0x105:					// page up
				if(!input_special){
					if(p_current->top_line != p_current->buffer_head){
						if(p_current->bottom_line == p_current->buffer_tail)
							p_current->msgtype = 3;
						p_current->bottom_line = p_current->top_line;
						msgScrollPage(p_current, SCROLL_PAGE_UP);
						msgDisplayPage(p_current);
					}
				}
				break;
			case 0x106:					// page down
				if(!input_special){
					if(p_current->bottom_line != p_current->buffer_tail){
						p_current->top_line = p_current->bottom_line;
						msgScrollPage(p_current, SCROLL_PAGE_DOWN);
						if(p_current->bottom_line == p_current->buffer_tail){
							msgScrollPage(p_current, SCROLL_PAGE_UP);
							p_current->new_messages = 0;
						}
						msgDisplayPage(p_current);
					}
				}
				break;
			case 0x101:
			case 0x107:					// home
				if(!input_special){
					if(p_current->top_line != p_current->buffer_head){
						do {
							p_current->bottom_line = p_current->top_line;
							msgScrollPage(p_current, SCROLL_PAGE_UP);
						} while(p_current->top_line != p_current->buffer_head);
						msgDisplayPage(p_current);
					}
				}
				break;
			case 0x104:
			case 0x108:					// end
				if(!input_special){
					if(p_current->bottom_line != p_current->buffer_tail){
						p_current->bottom_line = p_current->buffer_tail;
						msgScrollPage(p_current, SCROLL_PAGE_UP);
						p_current->new_messages = 0;
						msgDisplayPage(p_current);
					}
				}
				break;
			case 0x102:					// insert
				if(insert_mode){
					insert_mode = 0;
					msgCurrentPage(-1, NULL, MF_INSERT_OFF, 0);
				} else {
					insert_mode = 1;
					msgCurrentPage(-1, NULL, MF_INSERT_ON, 0);
				}

				break;
			case 0x103:					// delete
				if(input_pos < input_max){
					input_buffer[input_max] = '\0';
					strcpy(&input_buffer[input_pos], &input_buffer[input_pos+1]);
					input_max--;
					keyDisplayInput();
				}
				break;
			default:
				for(b = bind_head; b != NULL; b = b->next){
					if(keynames[b->key].keycode == c){
						strncpy(bindbody, b->body, LEN_COMMAND);
						cmdFromUser(bindbody);
						break;
					}
				}

				if(b == NULL){
					for(x = 0; keynames[x].keycode != c && keynames[x].name != NULL; x++);
					if(keynames[x].name == NULL)
						msgCurrentPage(-1, NULL, MF_NO_SUCH_BIND, 1, "???");
					else
						msgCurrentPage(-1, NULL, MF_NO_SUCH_BIND, 1, keynames[x].name);
				}
			}

		} else if(input_max < input_buffer_len){
			if(input_hidden) putchar('*');
			else putchar(c);

			keyInsert(c);
		} else _BEEP();
		break;

	case KDT_ESCSEQ:
		switch(key_data.data) {
		case 'A':				// up arrow
			if(!input_special){
				if(input_ln == NULL){
					if(input_temp_buffer)
						track_free(input_temp_buffer, "keyGet: temp buffer up 1");
					input_temp_buffer = (char *)track_malloc(input_max+1, "keyGet history up");
					input_buffer[input_max] = '\0';
					strcpy(input_temp_buffer, input_buffer);
				}
				input_ln = historyUp(input_ln);
				if(input_ln == NULL){
					strcpy(input_buffer, input_temp_buffer);
					track_free(input_temp_buffer, "keyGet: temp buffer up 2");
					input_temp_buffer = NULL;
				} else {
					strcpy(input_buffer, input_ln->text);
				}
				input_max = strlen(input_buffer);
				input_pos = input_max;

				if(input_max > width - input_base){
					input_scrolled = input_max - width + input_base;
					input_offset = width - input_base;
					keyDisplayInput();
				} else {
					input_scrolled = 0;
					input_offset = input_max;
					barsShow();
				}
			}
			break;

		case 'B':				// down arrow
			if(!input_special){
				if(input_ln == NULL){
					if(input_temp_buffer)
						track_free(input_temp_buffer, "keyGet: temp buffer down 1");
					input_temp_buffer = (char *)track_malloc(input_max+1, "keyGet history down");
					input_buffer[input_max] = '\0';
					strcpy(input_temp_buffer, input_buffer);
				}
				input_ln = historyDown(input_ln);
				if(input_ln == NULL){
					strcpy(input_buffer, input_temp_buffer);
					track_free(input_temp_buffer, "keyGet: temp buffer down 2");
					input_temp_buffer = NULL;
				} else {
					strcpy(input_buffer, input_ln->text);
				}
				input_max = strlen(input_buffer);
				input_pos = input_max;

				if(input_pos > width - input_base){
					input_scrolled = input_max - width + input_base;
					input_offset = width - input_base;
					keyDisplayInput();
				} else {
					input_scrolled = 0;
					input_offset = input_max;
					barsShow();
				}
			}
			break;

		case 'C':				// right arrow
			if(input_pos < input_max){
				input_pos++;
				if(input_base + input_offset >= width){
					input_scrolled++;
					keyDisplayInput();
				} else {
					input_offset++;
					_MOVE(input_base + input_offset, BAR_INPUT);

					fflush(stdout);
				}
			}
			break;

		case 'D':				// left arrow
			if(input_pos > 0){
				input_pos--;
				if(input_offset > 0){
					input_offset--;
					_MOVE(input_base + input_offset, BAR_INPUT);

					fflush(stdout);
				} else {
					input_scrolled--;
					keyDisplayInput();
				}
			}
			break;
		}
		break;
	}
	} while(1);
}


int keyGetCtrlC( ){
	struct key_data key_data;

	getkey(&key_data);
	if(key_data.type == KDT_CHAR && key_data.data == 0x3)
		return(1);
	return(0);
}
