//	cria - socket handler
//	Copyright (c) 2001-2010 Brand Huntsman
//


#include <unistd.h>		// select()
#include <sys/time.h>	// struct timeval and gettimeofday()
#include <sys/ioctl.h>	// window resize

#include "common.h"
#include "functions.h"


//////////////////////////////////////////////////////////////////////////

// global
double listen_timer, lag_timer;

// local

////////////////////////////////////////////////////////////////////////// PRIVATE


void resize_window( ){
	struct winsize ws;
	int x;

	if(ioctl(STDIN_FILENO, TIOCGWINSZ, (char *)&ws) < 0){
		msgGotoServer(NULL, MF_WINSIZE_ERROR, 0);
	} else {
		// clear input bar
		msgDisableCursor();
		msgDisableScrollRegion();
		_MOVE(1, BAR_INPUT);
		for(x = 0; x < width; x++) putchar(' ');

		// display "Resizing..."
		_MOVE(1, BAR_INPUT);
		printf("Resizing...");
		fflush(stdout);

		// recalculate line lengths
		if(width != ws.ws_col){
			width = ws.ws_col;
			msgRecalcLineLengths();
		}
		height = ws.ws_row;
		view_lines = height - 3;

		// refresh window contents
		msgScrollPage(p_current, SCROLL_PAGE_UP);
		pageShow(p_current);
	}
}


////////////////////////////////////////////////////////////////////////// GLOBAL


void listenInit( ){
	struct timeval tv;

	// listen_timer is initialized by plugHookTimer()
	gettimeofday(&tv, NULL);
	lag_timer = LAG_PING_TIME + (double)tv.tv_sec + (double)tv.tv_usec/1000000;
}


int listenSockets( int reconnect ){
	struct timeval timeout, tv;
	double new_time;
	s_network *network;
	fd_set set;
	int max_fd, n, s, failed, c;
	s_plug_socket *plug_sock, *plug_sock_tmp;
	s_plug_timer *plug_timer, *plug_timer_tmp;

	failed = 0;

	// clear all sockets
	FD_ZERO(&set);

	// load stdin socket
	FD_SET(0, &set);
	max_fd = 0;

	// load network sockets
	for(network = n_head; network != NULL; network = network->next){
		if((s = network->socket)){
			FD_SET(s, &set);
			if(s > max_fd) max_fd = s;
		}
	}

	// load plugin sockets
	for(plug_sock = plug_socket_head; plug_sock != NULL; plug_sock = plug_sock->next){
		if((s = plug_sock->socket)){
			FD_SET(s, &set);
			if(s > max_fd) max_fd = s;
		}
	}

	// listen
	timeout.tv_sec = 1;
	timeout.tv_usec = 0;
	n = select(max_fd+1, &set, NULL, NULL, &timeout);
	if(n > 0){
		// stdin key press
		if(FD_ISSET(0, &set)) keyGet();

		// network socket input
		for(network = n_head; network != NULL; network = network->next){
			if((s = network->socket)){
				if(FD_ISSET(s, &set)){
					if((c = netRecv(network)) == -1){
						if(reconnect)
							netReconnect(network);
						failed = -1;
					} else if(c > 0)
						cmdFromServer(network, network->buffer);
				}
			}
		}

		// plugin socket input
		plug_sock = plug_socket_head;
		while(plug_sock != NULL){
			plug_sock_tmp = plug_sock->next;
			if((s = plug_sock->socket)){
				if(FD_ISSET(s, &set)) (*plug_sock->callback)(s);
			}
			plug_sock = plug_sock_tmp;
		}
	}

	gettimeofday(&tv, NULL);
	new_time = (double)tv.tv_sec + (double)tv.tv_usec/1000000;

	if(plug_timer_head != NULL && new_time >= listen_timer){
		// process timers
		plug_timer = plug_timer_head;
		while(plug_timer != NULL){
			plug_timer_tmp = plug_timer->next;
			plug_timer->current--;
			if(plug_timer->current == 0){
				if((*plug_timer->callback)(plug_timer->id))			// call timer callback
					plugRemoveHookTimer(plug_timer->owner, plug_timer->id);	// don't repeat
				else plug_timer->current = plug_timer->seconds;			// reset timer
			}
			plug_timer = plug_timer_tmp;
		}
		listen_timer = new_time + 1.0;
	}

	for(network = n_head; network != NULL; network = network->next)
		if(network->current_timeout && new_time > network->current_timeout)
			netReconnect(network);

	if(resize_window_flag){
		resize_window();
		resize_window_flag = 0;
	}

  // TODO: update clock

	if(lag_timer <= new_time){
		for(network = n_head; network != NULL; network = network->next){
			if(network->socket && network->flags.x.registered){
				#ifdef __DEBUG__
				char msg[200];
				snprintf(msg, 200, "%s =%f", network->server_name, new_time);
				out_to_file("PING ", msg);
				#endif

				netSend(network, "PING %s =%f\n", network->server_name, new_time);
				if(network->pong_received){
					network->pong_received = 0;
				} else {
					network->pong_received = -1;
					barsShow();
				}
			}
		}
		lag_timer = LAG_PING_TIME + new_time;
	}

	return(failed);
}
