//	cria - main
//	Copyright (c) 2001-2003 Brand Huntsman
//


#include <fcntl.h>	// non-blocking I/O for stdin
#include <signal.h>	// signal handlers
#include <unistd.h>	// STDIN_FILENO
#include <sys/stat.h>	// stat() and mkdir()

#include "common.h"
#include "functions.h"

#define TC_TGETENT_SIZE 2048
#define TC_BUFFER tc_buffer
#define TC_BUFFER_ADDR &tc_buffer


//////////////////////////////////////////////////////////////////////////

// global
char version_string[15], *home;
int width, height, view_lines, kill_client, stdin_settings;
int has_window, resize_window_flag;
struct termios shell_settings, client_settings;
char *tc_beep, *tc_cursor_on, *tc_cursor_off, *tc_move;
char *tc_bold, *tc_underline, *tc_reverse, *tc_attr_off;
char *tc_region, *tc_scroll;

// local
char c_buffer[LEN_COMMAND+1];

////////////////////////////////////////////////////////////////////////// PRIVATE


void kill_handler( int nothing ){
	shutdownClient("Client killed by user");
}


void resize_handler( int nothing ){
	resize_window_flag = 1;
}


void process_command( ){
	char * c_ptr;

	c_ptr = strip_whitespace(c_buffer, SIDE_RIGHT);
	if(strlen(c_ptr)){
		historyAdd(c_ptr);
		cmdFromUser(c_ptr);
	}
}


void prepare_config( ){
	struct stat stat_buffer;
	char path[LEN_PATHNAME], buffer[LEN_PATHNAME+LEN_PATHNAME];

	home = getenv("HOME");
	if(home == NULL){
		printf("The environment does not have an entry for the user's home directory, unable to proceed.\n");
		exit(EXIT_FAILURE);
	}

	if(stat(CONFDIR "/", &stat_buffer) == -1){
		snprintf(path, LEN_PATHNAME, "%s/%s/", home, CONFDIR);
		if(stat(path, &stat_buffer) == -1){
			if(stat(CFGDIR, &stat_buffer) == -1){
				printf("The %s directory could not be found, unable to proceed.\n", CFGDIR);
				exit(EXIT_FAILURE);
			}
			printf("Copying files [%s] from %s to %s.\n", CFGFILES, CFGDIR, path);

			if(mkdir(path, 0700) == -1){
				printf("Can't create the %s directory, unable to proceed.\n", path);
				exit(EXIT_FAILURE);
			}

			snprintf(buffer, LEN_PATHNAME+LEN_PATHNAME, "cd %s && /bin/cp %s %s",
				CFGDIR, CFGFILES, path);
			if(system(buffer) == -1); // warn_unused_result -- commands failed if -1
		}
	}
}


////////////////////////////////////////////////////////////////////////// GLOBAL


int main( int argc, char *argv[] ){
	s_network *network;
	char *termtype, *tc_clrscr;
	char *tc_buffer;

	if(VERSION_CVS > 0)
		snprintf(version_string, 15, "%d.%d.%dcvs%d",
			VERSION_MAJOR, VERSION_MINOR, VERSION_SUB, VERSION_CVS);
	else
		snprintf(version_string, 15, "%d.%d.%d",
			VERSION_MAJOR, VERSION_MINOR, VERSION_SUB);

	if(argc > 1){
		printf("%s version %s, Copyright (c) %s Brand Huntsman\n",
			PROG_NAME, version_string, YEAR);
		printf("Usage: %s\n", argv[0]);
		exit(EXIT_FAILURE);
	}

	prepare_config();

	#ifdef __DEBUG__
	 track_init();
	#endif

	// init termcap
		termtype = getenv("TERM");
		if(termtype == NULL){
			printf("Environment variable TERM was not found, please set.\n");
			exit(EXIT_FAILURE);
		}
		tc_buffer = (char *)track_malloc(TC_TGETENT_SIZE, "main.c: tc_buffer");
		if(tgetent(TC_BUFFER, termtype) <= 0){
			printf("Invalid TERM type or termcap database was not found.\n");
			exit(EXIT_FAILURE);
		}

/*
{
char *tc = tgetstr("cs", TC_BUFFER_ADDR);
printf("\n\ncode = ");
if(tc == NULL) printf("NULL");
for(; tc != NULL && *tc != '\0'; tc++){
 if(*tc < 0x20) printf("(0x%X)", *tc);
 else printf("%c", *tc);
}
printf("\n");
exit(EXIT_FAILURE);
}
*/

		// clear screen
		tc_clrscr = tgetstr("cl", TC_BUFFER_ADDR);
		if(tc_clrscr == NULL)
			shutdownBrokenTerm(0);
		printf("%s", tc_clrscr);

		// get screen size
		width = height = -1;
		width = tgetnum("co");
		height = tgetnum("li");
		if(width == -1 || height == -1)
			shutdownBrokenTerm(0);
		view_lines = height - 3;

		// get required terminal capabilities
		#ifdef __LIBTERMCAP__
		if((tc_move = tgetstr("cm", TC_BUFFER_ADDR)) == NULL)
			shutdownBrokenTerm();
		if((tc_region = tgetstr("cs", TC_BUFFER_ADDR)) == NULL)
			shutdownBrokenTerm();
		if((tc_scroll = (char *)malloc(strlen(tc_move)+3)) == NULL)
			shutdownBrokenTerm();
		sprintf(tc_scroll, "%s\n", tc_move);
		#endif

		// used by ctrl-c
		kill_client = 0;

		// set stdin to non-blocking i/o
		stdin_settings = fcntl(0, F_GETFL);
		fcntl(0, F_SETFL, stdin_settings | O_NONBLOCK);

		// test for window
		if(!strcmp(termtype, "xterm")) has_window = 1;
		else if(!strcmp(termtype, "xterm-color")) has_window = 1;
		else if(!strcmp(termtype, "rxvt")) has_window = 1;
		else if(!strcmp(termtype, "cygwin")) has_window = 1;
		else has_window = 0;

		// used by listen.c and resize signal handler
		resize_window_flag = 0;

		// save settings for stdin
		tcgetattr(0, &shell_settings);
		client_settings = shell_settings;

		// disable echo, disable ctrl-key signals and disable canonical mode
		client_settings.c_lflag &= ~(ECHO|ISIG|ICANON);

		// set buffer size to 1
		client_settings.c_cc[VTIME] = 0;
		client_settings.c_cc[VMIN] = 1;

		// apply new settings
		tcsetattr(0, TCSANOW, &client_settings);

		// get other terminal capabilities
		tc_beep = tgetstr("bl", TC_BUFFER_ADDR);
		tc_cursor_on = tgetstr("ve", TC_BUFFER_ADDR);
		tc_cursor_off = tgetstr("vi", TC_BUFFER_ADDR);
		tc_bold = tgetstr("md", TC_BUFFER_ADDR);
		tc_underline = tgetstr("us", TC_BUFFER_ADDR);
		tc_reverse = tgetstr("mr", TC_BUFFER_ADDR);
		tc_attr_off = tgetstr("me", TC_BUFFER_ADDR);


	// window resize and kill handlers
		signal(SIGWINCH, resize_handler);
		signal(SIGHUP, kill_handler);


	// initialize sections
		listenInit();
		netInit();
		keyInit();
		cmdInit();
		// set.c
		msgInit();
		// nicks.c
		pageInit();
		historyInit();
		serversInit();
		barsInit();
		plugInit();
		ignoreInit();
		// user.c
		colorInit();
		// misc.c
		// mem-tracker.c

		configLoad(); // this must be below all other init sections


	// auto-connect to servers
		for(network = n_head; network != NULL; network = network->next)
			if(network->flags.x.autoconn)
				netConnect(network, NULL);


	// display help message
		msgCurrentPage(3, NULL, MF_HELP_WANTED, 0);


	// process stdin and network connections
		do {
			if(input_prompt_changed){
				do{
					promptSetup(C_IN_FG, 0, NULL,
						strlen(n_current->name) + 3,
						c_buffer, LEN_COMMAND, 0, 0, &process_command);
					input_prompt_changed = 0;
				}while(keyGet());
			}
			listenSockets(RECONNECT_ENABLED);
		} while(1);
}
