//	cria - memory tracker
//	Copyright (c) 2001-2003 Brand Huntsman
//


#ifdef __DEBUG__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "functions.h"

typedef struct s_memory {
	char *loc;
	void *base;
	uint size;
	struct s_memory *next;
} s_memory;

s_memory *memory_head;

void track_init( ){
	memory_head = NULL;
}


void *track_malloc( size_t size, char *location ){
	s_memory *m;
	void *x;

	x = malloc(size);

	m = (s_memory *)malloc(sizeof(s_memory));
	m->loc = (char *)malloc(strlen(location)+1);
	strcpy(m->loc, location);
	m->base = x;
	m->size = size;

	m->next = memory_head;
	memory_head = m;

	return(x);
}


void track_free( void *ptr, char *location ){
	s_memory *m, *pm;
	FILE *df;

	for(m = pm = memory_head; m != NULL; pm = m, m = m->next)
		if(m->base == ptr)
			break;
	if(m == NULL){
		if((df = fopen("MT.file", "a")) != NULL){
		#ifdef __64BIT__
			fprintf(df, "Invalid free 0x%LX '%s'\n",
				(uint64)ptr, location);
		#else
			fprintf(df, "Invalid free 0x%X '%s'\n",
				(uint)ptr, location);
		#endif

			fclose(df);
		}
	} else {
		if(m == memory_head) memory_head = m->next;
		else pm->next = m->next;
		free(m->loc);
		free(m);
	}

	free(ptr);
}


void track_dump( char *data, int datalen ){
	s_memory *m;
	FILE *df;

	df = fopen("MT.file", "a");

	if(df) fprintf(df, "Starting memory dump\n");
	msgServer(3, NULL, MF_DEBUG, 1, "Dumping memory contents, this may take a while...");

	for(m = memory_head; m != NULL; m = m->next){
		if(!strcmp(m->loc, "setMakeStr")){
		#ifdef __64BIT__
			if(df) fprintf(df, "0x%LX (%d) %s '%s'\n",
				(uint64)m->base, m->size, m->loc, (char *)m->base);
		#else
			if(df) fprintf(df, "0x%X (%d) %s '%s'\n",
				(uint)m->base, m->size, m->loc, (char *)m->base);
		#endif
		} else {
		#ifdef __64BIT__
			if(df) fprintf(df, "0x%LX (%d) %s\n",
				(uint64)m->base, m->size, m->loc);
		#else
			if(df) fprintf(df, "0x%X (%d) %s\n",
				(uint)m->base, m->size, m->loc);
		#endif
		}
	}

	if(df) fprintf(df, "Ending memory dump\n");
	msgServer(3, NULL, MF_DEBUG, 1, "Memory contents have been dumped to 'MT.file'");
	if(df) fclose(df);
}

#else

void track_dump( char *data, int datalen ){
	// declare dummy function
}

#endif
