//	cria - miscellaneous
//	Copyright (c) 2001-2004 Brand Huntsman
//


#include "common.h"
#include "functions.h"


//////////////////////////////////////////////////////////////////////////

// local

////////////////////////////////////////////////////////////////////////// PRIVATE

////////////////////////////////////////////////////////////////////////// GLOBAL


char *findchar( char *string, char c ){
	while(*string != c && *string != '\0') string++;
	return(string);
}


char *skipword( char *data ){
	data = findchar(data, ' ');
	if(*data != '\0'){ *data = '\0'; data++; }
	return(data);
}
char *skipnick( char *data ){
	data = findchar(data, '!');
	if(*data != '\0'){ *data = '\0'; data++; }
	return(data);
}
char *skipuserhost( char *data ){
	data = findchar(data, '=');
	if(*data != '\0'){
		if(*(data-1) == '*')
			*(data-1) = '\0';
		else
			*data = '\0';
		data++;
		if(*data == '+' || *data == '-')
			data++;
	}
	return(data);
}
char *skipstring( char *data ){
	while(*data != '"' && *data != '\0') data++;
	if(*data == '\0') return(NULL);
	data++;
	if(*data == ' '){ *data = '\0'; data++; }
	else if(*data != '\0') return(NULL);
	return(data);
}


void stringtolower( char *string ){
	for(; *string != '\0'; string++)
		if(*string >= 'A' && *string <= 'Z')
			*string |= 32;
}
void stringtoupper( char *string ){
	for(; *string != '\0'; string++)
		if(*string >= 'a' && *string <= 'z')
			*string &= ~32;
}


char *strip_whitespace( char *string, int side ){
	char *end;

	if(side == SIDE_RIGHT || side == SIDE_BOTH){
		end = string + strlen(string) - 1;
		while(end >= string && (*end == '\n' || *end == '\t' || *end == ' ')){
			*end = '\0';
			end--;
		}
	}

	if(side == SIDE_LEFT || side == SIDE_BOTH)
		while(*string != '\0' && (*string == '\n' || *string == '\t' || *string == ' '))
			string++;

	return(string);
}


int stringcompare( char *s1, char *s2 ){
	// case-insensitive string comparison
	// return 1 if match, else 0
	char c1, c2;

	for(; *s1 != '\0' && *s2 != '\0'; s1++, s2++){
		c1 = *s1;
		c2 = *s2;
		if(c1 >= 'A' && c1 <= 'Z') c1 |= 32;
		if(c2 >= 'A' && c2 <= 'Z') c2 |= 32;
		if(c1 != c2) return(0);
	}
	if(*s1 != *s2) return(0);
	return(1);
}
int stringNcompare( char *s1, char *s2, int max ){
	// case-insensitive string comparison with limit
	// return 1 if match, else 0
	char c1, c2;
	int n;

	for(n = 0; *s1 != '\0' && *s2 != '\0' && n != max; s1++, s2++, n++){
		c1 = *s1;
		c2 = *s2;
		if(c1 >= 'A' && c1 <= 'Z') c1 |= 32;
		if(c2 >= 'A' && c2 <= 'Z') c2 |= 32;
		if(c1 != c2) return(0);
	}
	if(n == max) return(1);
	if(*s1 != *s2) return(0);
	return(1);
}
char *findstring( char *haystack, char *needle ){
	// case-insensitive version of strstr()
	// return pointer to string or NULL if not found
	uint nlen = strlen(needle), hlen = strlen(haystack);

	for(; *haystack != '\0' && hlen >= nlen; haystack++, hlen--)
		if(stringNcompare(haystack, needle, nlen))
			return(haystack);

	return(NULL);
}
/*
char *findsubstring( char *haystack, char *needle, int max ){
	// case-insensitive version of strstr() with limit on the needle
	// return pointer to substring else NULL

	for(; *haystack != '\0'; haystack++)
		if(stringNcompare(haystack, needle, max))
			return(haystack);

   return(NULL);
}
*/
int stringcomparewild( char *string, char *wild ){
	// case-insensitive string comparison with wildcards
	// return 1 if match, else 0
	// modified from http://www.codeproject.com/string/wildcmp.asp

  // TODO: support + wildcards

	for(; *string && *wild != '*'; wild++, string++){
		char c1 = *string, c2 = *wild;
		if(c1 >= 'A' && c1 <= 'Z') c1 |= 32;
		if(c2 >= 'A' && c2 <= 'Z') c2 |= 32;
		if(c1 != c2 && c2 != '?') return(0);
	}

	while(*string){
		char *cp = NULL, *mp = NULL;
		if(*wild == '*'){
			if(!*++wild)
				return(1);

			mp = wild;
			cp = string + 1;
		} else {
			char c1 = *string, c2 = *wild;
			if(c1 >= 'A' && c1 <= 'Z') c1 |= 32;
			if(c2 >= 'A' && c2 <= 'Z') c2 |= 32;
			if(c1 == c2 || c2 == '?'){
				wild++;
				string++;
			} else {
				wild = mp;
				string = cp++;
			}
		}
	}

	while(*wild == '*')
		wild++;

	return(!*wild);
}


void stringrevcopy( char *dest, char *source ){
	// copy overlapping source to dest, copies terminating null
	char *pd, *ps;

	pd = dest + strlen(source);	// write pointer
	ps = source + strlen(source);	// read pointer

	for(; ps >= source; pd--, ps--)
		*pd = *ps;
}


int is_alphanumeric( char c ){
	if(c < '0' || c > 'z') return(0);
	if(c <= '9' || c >= 'a') return(1);
	if(c >= 'A' && c <= 'Z') return(1);
	return(0);
}
int is_ipaddr( char *string ){
	for(; *string != '\0'; string++)
		if(*string < '0' || *string > '9')
			if(*string != '.')
				return(0);
	return(1);
}


int chop_userhost( char *userhost ){
	char *host, *p;
	int n;

	host = findchar(userhost, '@') + 1;
	if(is_ipaddr(host)){
		return(0);
	} else {
		p = host + strlen(host) - 1;
		n = 0;
		while(*p != '@' && n < 2){
			p--;
			if(*p == '.') n++;
		}
		if(p > host){
			*host = '*';
			strcpy(host+1, p);
			return(1);
		}
	}
	return(0);
}


void beep_once( char *data ){
	// find beep character
	while(*data != '\0' && *data != '\007') data++;
	if(*data == '\0') return;

	if(var_public_beeps) _BEEP();
}


uint cria_atoui( char *value ){
	uint n, x;
	int len, d, power;
	char *end;

	for(end = value; (*end != '\0') && (*end != '\n') && (*end != ' ') && (*end != '\t'); end++);
	len = end - value;
	if(len > 20 || len == 0) return(0);

	n = 0;
	power = len - 1;
	if(power == 0) x = 1;
	else if(power == 1) x = 10;
	else {
		x = 10;
		while(power > 1){
			x *= 10;
			power--;
		}
	}
	while(x){
		if(*value >= '0' && *value <= '9')
			d = *value - '0';
		else return(0);
		n += d * x;
		x /= 10;
		value++;
	}
	return(n);
}


void out_to_file( char *prefix, char *message ){	// FOR TESTING PURPOSES
#ifdef __DEBUG__
	FILE *df;

	if((df = fopen("DEBUG.file", "a")) != NULL){
		fprintf(df, "%s%s\n", prefix, message);
		fclose(df);
	}
#endif
}


void line_to_file( s_line *line ){	// FOR TESTING PURPOSES
#ifdef __DEBUG__
	s_line_item *i;
	FILE *df;

	if((df = fopen("DEBUG.file", "a")) != NULL){
		if(line == NULL){
			fprintf(df, "LINE: null\n");
		} else {
		#ifdef __64BIT__
			fprintf(df, "LINE 0x%LX: 0x%LX 0x%LX {%s}",
				(uint64)line, (uint64)line->prev,
				(uint64)line->next, formats[line->type]);
		#else
			fprintf(df, "LINE 0x%X: 0x%X 0x%X {%s}",
				(uint)line, (uint)line->prev,
				(uint)line->next, formats[line->type]);
		#endif
			for(i = line->first_item; i != NULL; i = i->next)
				fprintf(df, " [%s]", i->item);
			fprintf(df, "\n");
		}
		fclose(df);
	}
#endif
}
