//	cria - message handling
//	Copyright (c) 2001-2010 Brand Huntsman
//


#include <stdarg.h>	// va_start/arg/end/list
#include <time.h>	// time() and localtime()

#include "common.h"
#include "functions.h"


#define MAX_COLOR_NAME 7
#define IS_ME(type) (type == MF_MESSAGE || type == MF_ACTION || type == MF_WALLOPS || type == MF_NOTICE)


//////////////////////////////////////////////////////////////////////////

// global
int suppress_messages;

// local
int scroll_region, cursor;

////////////////////////////////////////////////////////////////////////// PRIVATE


char *skipmirccolor( char *string ){
	if(*string < '0' || *string > '9') return(string);
	string++;
	if(*string < '0' || *string > '9'){
		if(*string != ',') return(string);
	} else string++;
	if(*string != ',') return(string);
	string++;
	if(*string < '0' || *string > '9') return(string);
	string++;
	if(*string < '0' || *string > '9') return(string);
	string++;
	return(string);
}


int msgGetItemLength( char *item ){
	int length;

	length = 0;
	while(*item != '\0'){

  // TODO: will this display ansi colors in the message?

		if(*item > '\0' && *item < ' '){
			switch(*item){
			case '\n':
				length = 80 * (length/80 + 1);
				break;
			case '\t':

  // TODO: calculate spaces until next tab stop, use var_tab_size

				length += TABSPACES;
				break;
			case '\002':
				// bold markers
				if(var_show_bold_markers)
					length++;
				break;
			case '\003':
				// mirc color markers
				item++;
				if(var_show_color_markers)
					length += 1 + skipmirccolor(item) - item;
				else
					item = skipmirccolor(item) - 1;
				break;
			case '\007':
				// beeps
				length++;
				break;
			case '\017':
				// underline markers
				if(var_show_underline_markers)
					length++;
				break;
			case '\020':
				// reverse video markers
				if(var_show_revvid_markers)
					length++;
				break;
			default:
				if(var_show_unknown_markers)
					length++;
			}
		} else length++;
		item++;
	}
	return(length);
}


int msgGetLineLength( s_line *line ){
	char c, *fs, *item_list[10], item_color[MAX_COLOR_NAME+1];
	s_line_item *i;
	int nr_items, x, xx, length, fs_len;
	float nr_lines;

	// format time
	if(var_timestamps) length = 6;
	else length = 0;

	// format network
	if(line->network != NULL)
		length += strlen(line->network->name) + 1;

	// get format string for line type
	fs = formats[line->type];

	// collect line items in array
	for(nr_items = 0, i = line->first_item; i != NULL && nr_items < 10; nr_items++, i = i->next)
		item_list[nr_items] = i->item;

	// count characters in line
	fs_len = strlen(fs);
	for(x = 0; x < fs_len; x++){
		if(fs[x] == '$'){
			c = fs[x+1];
			if(c >= '0' && c <= '9'){
				x++;
				xx = c - '0';
				if(xx < nr_items) length += msgGetItemLength(item_list[xx]);
				else length += 2;
			} else length++;
		} else if(fs[x] > '\0' && fs[x] < ' '){
			switch(fs[x]){
			case '\n':
				length = 80 * (length/80 + 1);
				break;
			case '\t':

  // TODO: calculate spaces until next tab stop, use var_tab_size

				length += TABSPACES;
				break;
			default:
				if(var_show_unknown_markers)
					length++;
			}
		} else if(fs[x] == '%'){
			xx = 0;
			do{
				x++;
				c = fs[x];
				if(c != '-') item_color[xx] = c;
				xx++;
			} while(c != '-' && xx <= MAX_COLOR_NAME);
			if(c != '-'){
				length++;
				x -= MAX_COLOR_NAME + 1;
			} else {
				xx--;
				item_color[xx] = '\0';
				switch(xx){
				case 2:
					if(!strcmp(item_color, "bg"));
					else if(!strcmp(item_color, "fg"));
					else { length++; x -= 3; }
					break;
				case 3:
					if(!strcmp(item_color, "sfg"));
					else if(!strcmp(item_color, "net"));
					else { length++; x -= 4; }
					break;
				case 4:
					if(!strcmp(item_color, "chan"));
					else if(!strcmp(item_color, "info"));
					else if(!strcmp(item_color, "mode"));
					else if(!strcmp(item_color, "host"));
					else if(!strcmp(item_color, "nick"));
					else if(!strcmp(item_color, "time"));
					else { length++; x -= 5; }
					break;
				case 5:
					if(!strcmp(item_color, "snick"));
					else if(!strcmp(item_color, "power"));
					else if(!strcmp(item_color, "error"));
					else { length++; x -= 6; }
					break;
				case 6:
					if(!strcmp(item_color, "hsnick"));
					else { length++; x -= 7; }
					break;
				case 7:
					if(!strcmp(item_color, "bracket"));
					else { length++; x -= 8; }
					break;
				}
			}
		} else length++;
	}


	// TODO: indent multiple lines by X chars


	// calculate nr_lines based on screen width
	nr_lines = ((float)length) / ((float)width);
	if(nr_lines > ((int)nr_lines)) nr_lines = ((int)nr_lines) + 1;

	return((int)nr_lines);
}


void msgFormatNames( char *names ){
	char *nick;
	int x;

	nick = names;
	x = 0;
	while(*nick != '\0'){
		if(x > 0) putchar(' ');
		names = skipword(nick);
		if(*nick == '@' || *nick == '%' || *nick == '+'){
			do_color(C_POWER); putchar(*nick);
			do_color(C_NICK); printf("%s", nick+1);
		} else {
			do_color(C_NICK); printf("%s", nick);
		}
		nick = names;
		if(*names != '\0') *(names-1) = ' ';
		x++;
	}
}


void printf_string( int type, char *string, int item ){
	int nlen, slen, t;
	char *nick, c;

	if(item == 1 && IS_ME(type)){
		nick = n_current->nickname;
		nlen = strlen(nick);
		slen = strlen(string);
	} else { nick = NULL; slen = nlen = 0; }

	c = '\0';
	while(*string != '\0'){

  // TODO: will this display ansi colors in the message?

		if(*string > '\0' && *string < ' '){
			switch(*string){
			case '\n':
				putchar('\n');
				break;
			case '\t':

  // TODO: print var_tab_size number of spaces

				for(t = 0; t < TABSPACES; t++) putchar(' ');
				break;
			case '\002':
				// display B for bold markers
				if(var_show_bold_markers){
					do_rev_on();
					putchar('B');
					do_rev_off();
				}
				break;
			case '\003':
				// display C for mirc color markers
				string++;
				if(var_show_color_markers){
					uint len = skipmirccolor(string) - string;

					do_rev_on();
					putchar('C');
					for(; len > 0; len--, string++) putchar(*string);
					string--;
					do_rev_off();
				} else {
					string = skipmirccolor(string) - 1;
				}
				break;
			case '\007':
				// display G for beeps
				do_rev_on();
				putchar('G');
				do_rev_off();
				break;
			case '\017':
				// display U for underlines
				if(var_show_underline_markers){
					do_rev_on();
					putchar('U');
					do_rev_off();
				}
				break;
			case '\020':
				// display V for reverse video
				if(var_show_revvid_markers){
					do_rev_on();
					putchar('V');
					do_rev_off();
				}
				break;
			default:
				// display * for unknown markers
				if(var_show_unknown_markers){
					do_rev_on();
					putchar('*');
					do_rev_off();
				}
			}
			c = '\0';
		} else if(nick && slen >= nlen){
			if(stringNcompare(string, nick, nlen) && !is_alphanumeric(c)
			   && !is_alphanumeric(*(string+nlen))){
				e_color_item cc = get_color();

				c = *(string+nlen);

				do_color(C_HSNICK);
				printf("%.*s", nlen, string);
				do_color(cc);
				string += nlen - 1;
				slen -= nlen - 1;
			} else {
				c = *string;
				putchar(*string);
			}
			slen--;
		} else {
			c = *string;
			putchar(*string);
		}
		string++;
	}
}


void msgFormatLine( s_line *line ){
	char item_color[10], c, *fs, *item_list[10];
	int nr_items, x, xx, fs_len, t;
	s_line_item *i;

//////////////////////////////////////////////////////////
if(line == NULL){										//
FILE *df;												//
if((df = fopen("DEBUG-nullpagelines", "a")) != NULL){	//
fprintf(df, "msgFormatLine(line=NULL)\n");				//
fclose(df);												//
}														//
}														//
//////////////////////////////////////////////////////////
	if(line == NULL) return;							// malenfant

	// format time
	if(var_timestamps){
		struct tm *tp = localtime(&(line->ct));
		do_color(C_TIME);
		printf("%02d:%02d ", tp->tm_hour, tp->tm_min);
	}

	// format network
	if(line->network != NULL){
		do_color(C_NETWORK);
		printf("%s ", line->network->name);
	}

	// get format string for line type
	fs = formats[line->type];

	// collect line items in array
	for(nr_items = 0, i = line->first_item; i != NULL && nr_items < 10; nr_items++, i = i->next)
		item_list[nr_items] = i->item;


	// TODO: indent multiple lines by X chars
	// TODO: only display visible portion of line


	// display formatted line
	fs_len = strlen(fs);
	for(x = 0; x < fs_len; x++){
		if(fs[x] == '$'){
			c = fs[x+1];
			if(c >= '0' && c <= '9'){
				x++;
				xx = c - '0';
				if(xx < nr_items){
					if(line->type == MF_NAMES && xx == 1){
						msgFormatNames(item_list[xx]);
					} else printf_string(line->type, item_list[xx], xx);
				} else printf("$%c", c);
			} else putchar('$');
		} else if(fs[x] > '\0' && fs[x] < ' '){
			switch(fs[x]){
			case '\n':
				putchar('\n');
				break;
			case '\t':

  // TODO: print var_tab_size number of spaces

				for(t = 0; t < TABSPACES; t++) putchar(' ');
				break;
			default:
				if(var_show_unknown_markers){
					do_rev_on();
					putchar('*');
					do_rev_off();
				}
			}
		} else if(fs[x] == '%'){
			xx = 0;
			do{
				x++;
				c = fs[x];
				if(c != '-') item_color[xx] = c;
				xx++;
			} while(c != '-' && xx <= MAX_COLOR_NAME);
			if(c != '-'){
				putchar('%');
				x -= MAX_COLOR_NAME + 1;
			} else {
				xx--;
				item_color[xx] = '\0';
				switch(xx){
				case 2:
					if(!strcmp(item_color, "bg"))
						do_color(C_BG);
					else if(!strcmp(item_color, "fg"))
						do_color(C_FG);
					else { putchar('%'); x -= 3; }
					break;
				case 3:
					if(!strcmp(item_color, "sfg"))
						do_color(C_SFG);
					else if(!strcmp(item_color, "net"))
						do_color(C_NETWORK);
					else { putchar('%'); x -= 4; }
					break;
				case 4:
					if(!strcmp(item_color, "chan"))
						do_color(C_CHANNEL);
					else if(!strcmp(item_color, "info"))
						do_color(C_INFO);
					else if(!strcmp(item_color, "mode"))
						do_color(C_MODE);
					else if(!strcmp(item_color, "host"))
						do_color(C_HOST);
					else if(!strcmp(item_color, "nick"))
						do_color(C_NICK);
					else if(!strcmp(item_color, "time"))
						do_color(C_TIME);
					else { putchar('%'); x -= 5; }
					break;
				case 5:
					if(!strcmp(item_color, "snick"))
						do_color(C_SNICK);
					else if(!strcmp(item_color, "power"))
						do_color(C_POWER);
					else if(!strcmp(item_color, "error"))
						do_color(C_ERROR);
					else { putchar('%'); x -= 6; }
					break;
				case 6:
					if(!strcmp(item_color, "hsnick"))
						do_color(C_HSNICK);
					else { putchar('%'); x -= 7; }
					break;
				case 7:
					if(!strcmp(item_color, "bracket"))
						do_color(C_BRACKET);
					else { putchar('%'); x -= 8; }
					break;
				}
			}
		} else putchar(fs[x]);
	}
}


void msgDisplayLine( s_line *line ){
	msgDisableCursor();

//////////////////////////////////////////////////////////
if(line == NULL){										//
FILE *df;												//
if((df = fopen("DEBUG-nullpagelines", "a")) != NULL){	//
fprintf(df, "msgDisplayLine(line=NULL)\n");				//
fclose(df);												//
}														//
}														//
//////////////////////////////////////////////////////////
	if(line == NULL) return;							// malenfant

	msgEnableScrollRegion();
	_SCROLL();

	msgFormatLine(line);

	msgDisableScrollRegion();
	_MOVE(input_base + input_offset, BAR_INPUT);
	do_color(C_IN_FG);

	msgEnableCursor();

	fflush(stdout);
}


int msgScanSelf( s_network *network, char *string ){
	int nlen, slen;
	char *nick, c;

	nick = network->nickname;
	if(nick == NULL) return(0);
	nlen = strlen(nick);
	if(nlen == 0) return(0); // just in case
	slen = strlen(string);
	c = '\0';
	while(slen >= nlen){
		if(stringNcompare(string, nick, nlen)){
			if(!is_alphanumeric(c) && !is_alphanumeric(*(string+nlen)))
				return(1);
		}
		c = *string;
		string++;
		slen--;
	}
	return(0);
}


void msgPageRaw( s_page *page, int msgtype, s_network *network, int type, int count, va_list ap ){
	// line items must be passed in reverse order
	s_line *line;
	char *item;

	// add to selected page
		line = bufferAddLine(page, network, type);

	// store d1...
		while(count){
			item = va_arg(ap, char*);
			if(item == NULL){
				item = (char *)track_malloc(1, "msgPage null item");
				*item = '\0';
			}
			bufferAddLineItem(line, item);
			count--;
		}

	line->nr_lines = msgGetLineLength(line);

	if(IS_ME(line->type)) if(msgScanSelf(page->network, line->first_item->next->item)) msgtype = 0;

	if(page->bottom_line == page->buffer_tail){
		// page is not scrolled, recalculate top_line
		msgScrollPage(page, SCROLL_PAGE_UP);
	}

	if(p_current == page && page->bottom_line == page->buffer_tail){
		// current page is scrolled to bottom
		msgDisplayLine(line);
	} else if(p_current == page && page->nr_lines >= var_buffer_max
		&& page->top_line == page->buffer_head){
		// current page is scrolled to top and has been clipped
		if(msgtype == -1)
			page->new_messages--;		// don't increment message counter
		else if(page->msgtype > msgtype)
			page->msgtype = msgtype;	// change message type

		msgDisplayPage(page);
	} else {
		if(msgtype == -1)
			page->new_messages--;		// don't increment message counter
		else if(page->msgtype > msgtype)
			page->msgtype = msgtype;	// change message type

		if(page->flags.x.hidden){
			page->flags.x.hidden = 0;	// unhide page if hidden
			page->network->nr_hidden_pages--;
			nr_hidden_pages--;

			if(var_beep_unhide) _BEEP();
		}

		barsShow();
	}
}


void msgPageNewDay( s_page *page, int msgtype, s_network *network, int type, int count, ... ){
	va_list ap;

	va_start(ap, count);
	msgPageRaw(page, msgtype, network, type, count, ap);
	va_end(ap);
}


////////////////////////////////////////////////////////////////////////// GLOBAL


void msgInit( ){
	suppress_messages = 0;
	scroll_region = 0;
	msgEnableScrollRegion();
	cursor = 1;
}


void msgNewDay( s_page *page, int force ){
	struct tm *tp;
	time_t ct;
	char msg[100];

	if(*page->name == '!' && page != p_server) return;

	ct = time(NULL);
	tp = localtime(&ct);

	if(tp->tm_yday != page->yday || force){
		page->yday = tp->tm_yday;

		strftime(msg, 100, var_time_format, tp);

		msgPageNewDay(page, -1, NULL, MF_INFO, 1, msg);
	}
}


void msgRecalcLineLengths( ){
	s_page *page;
	s_line *line;

	for(page = p_head; page != NULL; page = page->next)
		for(line = page->buffer_head; line != NULL; line = line->next)
			line->nr_lines = msgGetLineLength(line);
}


void msgScrollPage( s_page *page, int direction ){
	s_line *ln, *head_ln, *tail_ln;
	int y;

	y = view_lines;

	if(direction == SCROLL_PAGE_DOWN){
		// scroll down, calculate bottom_line

		tail_ln = page->buffer_tail;
		for(ln = page->top_line; ln != tail_ln && y > 0; ln = ln->next)
			y -= ln->nr_lines;
		if(y <= 0) page->bottom_line = ln->prev;
		else page->bottom_line = ln;
	} else {
		// scroll up, calculate top_line

		head_ln = page->buffer_head;
		for(ln = page->bottom_line; ln != head_ln && y > 0; ln = ln->prev)
			y -= ln->nr_lines;
		if(y <= 0) page->top_line = ln->next;
		else page->top_line = ln;
	}

	if(page->bottom_line != page->buffer_tail){
		y = 1;
		for(ln = page->bottom_line->next; ln != page->buffer_tail; ln = ln->next)
			y++;
		page->nr_lines_below = y;
		if(page->nr_lines_below < page->new_messages)
			page->new_messages = page->nr_lines_below;
	}
}


void msgDisplayPage( s_page *page ){
	s_line *ln, *last_ln;
	int y;

	if(p_current == NULL) return;

	msgDisableCursor();

	// clear screen
	msgEnableScrollRegion();
	_SCROLL();
	for(y=1; y<view_lines; y++) putchar('\n');

	if(page->nr_lines > 0){
		ln = page->top_line;
		last_ln = page->bottom_line;

		// display each visible line
		for(; ln != last_ln; ln = ln->next){
//		for(; ln != last_ln && ln != NULL; ln = ln->next){			// malenfant
//////////////////////////////////////////////////////////
if(ln == NULL){											//
FILE *df;												//
if((df = fopen("DEBUG-nullpagelines", "a")) != NULL){	//
fprintf(df, "msgDisplayPage() ln=NULL\n");				//
fclose(df);												//
}														//
break;													//
}														//
//////////////////////////////////////////////////////////
			_SCROLL();
			msgFormatLine(ln);
			fflush(stdout);
		}
		_SCROLL();
		msgFormatLine(last_ln);
		fflush(stdout);
	}

	barsShow(); // turns cursor on
}


void msgGotoServer( s_network *network, int type, int count, ... ){
	// line items must be passed in reverse order
	s_line *line;
	va_list ap;
	char *item;

	if(suppress_messages) return;

	msgNewDay(p_server, 0);

	// add to !server page
		line = bufferAddLine(p_server, network, type);

	// store d1...
		va_start(ap, count);
		while(count){
			item = va_arg(ap, char*);
			if(item == NULL){
				item = (char *)track_malloc(1, "msgGotoServer null item");
				*item = '\0';
			}
			bufferAddLineItem(line, item);
			count--;
		}
		va_end(ap);

	line->nr_lines = msgGetLineLength(line);

	if(p_current == p_server){
		if(p_server->bottom_line == p_server->buffer_tail){
			// current page is scrolled to bottom, recalculate top_line
			msgScrollPage(p_server, SCROLL_PAGE_UP);
			msgDisplayLine(line);
		} else {
			// scroll to bottom
			p_server->bottom_line = p_server->buffer_tail;
			msgScrollPage(p_server, SCROLL_PAGE_UP);
			pageShow(p_server);
		}
	} else {
		if(p_server->bottom_line != p_server->buffer_tail){
			// scroll to bottom
			p_server->bottom_line = p_server->buffer_tail;
			msgScrollPage(p_server, SCROLL_PAGE_UP);
		}
		pageShow(p_server);
	}
}


void msgServer( int msgtype, s_network *network, int type, int count, ... ){
	// line items must be passed in reverse order
	s_line *line;
	va_list ap;
	char *item;

	if(suppress_messages) return;

	msgNewDay(p_server, 0);

	// add to !server page
		line = bufferAddLine(p_server, network, type);

	// store d1...
		va_start(ap, count);
		while(count){
			item = va_arg(ap, char*);
			if(item == NULL){
				item = (char *)track_malloc(1, "msgServer null item");
				*item = '\0';
			}
			bufferAddLineItem(line, item);
			count--;
		}
		va_end(ap);

	line->nr_lines = msgGetLineLength(line);

	if(p_server->bottom_line == p_server->buffer_tail){
		// page is not scrolled, recalculate top_line
		msgScrollPage(p_server, SCROLL_PAGE_UP);
	}

	if(p_current == p_server && p_server->bottom_line == p_server->buffer_tail){
		// current page is scrolled to bottom
		msgDisplayLine(line);
	} else if(p_current == p_server && p_server->nr_lines >= var_buffer_max
		&& p_server->top_line == p_server->buffer_head){
		// current page is scrolled to top and has been clipped
		if(msgtype == -1)
			p_server->new_messages--;	// don't increment message counter
		else if(p_server->msgtype > msgtype)
			p_server->msgtype = msgtype;	// change message type

		msgDisplayPage(p_server);
	} else {
		if(msgtype == -1)
			p_server->new_messages--;	// don't increment message counter
		else if(p_server->msgtype > msgtype)
			p_server->msgtype = msgtype;	// change message type

		barsShow();
	}
}


void msgCurrentPage( int msgtype, s_network *network, int type, int count, ... ){
	// line items must be passed in reverse order
	s_line *line;
	va_list ap;
	char *item;

	msgNewDay(p_current, 0);

	// add to p_current
		line = bufferAddLine(p_current, network, type);

	// store d1...
		va_start(ap, count);
		while(count){
			item = va_arg(ap, char*);
			if(item == NULL){
				item = (char *)track_malloc(1, "msgCurrentPage null item");
				*item = '\0';
			}
			bufferAddLineItem(line, item);
			count--;
		}
		va_end(ap);

	line->nr_lines = msgGetLineLength(line);

	if(IS_ME(line->type)) if(msgScanSelf(n_current, line->first_item->next->item)) msgtype = 0;

	if(p_current->bottom_line == p_current->buffer_tail){
		// current page is scrolled to bottom, recalculate top_line
		msgScrollPage(p_current, SCROLL_PAGE_UP);
		msgDisplayLine(line);
	} else if(p_current->nr_lines >= var_buffer_max
		&& p_current->top_line == p_current->buffer_head){
		// current page is scrolled to top and has been clipped
		if(msgtype == -1)
			p_current->new_messages--;	// don't increment message counter
		else if(p_current->msgtype > msgtype)
			p_current->msgtype = msgtype;	// change message type

		msgDisplayPage(p_current);
	} else {
		if(msgtype == -1)
			p_current->new_messages--;	// don't increment message counter
		else if(p_current->msgtype > msgtype)
			p_current->msgtype = msgtype;	// change message type

		barsShow();
	}
}


s_page *msgUnkPage( s_network *n, char *p, int msgtype, s_network *network, int type, int count, ... ){
	// line items must be passed in reverse order
	va_list ap;
	s_page *page;

	page = pageCreate(n, p, msg_host);

	if(page){
		msgNewDay(page, 0);

		va_start(ap, count);
		msgPageRaw(page, msgtype, network, type, count, ap);
		va_end(ap);
	}

	return(page);
}


void msgPage( s_page *page, int msgtype, s_network *network, int type, int count, ... ){
	// line items must be passed in reverse order
	va_list ap;

	if(page){
		msgNewDay(page, 0);

		va_start(ap, count);
		msgPageRaw(page, msgtype, network, type, count, ap);
		va_end(ap);
	}
}


void msgEnableScrollRegion( ){
	if(!scroll_region){
		scroll_region = 1;
		_SCROLL_REGION(0, BAR_VIEW);
	}
}
void msgDisableScrollRegion( ){
	if(scroll_region){
		scroll_region = 0;
		_SCROLL_REGION(BAR_PAGE, BAR_INPUT);
	}
}


void msgEnableCursor( ){
	if(!cursor && tc_cursor_on){
		cursor = 1;
		printf("%s", tc_cursor_on);
	}
}
void msgDisableCursor( ){
	if(cursor && tc_cursor_off){
		cursor = 0;
		printf("%s", tc_cursor_off);
	}
}
