//	cria - network connections
//	Copyright (c) 2001-2005 Brand Huntsman
//


#include <unistd.h>			// close() socket
#include <sys/socket.h>		// socket connecting
#include <netinet/in.h>		// struct sockaddr
#include <arpa/inet.h>		// socket connecting
#include <netdb.h>			// dns
#include <stdarg.h>			// va_start/arg/end/list
#include <sys/time.h>		// struct timeval and gettimeofday()

#include "common.h"
#include "functions.h"

#ifndef bzero
  #include <strings.h>		// bzero()
#endif


//////////////////////////////////////////////////////////////////////////

// global
uint proxy, proxy_ip, proxy_port;

// local
static int got_password;
int got_nick;
char cmd_buffer[LEN_BUFFER_MSG];

////////////////////////////////////////////////////////////////////////// PRIVATE


static void password_callback( ){
	got_password = 1;
}

void nick_callback( ){
	got_nick = 1;
}


int network_connect( s_network *network, s_server *server ){
	char p_buffer[100], m_buffer[150], nickname[LEN_NICKNAME+1];
	uint src;
	int recon_failed;

	msgServer(3, network, MF_SERVER_CONNECTING, 1, server->name);

	if(netGetHost(&src, server->name) == 0){
		snprintf(m_buffer, 150, "Could not resolve hostname [%s], skipping", server->name);
		msgServer(0, network, MF_ERROR, 1, m_buffer);
	} else {
		if((network->socket = netOpen(src, server->port)) == 0){
			snprintf(m_buffer, 150, "Could not allocate socket [%s], skipping", server->name);
			msgServer(0, network, MF_ERROR, 1, m_buffer);
		} else {
			msgServer(3, network, MF_SERVER_CONNECTED, 1, server->name);

			if(network->flags.x.password && network->password != NULL)
				if(strlen(network->password) > 0)
					netSend(network, "PASS %s\n", network->password);

			network->flags.x.registered = 0;
			network->umode = network->requested_umode;
			netSend(network, "USER %s 0 * :%s\n", var_username, var_realname);

			if(network->server_name != NULL){
				track_free(network->server_name, "network_connect: server name");
				network->server_name = NULL;
			}

			if(network->nickname == NULL)
				setMakeStr(&network->nickname, network->requested_nickname);

			while(strlen(network->nickname) == 0){
				snprintf(p_buffer, 100, "Enter nickname for %s: ", network->name);
				promptSetup(C_IN_NICK, C_IN_INFO, p_buffer, strlen(p_buffer),
					nickname, LEN_NICKNAME, 0, 1, &nick_callback);

				got_nick = 0;
				recon_failed = 0;
				while(!got_nick && recon_failed != -1)
					recon_failed = listenSockets(RECONNECT_DISABLED);

				if(recon_failed == -1) return(-1);

				if(strlen(nickname) > 0){
					if(findchar(nickname, ' ') != (nickname + strlen(nickname))){
						msgServer(0, network, MF_ERROR, 1, "Nicknames can't contain spaces");
					} else setMakeStr(&network->nickname, nickname);
				} else if(input_aborted) break;
			}

			network->current_timeout = network->last_timeout = 0;

			if(network->nickname != NULL){
				netSend(network, "NICK %s\n", network->nickname);

				network->flags.x.away = 0;
				network->pong_received = 1;

				_UPDATE_TITLEBAR()

				return(1);
			} else {
				msgServer(3, network, MF_ERROR, 1, "Connection aborted");
				network->umode = 0;				// clear usermode on infobar
				close(network->socket);
				network->socket = 0;
				return(0);
			}
		}
	}
	return(-1);
}


////////////////////////////////////////////////////////////////////////// GLOBAL


void netInit( ){
	proxy = 0;
}


int netRecv( s_network *network ){
	// returns -1 if failed
	int count, r;
	char c[2], *buffer;

	count = 0;
	buffer = &network->buffer[network->buffer_offset];
	do {
		if(count == LEN_BUFFER_NET){
			network->buffer_offset = 0;
			msgServer(0, network, MF_ERROR, 1, "Network buffer overflow, disconnecting");
			return(-1);
		}
		if((r = recv(network->socket, c, 1, 0)) < 0){
			network->buffer_offset = 0;
			msgServer(0, network, MF_ERROR, 1, "Lost connection to server");
			return(-1);
		} else if(r == 0){
			if(count == 0){
				msgServer(0, network, MF_ERROR, 1, "Received empty packet, disconnecting");
				return(-1);
			}
			network->buffer_offset += count;
			return(0);
		}
		buffer[count] = c[0];
		count++;
	} while(c[0] != '\n');

	// irc lines end with \r\n, replace with \0

	count--;
	while(buffer[count] == '\n' || buffer[count] == '\r' || buffer[count] == ' ' || buffer[count] == '\t'){
		buffer[count] = '\0';
		count--;
	}

	network->buffer_offset = 0;
	return(count);
}


int netSend( s_network *network, char *data, ... ){
	// returns -1 if failed
	int count;
	va_list ap;

	va_start(ap, data);
	vsnprintf(cmd_buffer, LEN_BUFFER_MSG, data, ap);
	va_end(ap);

	if(network->socket){
		count = send(network->socket, cmd_buffer, strlen(cmd_buffer), 0);

  // TODO: should count be compared against strlen(cmd_buffer) and looped if all data wasn't sent?

		return(count);
	}

	msgGotoServer(network, MF_NETWORK_OFFLINE, 0);

	return(-1);
}


int netGetHost( uint *src, char *hostname ){
	// 0, invalid hostname
	struct hostent *res;
	struct in_addr inp;

	if((res = gethostbyname(hostname)) != NULL){
		memcpy(src, res->h_addr, res->h_length);
		return(1);
	}

	if(!inet_aton(hostname, &inp)) return(0);
	*src = inp.s_addr;
	return(1);
}


int netOpen( uint src, int port ){
	// -1, no socket
	//  0, no connection
	struct sockaddr_in remote;
	int s;

	if((s = socket(AF_INET, SOCK_STREAM, IPPROTO_IP)) < 0)
		return(-1);

	memset(&remote, 0, sizeof(struct sockaddr_in));
	remote.sin_family = AF_INET;
	if(proxy){
		remote.sin_port = htons(proxy_port);
		remote.sin_addr.s_addr = proxy_ip;
	} else {
		remote.sin_port = htons(port);
		remote.sin_addr.s_addr = src;
	}
	#ifdef __CHAR_BZERO__
		bzero((char *)&(remote.sin_zero), 8);
	#else
		bzero(&(remote.sin_zero), 8);
	#endif

	if(connect(s, (struct sockaddr *)&remote, sizeof(struct sockaddr_in)) == 0){
		if(proxy){
			char buffer[9];

			buffer[0] = proxy; buffer[1] = 1;
			buffer[2] = port>>8; buffer[3] = port&0xff;
			buffer[7] = src>>24; buffer[6] = (src>>16)&0xff;
			buffer[5] = (src>>8)&0xff; buffer[4] = src&0xff;
			buffer[8] = '\0';
			send(s, buffer, 9, 0);

			if(recv(s, buffer, 8, 0) > 0)
				if(buffer[1] == (char)90)
					return(s);
			close(s);
			return(0);
		}
		return(s);
	}

	close(s);
	return(0);
}


int netConnect( s_network *network, s_server *server ){
	struct timeval tv;
	char msg[100];

	network->current_timeout = network->last_timeout = 0;		// reset timeout counter

	if(network->socket){
		network->umode = 0;					// clear usermode on infobar
		close(network->socket);
		network->socket = 0;
	}
	if(network->server_name){
		track_free(network->server_name, "netConnect: server name");
		network->server_name = NULL;
	}

	// if no server selected then use head
	if(server == NULL) server = network->server_head;
	network->s_current = server;

	// make sure selected server or head is valid
	if(server != NULL){
		if(network->flags.x.password && (network->password == NULL || strlen(network->password) == 0)){
			char password[33];
			int recon_failed;

			promptSetup(C_IN_FG, C_IN_INFO, "Enter password: ", strlen("Enter password: "),
				password, 32, 1, 1, &password_callback);
			got_password = 0;
			recon_failed = 0;
			while(!got_password && recon_failed != -1)
				recon_failed = listenSockets(RECONNECT_DISABLED);
			if(strlen(password) > 0)
				setMakeStr(&network->password, password);
		}

		if(network_connect(network, server) != -1){
			if(network == n_current) barsShow();
			return(1);
		}

		// connection failed, close socket and setup reconnection timer

		if(network->socket){
			network->umode = 0;				// clear usermode on infobar
			close(network->socket);
			network->socket = 0;
		}

		if(network->last_timeout < MAX_RECONNECT_TIMEOUT)
			network->last_timeout = 5;			// set timeout period to 5 seconds
		gettimeofday(&tv, NULL);
		network->current_timeout = (double)tv.tv_sec + (double)tv.tv_usec/1000000 + network->last_timeout;

		snprintf(msg, 100, "Waiting %d seconds before reconnecting", network->last_timeout);
		msgServer(0, network, MF_ERROR, 1, msg);
	} else {
		// no servers in network or invalid server, stop reconnecting
		network->current_timeout = 0;
		msgServer(0, network, MF_ERROR, 1, "No servers have been set for this network.");
	}
	return(0);
}


int netReconnect( s_network *network ){
	s_server *server;
	struct timeval tv;
	char msg[100];

	network->current_timeout = 0;		// reset timeouts

	if(network->socket){
		network->umode = 0;					// clear usermode on infobar
		close(network->socket);
		network->socket = 0;
	}
	if(network->server_name){
		track_free(network->server_name, "netReconnect: server name");
		network->server_name = NULL;
	}

	// get next server
	if(network->s_current){
		server = network->s_current->next;
		if(server == NULL) server = network->server_head;
		network->s_current = server;
	} else server = NULL;
	if(network == n_current) barsShow();

	if(server != NULL){
		if(network_connect(network, server) != -1){
			if(network == n_current) barsShow();
			return(1);
		}

		// reconnection failed, close socket and setup another reconnection timer

		if(network->socket){
			network->umode = 0;				// clear usermode on infobar
			close(network->socket);
			network->socket = 0;
		}

		if(network->last_timeout < MAX_RECONNECT_TIMEOUT)
			network->last_timeout += 5;			// add 5 seconds to timeout period
		gettimeofday(&tv, NULL);
		network->current_timeout = (double)tv.tv_sec + (double)tv.tv_usec/1000000 + network->last_timeout;

		snprintf(msg, 100, "Waiting %d seconds before reconnecting", network->last_timeout);
		msgServer(0, network, MF_ERROR, 1, msg);
	} else {
		// no servers in network, stop reconnecting
		network->current_timeout = 0;
	}
	return(0);
}
