//	cria - page handling
//	Copyright (c) 2001-2004 Brand Huntsman
//


#include "common.h"
#include "functions.h"


//////////////////////////////////////////////////////////////////////////

// global
s_page *p_head, *p_current, *p_server;
int nr_hidden_pages;

// local

////////////////////////////////////////////////////////////////////////// PRIVATE

////////////////////////////////////////////////////////////////////////// GLOBAL


void pageInit( ){
	p_head = NULL;
	p_server = pageCreate(NULL, "!server", "");
	p_current = p_server;
	nr_hidden_pages = 0;
}


s_page *pageCreate( s_network *network, char *page_name, char *host_key ){
	s_page *p;

	p = pageGet(network, page_name);
	if(p != NULL) return(p);

	p = (s_page *)track_malloc(sizeof(s_page), "pageCreate -");
	p->name = (char *)track_malloc(strlen(page_name)+1, "pageCreate name");
	strcpy(p->name, page_name);
	if(host_key != NULL){
		p->host_key = (char *)track_malloc(strlen(host_key)+1, "pageCreate host_key");
		strcpy(p->host_key, host_key);
	} else {
		if(*page_name != '#' && *page_name != '!'){
			netSend(network, "USERHOST %s\n", page_name);
		}
		p->host_key = NULL;
	}
	p->topic = (char *)track_malloc(1, "pageCreate topic");
	p->topic[0] = '\0';
	p->mode = 0;
	p->limit = 0;
	p->flags.all = 0;
	if(*page_name != '#') p->flags.x.joined = 1;
	// p->yday is set below
	p->network = network;
	p->buffer_head = p->buffer_tail = NULL;
	p->nr_lines = 0;
	p->nr_lines_below = 0;
	p->msgtype = 3;
	p->new_messages = 0;
	p->top_line = p->bottom_line = NULL;
	p->nicklist = NULL;
	p->invite_channel = NULL;

	if(p_head == NULL) p_current = p;

	p->next = p_head;
	p_head = p;

	if(network != NULL)
		network->nr_pages++;

	// reset page bar
	left_page = p;
	pb_left_arrow = 0;

	// set p->yday and display date timestamp
	msgNewDay(p, 1);

	return(p);
}


void pageFree( s_page *page ){
	s_page *p, *pp;

	if(page == NULL || page == p_server) return;

	for(p = pp = p_head; p != NULL; pp = p, p = p->next)
		if(p == page)
			break;

	if(page->network != NULL)
		page->network->nr_pages--;

	if(page == p_head){
		p_head = page->next;
		if(page == left_page){
			left_page = p_head;
			pb_left_arrow = 0;
		}
		if(page == p_current) pageShow(p_head);
		else barsShow();
	} else {
		pp->next = page->next;
		if(page == left_page){
			left_page = pp;
			pb_left_arrow = 0;
			for(p = p_head; p != pp; p = p->next){
				if(!p->flags.x.hidden){
					pb_left_arrow = 1;
					break;
				}
			}
		}
		if(page == right_page && page->next == NULL){

  // TODO: last page, shift bar right if possible

		}
		if(page == p_current) pageShow(pp);
		else barsShow();
	}

	track_free(page->name, "pageFree: name");
	if(page->host_key) track_free(page->host_key, "pageFree: host_key");
	if(page->topic) track_free(page->topic, "pageFree: topic");
	bufferClear(page);
	if(page->nicklist) nicksFree(page);
	if(page->invite_channel) track_free(page->invite_channel, "pageFree: invite channel");
	track_free(page, "pageFree:");
}


void pageShow( s_page *page ){
	if(p_current->bottom_line == p_current->buffer_tail)
		p_current->msgtype = 3;				// reset current page to 3, if not scrolled

	if(page == NULL) page = p_head;

	while(page->flags.x.hidden){
		page = page->next;					// jump to next if page is hidden
		if(page == NULL) page = p_head;
	}
	p_current = page;

	if(page->network != NULL) n_current = page->network;

	if(page->bottom_line == page->buffer_tail){		// if not scrolled
		page->msgtype = 3;							// 	reset to 3
		page->new_messages = 0;						// 	clear counter
	}

	_UPDATE_TITLEBAR()

	msgDisplayPage(page);
}


s_page *pageGet( s_network *network, char *page_name ){
	s_page *p;

	for(p = p_head; p != NULL; p = p->next)
		if(p->network == network && stringcompare(p->name, page_name))
			break;
	return(p);
}


void bufferFreeLine( s_line *line ){
	s_line_item *i, *ni;

	i = line->first_item;
	while(i != NULL){
		ni = i->next;
		track_free(i->item, "bufferFreeLine: item text");
		track_free(i, "bufferFreeLine: item");
		i = ni;
	}
	track_free(line, "bufferFreeLine:");
}
void bufferClear( s_page *page ){
	s_line *ln, *temp;

	ln = page->buffer_head;
	while(ln != NULL){
		temp = ln->next;
		bufferFreeLine(ln);
		ln = temp;
	}

	page->buffer_head = NULL;
	page->buffer_tail = NULL;
	page->nr_lines = 0;
	page->top_line = NULL;
	page->bottom_line = NULL;
}


s_line *bufferAddLine( s_page *page, s_network *network, int type ){
	s_line *ln, *ln2;

	// allocate line
	ln = (s_line *)track_malloc(sizeof(s_line), "bufferAddLine -");
	ln->ct = time(NULL);
	ln->network = network;
	ln->type = type;
	ln->next = NULL;
	ln->first_item = NULL;

	// add line to list
	if(page->buffer_head == NULL){
		page->buffer_head = ln;
		ln->prev = NULL;
		page->nr_lines++;
	} else {
		ln->prev = page->buffer_tail;
		page->buffer_tail->next = ln;

		// clip buffer, >= will prevent a bug from becoming a bigger bug
		if(page->nr_lines >= var_buffer_max){
			ln2 = page->buffer_head;
			page->buffer_head = ln2->next;
			ln2->next->prev = NULL;
			if(page->top_line == ln2){
				// page is scrolled to top
				page->top_line = ln2->next;
				msgScrollPage(page, SCROLL_PAGE_DOWN);

  // TODO: will fail if buffer size is less than one page
  //		because nr_lines hasn't been calculated for new line nor
  //		has buffer_tail been set at this point.

			}
			bufferFreeLine(ln2);
		} else page->nr_lines++;
	}

	if(page->bottom_line == page->buffer_tail)
		page->bottom_line = ln;
	else
		page->nr_lines_below++;
	page->buffer_tail = ln;

	if(page != p_current || p_current->bottom_line != p_current->buffer_tail) page->new_messages++;
	return(ln);
}


void bufferResizePage( s_page *page ){
	s_line *ln, *new_top;
	int x;

	if(page == NULL) return;

	x = 0;
	new_top = page->top_line;
	while(page->nr_lines > var_buffer_max){
		x++;
		ln = page->buffer_head;
		if(new_top == ln)
			new_top = ln->next;
		page->buffer_head = ln->next;
		ln->next->prev = NULL;
		bufferFreeLine(ln);
		page->nr_lines--;
	}
	if(x){
		if(new_top != page->top_line){
			page->top_line = new_top;
			msgScrollPage(page, SCROLL_PAGE_DOWN);
		}
		if(page == p_current)
			pageShow(p_current);
	}
}


void bufferResizeAll( ){
	s_page *page;

	for(page = p_head; page != NULL; page = page->next)
		bufferResizePage(page);
}


void bufferAddLineItem( s_line *line, char *item_text){
	s_line_item *i;

	i = (s_line_item *)track_malloc(sizeof(s_line_item), "bufferAddLineItem -");
	i->item = (char *)track_malloc(strlen(item_text)+1, "bufferAddLineItem text");
	strcpy(i->item, item_text);

	if(line->first_item == NULL)
		i->next = NULL;
	else
		i->next = line->first_item;
	line->first_item = i;
}
