//	cria - plugins
//	Copyright (c) 2002-2003 Brand Huntsman
//


#ifndef COREPLUGS
 #include <dlfcn.h>		// dynamic loading
#endif
#include <sys/time.h>	// gettimeofday()

#include "common.h"
#include "functions.h"


//////////////////////////////////////////////////////////////////////////

// global
s_plug *plug_head;
s_plug_set *plug_set_head;
s_plug_socket *plug_socket_head;
s_plug_user *plug_user_head;
s_plug_server *plug_server_head;
s_plug_server_int *plug_server_int_head;
s_plug_timer *plug_timer_head;

// local

////////////////////////////////////////////////////////////////////////// PRIVATE


#ifdef COREPLUGS
	s_coreplug *lookup_coreplug( char *name ){
		s_coreplug *cp;
		int x = 0;

		while((cp = &coreplugs[x])->name != NULL){
			if(!strcmp(name, cp->name))
				return(cp);
			x++;
		}
		return(NULL);
	}
#endif


////////////////////////////////////////////////////////////////////////// GLOBAL


void plugInit( ){
	plug_head = NULL;
	plug_set_head = NULL;
	plug_socket_head = NULL;
	plug_user_head = NULL;
	plug_server_head = NULL;
	plug_server_int_head = NULL;
	plug_timer_head = NULL;
}


int plugLoad( char *name ){
	s_plug *plug;
	char buffer[LEN_PATHNAME];
	int (*plugInit)(s_plug *self);

	// is plugin already loaded?
	for(plug = plug_head; plug != NULL; plug = plug->next){
		if(stringcompare(plug->name, name)){
			msgGotoServer(NULL, MF_PLUG_DUP, 1, name);
			return(1);
		}
	}

	// create plugin node
	plug = (s_plug *)track_malloc(sizeof(s_plug), "plugLoad -");
	plug->name = (char *)track_malloc(strlen(name)+1, "plugLoad name");
	strcpy(plug->name, name);
	plug->deps = NULL;

	plug->next = plug_head;
	plug_head = plug;

	// load plugin
	#ifndef COREPLUGS
		snprintf(buffer, LEN_PATHNAME, "%s/%s.so", PLUGDIR, name);
		if((plug->handle = dlopen(buffer, RTLD_NOW|RTLD_GLOBAL)) == NULL){
			msgGotoServer(NULL, MF_ERROR, 1, dlerror());
	#else
		if((plug->cp = lookup_coreplug(name)) == NULL){
			msgGotoServer(NULL, MF_ERROR, 1, "no such plugin");
	#endif

			plug_head = plug->next;
			track_free(plug->name, "plugLoad: name");
			// no deps yet
			track_free(plug, "plugLoad:");
			return(0);
		}

	// initialize plugin
	#ifndef COREPLUGS
		snprintf(buffer, LEN_PATHNAME, "%sInit", name);
		plugInit = dlsym(plug->handle, buffer);
	#else
		plugInit = plug->cp->init;
	#endif
	if(!(*plugInit)(plug)){
		snprintf(buffer, LEN_PATHNAME, "Initialization failed for plugin %s", name);
		msgGotoServer(NULL, MF_ERROR, 1, buffer);
		plugUnload(plug);
		return(0);
	}

	msgGotoServer(NULL, MF_PLUG_ON, 1, name);
	return(1);
}


int plugUnload( s_plug *plug ){
	// -1, null plugin
	// 0, cant unplug
	// 1, unplugged
	s_plug *p;
	s_plug_dep *pd;
	s_plug_set *pset;
	s_plug_socket *psk;
	s_plug_user *pu;
	s_plug_server *ps;
	s_plug_server_int *psi;
	s_plug_timer *pt;
	#ifndef COREPLUGS
		char buffer[LEN_PATHNAME];
	#endif
	void (*plugUnload)();

	if(plug == NULL) return(-1);

	pd = plug->deps;
	if(pd != NULL){
		do{
			msgGotoServer(NULL, MF_PLUG_INUSE, 2, pd->plug->name, plug->name);
			pd = pd->next;
		} while(pd != NULL);
		return(0);
	}

	// remove from list
	if(plug == plug_head){
		plug_head = plug->next;
	} else {
		for(p = plug_head; p->next != plug; p = p->next);
		p->next = plug->next;
	}

	// remove hooks
	for(pset = plug_set_head; pset != NULL; pset = pset->next)
		if(pset->owner == plug)
			plugRemoveHookSet(plug, pset->vname);
	for(psk = plug_socket_head; psk != NULL; psk = psk->next)
		if(psk->owner == plug)
			plugRemoveHookSocket(plug, psk->socket);
	for(pu = plug_user_head; pu != NULL; pu = pu->next)
		if(pu->owner == plug)
			plugRemoveHookUser(plug, pu->command);
	for(ps = plug_server_head; ps != NULL; ps = ps->next)
		if(ps->owner == plug)
			plugRemoveHookServer(plug, ps->command);
	for(psi = plug_server_int_head; psi != NULL; psi = psi->next)
		if(psi->owner == plug)
			plugRemoveHookServerInt(plug, psi->command);
	for(pt = plug_timer_head; pt != NULL; pt = pt->next)
		if(pt->owner == plug)
			plugRemoveHookTimer(plug, pt->id);

	// unload plugin
	#ifndef COREPLUGS
		snprintf(buffer, LEN_PATHNAME, "%sUnload", plug->name);
		plugUnload = dlsym(plug->handle, buffer);
		plugUnload();
		dlclose(plug->handle);
	#else
		plugUnload = plug->cp->unload;
		plugUnload();
	#endif

	msgGotoServer(NULL, MF_PLUG_OFF, 1, plug->name);

	// remove node
	track_free(plug->name, "plugUnload: name");
	track_free(plug, "plugUnload:");

	return(1);
}


s_plug *plugGet( char *name ){
	s_plug *plug;

	for(plug = plug_head; plug != NULL; plug = plug->next)
		if(stringcompare(plug->name, name))
			break;
	return(plug);
}


////////////////////


int plugRequire( s_plug *self, char *name ){
	// require another plugin

	return 0;
}


void plugRemoveRequire( s_plug *self, char *name ){
	// remove plugin from required list
}


int plugVersion( int v_major, int v_minor, int v_sub, int v_cvs ){
	if(ver_major > v_major) return(1);
	if(ver_minor > v_minor) return(1);
	if(ver_sub > v_sub) return(1);
	if(ver_cvs >= v_cvs) return(1);
	return(0);
}


////////////////////


int plugHookSet( s_plug *self, char *name, int type, void *value ){
	// register plugin variable
	s_plug_set *plug;

	for(plug = plug_set_head; plug != NULL; plug = plug->next)
		if(!strcmp(plug->vname, name))
			return(0);

	plug = (s_plug_set *)track_malloc(sizeof(s_plug_set), "plugHookSet -");
	plug->owner = self;
	plug->vname = (char *)track_malloc(strlen(name)+1, "plugHookSet vname");
	strcpy(plug->vname, name);
	plug->type = type;
	plug->value = value;

	plug->next = plug_set_head;
	plug_set_head = plug;

	return(1);
}


int plugHookSocket( s_plug *self, int socket, int (*callback)(int socket) ){
	// register callback for socket
	s_plug_socket *plug;

	for(plug = plug_socket_head; plug != NULL; plug = plug->next)
		if(plug->socket == socket)
			return(0);

	plug = (s_plug_socket *)track_malloc(sizeof(s_plug_socket), "plugHookSocket -");
	plug->owner = self;
	plug->socket = socket;
	plug->callback = callback;

	plug->next = plug_socket_head;
	plug_socket_head = plug;

	return(1);
}


int plugHookUser( s_plug *self, char *command, uint cmdflags, int (*callback)(char *data, int datalen) ){
	// register callback for user command
	s_plug_user *plug;

	stringtolower(command);

	for(plug = plug_user_head; plug != NULL; plug = plug->next)
		if(!strcmp(plug->command, command))
			return(0);

	plug = (s_plug_user *)track_malloc(sizeof(s_plug_user), "plugHookUser -");
	plug->owner = self;
	plug->command = (char *)track_malloc(strlen(command)+1, "plugHookUser command");
	strcpy(plug->command, command);
	plug->cmdflags = cmdflags;
	plug->callback = callback;

	plug->next = plug_user_head;
	plug_user_head = plug;

	return(1);
}


int plugHookServer( s_plug *self, char *command, int (*callback)(s_network *network, char *from, char *data, int datalen) ){
	// register callback for server command
	s_plug_server *plug;

	plug = (s_plug_server *)track_malloc(sizeof(s_plug_server), "plugHookServer -");
	plug->owner = self;
	plug->command = (char *)track_malloc(strlen(command)+1, "plugHookServer command");
	strcpy(plug->command, command);
	plug->callback = callback;

	plug->next = plug_server_head;
	plug_server_head = plug;

	return(1);
}


int plugHookServerInt( s_plug *self, int command, int (*callback)(s_network *network, char *from, char *data, int datalen) ){
	// register callback for server integer command
	s_plug_server_int *plug;

	for(plug = plug_server_int_head; plug != NULL; plug = plug->next)
		if(plug->command == command)
			return(0);

	plug = (s_plug_server_int *)track_malloc(sizeof(s_plug_server_int), "plugHookServerInt -");
	plug->owner = self;
	plug->command = command;
	plug->callback = callback;

	plug->next = plug_server_int_head;
	plug_server_int_head = plug;

	return(1);
}


int plugHookTimer( s_plug *self, uint id, uint seconds, int (*callback)(uint id) ){
	// register callback for timer
	s_plug_timer *plug;
	struct timeval tv;

	for(plug = plug_timer_head; plug != NULL; plug = plug->next)
		if(plug->owner == self && plug->id == id)
			return(0);

	plug = (s_plug_timer *)track_malloc(sizeof(s_plug_timer), "plugHookTimer -");
	plug->owner = self;
	plug->id = id;
	plug->seconds = seconds;
	plug->current = seconds;
	plug->callback = callback;

	plug->next = plug_timer_head;
	if(plug_timer_head == NULL){
		gettimeofday(&tv, NULL);
		listen_timer = (double)tv.tv_sec + (double)tv.tv_usec/1000000 + 1.0;
	}
	plug_timer_head = plug;

	return(1);
}


////////////////////


int plugRemoveHookSet( s_plug *self, char *name ){
	s_plug_set *plug, *old;

	for(old = NULL, plug = plug_set_head; plug != NULL; old = plug, plug = plug->next){
		if(!strcmp(plug->vname, name) && plug->owner == self){
			if(old != NULL) old->next = plug->next;
			else plug_set_head = plug->next;
			track_free(plug->vname, "plugRemoveHookSet: vname");
			track_free(plug, "plugRemoveHookSet:");
			return(1);
		}
	}
	return(0);
}


int plugRemoveHookSocket( s_plug *self, int socket ){
	s_plug_socket *plug, *old;

	for(old = NULL, plug = plug_socket_head; plug != NULL; old = plug, plug = plug->next){
		if(plug->socket == socket && plug->owner == self){
			if(old != NULL) old->next = plug->next;
			else plug_socket_head = plug->next;
			track_free(plug, "plugRemoveHookSocket:");
			return(1);
		}
	}
	return(0);
}


int plugRemoveHookUser( s_plug *self, char *command ){
	s_plug_user *plug, *old;

	for(old = NULL, plug = plug_user_head; plug != NULL; old = plug, plug = plug->next){
		if(!strcmp(plug->command, command) && plug->owner == self){
			if(old != NULL) old->next = plug->next;
			else plug_user_head = plug->next;
			track_free(plug->command, "plugRemoveHookUser: command");
			track_free(plug, "plugRemoveHookUser:");
			return(1);
		}
	}
	return(0);
}


int plugRemoveHookServer( s_plug *self, char *command ){
	s_plug_server *plug, *old;

	for(old = NULL, plug = plug_server_head; plug != NULL; old = plug, plug = plug->next){
		if(!strcmp(plug->command, command) && plug->owner == self){
			if(old != NULL) old->next = plug->next;
			else plug_server_head = plug->next;
			track_free(plug->command, "plugRemoveHookServer: command");
			track_free(plug, "plugRemoveHookServer:");
			return(1);
		}
	}
	return(0);
}


int plugRemoveHookServerInt( s_plug *self, int command ){
	s_plug_server_int *plug, *old;

	for(old = NULL, plug = plug_server_int_head; plug != NULL; old = plug, plug = plug->next){
		if(plug->command == command && plug->owner == self){
			if(old != NULL) old->next = plug->next;
			else plug_server_int_head = plug->next;
			track_free(plug, "plugRemoveHookServerInt:");
			return(1);
		}
	}
	return(0);
}


int plugRemoveHookTimer( s_plug *self, uint id ){
	s_plug_timer *plug, *old;

	for(old = NULL, plug = plug_timer_head; plug != NULL; old = plug, plug = plug->next){
		if(plug->id == id && plug->owner == self){
			if(old != NULL) old->next = plug->next;
			else plug_timer_head = plug->next;
			track_free(plug, "plugRemoveHookTimer:");
			return(1);
		}
	}
	return(0);
}
