//	cria - network and server handling
//	Copyright (c) 2001-2004 Brand Huntsman
//


#include "common.h"
#include "functions.h"


//////////////////////////////////////////////////////////////////////////

// global
s_network *n_head, *n_current;

// local

////////////////////////////////////////////////////////////////////////// PRIVATE

////////////////////////////////////////////////////////////////////////// GLOBAL


void serversInit( ){
	n_head = NULL;
	n_current = NULL;
}


s_network *networkCreate( char *network, char *nickname, uint64 umode, uint auto_connect ){
	s_network *n;

	n = (s_network *)track_malloc(sizeof(s_network), "networkCreate -");
	n->name = (char *)track_malloc(strlen(network)+1, "networkCreate name");
	strcpy(n->name, network);
	n->requested_nickname = (char *)track_malloc(strlen(nickname)+1, "networkCreate requested_nickname");
	strcpy(n->requested_nickname, nickname);
	n->nickname = NULL;
	n->password = NULL;
	n->requested_umode = umode;
	n->umode = umode;
	n->socket = 0;
	n->current_timeout = 0;
	n->last_timeout = 0;
	n->server_name = NULL;
	n->s_current = NULL;
	n->server_head = NULL;
	n->nr_pages = 0;
	n->nr_hidden_pages = 0;
	n->buffer_offset = 0;
	// n->buffer[]
	n->flags.all = 0;
	n->flags.x.firstconn = 1;
	if(auto_connect) n->flags.x.autoconn = 1;
	n->local_ip_addr = NULL;
	n->lag = 0;
	// n->pong_received = 1;	set by network_connect()

	n->server_version = NULL;
	n->user_modes = NULL;
	n->channel_modes = NULL;
	n->server_features = NULL;

	n->next = n_head;
	n_head = n;

	return(n);
}


void networkFree( s_network *network ){
	s_network *n, *np;

	for(n = np = n_head; n != NULL; np = n, n = n->next)
		if(n == network)
			break;

	if(network == n_head) n_head = network->next;
	np->next = network->next;

	track_free(network->name, "networkFree: name");
	track_free(network->requested_nickname, "networkFree: requested nickname");
	if(network->nickname) track_free(network->nickname, "networkFree: nickname");
	if(network->server_name) track_free(network->server_name, "networkFree: server name");
	if(network->local_ip_addr) track_free(network->local_ip_addr, "networkFree: local ip addr");
	serverFreeAll(network);
	track_free(network, "networkFree:");
}


s_network *networkGet( char *network_name ){
	s_network *n;

	for(n = n_head; n != NULL; n = n->next)
		if(stringcompare(n->name, network_name))
			return(n);
	return(NULL);
}


s_server *serverCreate( s_network *network, char *server, uint port ){
	s_server *s;

	s = (s_server *)track_malloc(sizeof(s_server), "serverCreate -");
	s->name = (char *)track_malloc(strlen(server)+1, "serverCreate name");
	strcpy(s->name, server);
	s->port = port;

	s->next = network->server_head;
	network->server_head = s;

	if(network->s_current == NULL) network->s_current = s;

	return(s);
}


void serverFree( s_network *network, char *server ){

  // TODO:

}


void serverFreeAll( s_network *network ){
	s_server *s, *s2;

	s = network->server_head;
	while(s != NULL){
		s2 = s->next;
		track_free(s->name, "serverFreeAll: name");
		track_free(s, "serverFreeAll:");
		s = s2;
	}
	network->server_head = NULL;
}


uint64 umodeToInt( uint64 umode, char *ms ){
	uint x, mode_switch, ms_len, bit;

	mode_switch = 1;
	ms_len = strlen(ms);
	for(x = 0; x < ms_len; x++){
		switch(ms[x]){
		case '+': mode_switch = 1; break;
		case '-': mode_switch = 0; break;
		default:
			if(ms[x] >= 'a' && ms[x] <= 'z') bit = ms[x] - 'a';
			else if(ms[x] >= 'A' && ms[x] <= 'Z') bit = ms[x] - 'A' + 26;
			else break;

			if(mode_switch) umode |= (1LL << bit);
			else umode &= ~(1LL << bit);
		}
	}

	return(umode);
}


char *modeToString( char *ms, uint64 mode ){
	uint bit;
	char c[2];

	ms[0] = '+';
	ms[1] = '\0';
	c[0] = 'a';
	c[1] = '\0';

	c[0] = 'a';
	for(bit = 0; bit < 26; bit++){
		if(mode & (1LL << bit)) strcat(ms, c);		// ms[55]
		c[0]++;
	}

	c[0] = 'A';
	for(bit = 26; bit < 52; bit++){
		if(mode & (1LL << bit)) strcat(ms, c);		// ms[55]
		c[0]++;
	}

	// old cmode order = spimtnrcCPFlk
	// old umode order = Cxeiw

	if(ms[1] == '\0') ms[0] = '\0';
	return(ms);
}


uint64 cmodeToInt( s_page *page, char *ms ){
	uint x, n, mode_switch, ms_len, bit;
	uint64 cmode;
	char *arg;

	if(page == NULL) return(0);

	arg = skipword(ms); // first mode argument

	cmode = page->mode;
	mode_switch = 1;
	ms_len = strlen(ms);
	for(x = 0; x < ms_len; x++){
		n = findchar(arg, ' ') - arg;
		switch(ms[x]){
		case '+': mode_switch = 1; break;
		case '-': mode_switch = 0; break;
		case 'l':
			if(mode_switch) cmode |= MODE_l; else cmode &= ~MODE_l;
			page->limit = cria_atoui(arg);
			arg = skipword(arg);
			break;
		case 'k':
			if(mode_switch) cmode |= MODE_k; else cmode &= ~MODE_k;
			setMakeStr(&page->host_key, arg);
			arg = skipword(arg);
			break;
		case 'o':
			if(strlen(page->network->nickname) == n && stringNcompare(arg, page->network->nickname, n)){
				if(mode_switch) page->flags.x.chanop = 1;
				else page->flags.x.chanop = 0;
			}
			arg = skipword(arg);
			break;
		case 'h':
			if(strlen(page->network->nickname) == n && stringNcompare(arg, page->network->nickname, n)){
				if(mode_switch) page->flags.x.halfchanop = 1;
				else page->flags.x.halfchanop = 0;
			}
			arg = skipword(arg);
			break;
		case 'v':
			if(strlen(page->network->nickname) == n && stringNcompare(arg, page->network->nickname, n)){
				if(mode_switch) page->flags.x.voice = 1;
				else page->flags.x.voice = 0;
			}
			arg = skipword(arg);
			break;
		case 'q':

  // TODO: this isn't stable as the user@host might not match

			n = findchar(arg, '!') - arg;
			if(strlen(page->network->nickname) == n && stringNcompare(arg, page->network->nickname, n)){
				if(mode_switch) page->flags.x.quiet = 1;
				else page->flags.x.quiet = 0;
			}
			arg = skipword(arg);
			break;
		case 'b':
		case 'e':
			arg = skipword(arg);
			break;
		default:
			if(ms[x] >= 'a' && ms[x] <= 'z') bit = ms[x] - 'a';
			else if(ms[x] >= 'A' && ms[x] <= 'Z') bit = ms[x] - 'A' + 26;
			else break;

			if(mode_switch) cmode |= (1LL << bit);
			else cmode &= ~(1LL << bit);
		}
	}
	page->mode = cmode;
	return(cmode);
}
