//	cria - configuration variables
//	Copyright (c) 2001-2003 Brand Huntsman and Nathaniel Pinckney
//


#include <stdarg.h>	// va_start/end/list

#include "common.h"
#include "functions.h"


//////////////////////////////////////////////////////////////////////////

// global

// local

////////////////////////////////////////////////////////////////////////// PRIVATE


void *findValue( int type, char *name ){
	s_set_item *item;
	s_plug_set *p;

	switch(type){
		case SET_NUM:	item = setIntItems; break;
		case SET_CHAR:	item = setCItems; break;
		case SET_STR:	item = setStrItems; break;
		case SET_COLOR:	item = setColorItems; break;
		case SET_MF:	item = setMFItems; break;
		default:	return(NULL);
	}

	while(item->name != NULL){
		if(!strcmp(item->name, name)) return(item->value);
		item = &item[1];
	}

	// scan plugins
	for(p = plug_set_head; p != NULL; p = p->next)
		if((type == p->type) && !strcmp(p->vname, name))
			return(p->value);

	return(NULL);
}


int findWild( int type, char *name, void (*output)(char *name, void *value) ){
	int len, a;
	s_set_item *item;
	s_plug_set *p;

	switch(type){
		case SET_NUM:	item = setIntItems; break;
		case SET_CHAR:	item = setCItems; break;
		case SET_STR:	item = setStrItems; break;
		case SET_COLOR:	item = setColorItems; break;
		case SET_MF:	item = setMFItems; break;
		default:	return(0);
	}

	a = 0;
	len = strlen(name);

	// Since we don't want to implement full patern matching, little hackish
	if(name[--len] == '*') {  // note the --len
		while(item->name){
			if(!strncmp(item->name, name, len)){
				output(item->name, item->value);
				a++;
			}
			item = &item[1];
		}

		// scan plugins
		for(p = plug_set_head; p != NULL; p = p->next){
			if((type == p->type) && !strncmp(p->vname, name, len)){
				output(p->vname, p->value);
				a++;
			}
		}
	} else {
		while(item->name != NULL){
			if(!strcmp(item->name, name)){
				output(item->name, item->value);
				a++;
			}
			item = &item[1];
		}

		// scan plugins
		for(p = plug_set_head; p != NULL; p = p->next){
			if((type == p->type) && !strcmp(p->vname, name)){
				output(p->vname, p->value);
				a++;
			}
		}
	}

	return(a);
}


e_color_name setMakeColor( char *value ){
	e_color_name color;

	stringtoupper(value);

	for(color=0; color<8; color++){
		if(!strcmp(value, color_list[color])) return(color);
	}
	msgGotoServer(NULL, MF_INVALID_COLOR, 1, value);
	return(BLACK);
}


int set_match_attr( char *attr ){
	stringtoupper(attr);

	if(*attr == '\0') return(0);
	if(!strcmp(attr, "BOLD")) return(BOLD);
	if(!strcmp(attr, "B")) return(BOLD);
	if(!strcmp(attr, "UNDERLINE")) return(UNDERLINE);
	if(!strcmp(attr, "U")) return(UNDERLINE);
	if(!strcmp(attr, "REVERSE-VIDEO")) return(REVERSE);
	if(!strcmp(attr, "REVERSE")) return(REVERSE);
	if(!strcmp(attr, "R")) return(REVERSE);
	if(!strcmp(attr, "NONE")) return(0);
	msgGotoServer(NULL, MF_INVALID_ATTR, 1, attr);
	return(0);
}


int setMakeAttr( char *value ){
	char *v2, *v3;

	if(*value == '\0') return(0);

	v2 = findchar(value, ',');
	if(*v2 != '\0'){
		*v2 = '\0';
		v2++;
		v3 = findchar(v2, ',');
		if(*v3 != '\0'){
			*v3 = '\0';
			v3++;
		}
		return(set_match_attr(value) | set_match_attr(v2) | set_match_attr(v3));
	}
	return(set_match_attr(value) | set_match_attr(v2));
}


//////////////////// Sets


void *setInt( char *name, char *value ){
	int num;
	void *item_value;
	char buffer[16];

	if((item_value = findValue(SET_NUM, name)) == NULL) return(NULL);

	// Found.. now validate
	if(setMakeNum(&num, value) == -1){
		msgGotoServer(NULL, MF_INVALID_SET_NUM, 2, value, name);
		return(NULL);
	}

	*(int *)item_value = num;
	snprintf(buffer, 16, "%d", *(int *)item_value);
	msgServer(3, NULL, MF_SET_ITEM_VALUE, 2, buffer, name);
	return(item_value);
}


void *setMF( char *name, char *value ){
	void *item_value;
	char *v;

	if(strncmp(name, "mf", 2) != 0) return(NULL);
	if((item_value = findValue(SET_MF, name+2)) == NULL) return(NULL);

	// found... any validation (of format str) here?

	// convert C-style control codes to single byte
	for(v = value; *v != '\0'; v++){
		if(*v == '\\'){
			switch(*(v + 1)){
			case 'n':	*v = '\n';
					strcpy(v + 1, v + 2);
					break;
			case 't':	*v = '\t';
					strcpy(v + 1, v + 2);
					break;
			case '\\':	*v = '\\';
					strcpy(v + 1, v + 2);
					break;
			}
		}
	}
	setMakeStr((char **)item_value, value);
	msgServer(3, NULL, MF_SET_ITEM_VALUE, 2, *(char **)item_value, name);
	return(item_value);
}


void *setStr( char *name, char *value ){
	void *item_value;

	if((item_value = findValue(SET_STR, name)) == NULL) return(NULL);

	// found... any validation here?
	setMakeStr(item_value, value);
	msgServer(3, NULL, MF_SET_ITEM_VALUE, 2, *(char **)item_value, name);
	return(item_value);
}


void *setC( char *name, char *value ){
	void *item_value;
	char buffer[2];

	if((item_value = findValue(SET_CHAR, name)) == NULL) return(NULL);

	// Validation
	if(strlen(value) != 1){
		msgGotoServer(NULL, MF_INVALID_SET_CHAR, 2, value, name);
		return(NULL);
	}

	*(char *)item_value = value[0];
	buffer[0] = *(char *)item_value;
	buffer[1] = '\0';
	msgServer(3, NULL, MF_SET_ITEM_VALUE, 2, buffer, name);
	return(item_value);
}


void *setColor( char *name, char *value ){
	s_color *color;
	char *bg_c, *attr_c, attrname[50];
	int attr;

	if(strncmp(name, "color_", 6) != 0) return(NULL);
	if((color = (s_color *)findValue(SET_COLOR, name+6)) == NULL) return(NULL);

	if(var_color){
		// split value into 3 components
		bg_c = skipword(value);
		attr_c = skipword(bg_c);
		color->fg = setMakeColor(value);
		color->bg = setMakeColor(bg_c);
		color->attr = setMakeAttr(attr_c);
//		if(color->bg != BLACK) color->attr |= BOLD;
	} else {
		color->attr = setMakeAttr(value);
	}

	attr = color->attr;
	attrname[0] = '\0';
	if(attr & BOLD) strncat(attrname, "BOLD", 50-strlen(attrname));
	if(attr & UNDERLINE){
		if(attr & BOLD) strncat(attrname, ",", 50-strlen(attrname));
		strncat(attrname, "UNDERLINE", 50-strlen(attrname));
	}
	if(attr & REVERSE){
		if(attr & BOLD || attr & UNDERLINE) strncat(attrname, ",", 50-strlen(attrname));
		strncat(attrname, "REVERSE-VIDEO", 50-strlen(attrname));
	}
	if(*attrname == '\0') strncpy(attrname, "NONE", 50);
	if(var_color)
		msgServer(3, NULL, MF_SET_COLOR, 4, attrname, color_list[color->bg],
			color_list[color->fg], name);
	else msgServer(3, NULL, MF_SET_ATTR, 2, attrname, name);

	return(color);
}


//////////////////// Gets


void getStr_output(char *name, void *value){
	if(*(char **)value != NULL)
		msgServer(3, NULL, MF_ITEM_VALUE, 2, *(char**)value, name);
	else
		msgServer(3, NULL, MF_ITEM_VALUE_NULL, 1, name);
}
void getInt_output(char *name, void *value){
	char buffer[16];

	snprintf(buffer, 16, "%d", *(char *)value);
	msgServer(3, NULL, MF_ITEM_VALUE, 2, buffer, name);
}
void getC_output(char *name, void *value){
	char buffer[2];

	buffer[0] = *(char *)value;
	buffer[1] = '\0';
	msgServer(3, NULL, MF_ITEM_VALUE, 2, buffer, name);
}


void getMF_output(char *name, void *value){
	char buffer[LEN_FORMAT_NAME];

	snprintf(buffer, LEN_FORMAT_NAME, "mf%s", name);
	if(*(char **)value != NULL)
		msgServer(3, NULL, MF_ITEM_VALUE, 2, *(char **)value, buffer);
	else
		msgServer(3, NULL, MF_ITEM_VALUE_NULL, 1, buffer);
}
int getMF( char *name ){
	if(strncmp(name, "mf", 2) != 0) return(0);
	return findWild(SET_MF, name+2, getMF_output);
}


void getColor_output(char *name, void *value){
	char fullname[LEN_COLOR_NAME], attrname[50];
	int attr;

	snprintf(fullname, LEN_COLOR_NAME, "color_%s", name);
	attr = ((s_color *)value)->attr;
	*attrname = '\0';
	if(attr & BOLD) strncat(attrname, "BOLD", 50-strlen(attrname));
	if(attr & UNDERLINE){
		if(attr & BOLD) strncat(attrname, ",", 50-strlen(attrname));
		strncat(attrname, "UNDERLINE", 50-strlen(attrname));
	}
	if(attr & REVERSE){
		if(attr & BOLD || attr & UNDERLINE) strncat(attrname, ",", 50-strlen(attrname));
		strncat(attrname, "REVERSE-VIDEO", 50-strlen(attrname));
	}
	if(*attrname == '\0') strncpy(attrname, "NONE", 50);
	if(var_color)
		msgServer(3, NULL, MF_COLOR, 4, attrname, color_list[((s_color *)value)->bg],
			color_list[((s_color *)value)->fg], fullname);
	else msgServer(3, NULL, MF_ATTR, 2, attrname, fullname);
}
int getColor( char *name ){
	if(strncmp(name, "color_", 6) != 0) return(0);
	return findWild(SET_COLOR, name+6, getColor_output);
}


////////////////////////////////////////////////////////////////////////// GLOBAL


void setRawColor( e_color_item i, e_color_name fg, e_color_name bg, int attr ){
	colors[i].fg = fg;
	colors[i].bg = bg;
	colors[i].attr = attr;
}


int setMakeNum( int *num, char *value ){
	// if ON, OFF, TRUE, FALSE, YES, NO set num to 0 or 1 and return 0
	char *end;

	*num = (int)strtol(value, &end, 10);
	if(end == value + strlen(value)) return(1);

	stringtolower(value);

	if(!strcmp(value, "on")) *num = 1; 
	else if(!strcmp(value, "off")) *num = 0;
	else if(!strcmp(value, "true")) *num = 1;
	else if(!strcmp(value, "false")) *num = 0;
	else if(!strcmp(value, "yes")) *num = 1;
	else if(!strcmp(value, "no")) *num = 0;
	else return(-1);

	return(0);
}


void setMakeStr( char **item, char *value ){
	char *copy;

	if(*item != NULL) track_free(*item, "setMakeStr");

	copy = (char *)track_malloc(strlen(value)+1, "setMakeStr");
	strcpy(copy, value);
	*item = copy;
}


void *setItem( char *item, char *value ){
	void *ret;
	stringtolower(item);

	// still don't like how we are doing this.
	if((ret = setInt(item, value))){
	} else if((ret = setC(item, value))){
	} else if((ret = setStr(item, value))){
	} else if((ret = setColor(item, value))){
	} else if((ret = setMF(item, value))){
	} else {
		msgGotoServer(NULL, MF_INVALID_SET_UNK, 2, value, item);
		return(NULL);
	}
	return(ret);
}


void getItem( char *name ){
	int a;

	stringtolower(name);

	a = 0;
	// These nots are kinda hacks.. oh well.  just keeping return values constitent, work on this....
	a += findWild(SET_STR, name, getStr_output);
	a += findWild(SET_NUM, name, getInt_output);
	a += findWild(SET_CHAR, name, getC_output);
	a += getMF(name);
	a += getColor(name);
	if(a == 0) msgGotoServer(NULL, MF_UNKNOWN_SET, 1, name);

   // TODO: if item ends with * then display all matches
	// I think we'll do this with strncmp for all entrys.
	// That way it doesn't have to be alphabetized.
	// How do we want to do this?  I might have to put all the tables
	// above together.
	// One possible way to do this is to implement ranges for types.
	// it would be sorta like this for each range.  (Inclusive-low, exclusive-high)
	// stuct range {
	//	int low;
	// 	int high;
	//	func fValidation;
	//	func fSet;
	//	func fGet;
	// }
}
