//	cria - user commands
//	Copyright (c) 2001-2005 Brand Huntsman
//


#include <sys/time.h>		// gettimeofday() for ping
#include <time.h>			// strftime()
#include <unistd.h>			// close() socket and /LOAD usleep()
#include <fcntl.h>			// non-blocking I/O for stdin
#include <sys/utsname.h>	// OS name/version and cpu
#include <sys/socket.h>		// userProxy()
#include <netinet/in.h>		// userProxy()
#include <arpa/inet.h>		// userProxy()
#include <dirent.h>			// userPlugList()

#include "common.h"
#include "functions.h"


//////////////////////////////////////////////////////////////////////////

// global
int total_bans, total_quiet, display_ban_totals;

// local
static int got_password;

////////////////////////////////////////////////////////////////////////// PRIVATE


static void password_callback( ){
	got_password = 1;
}


////////////////////////////////////////////////////////////////////////// GLOBAL


void userAbout( char *data, int datalen ){ // display about info
	// ABOUT
	char buffer[200];

	// ignore data

	snprintf(buffer, 200, "-- %s %s Copyright (c) %s by Brand Huntsman\n"
		"-- Additional coding by Nathaniel Pinckney and Leif Ames\n"
		"-- Translations by Niklas Klgel, Gerardo Canosa and Jukka Liukkonen",
		PROG_NAME, version_string, YEAR);
	msgGotoServer(NULL, MF_DEBUG, 1, buffer);
}


void userAdmin( char *data, int datalen ){ // get server admin info
	// ADMIN [server]

	netSend(n_current, "ADMIN :%s\n", data);
}


void userAlias( char *data, int datalen ){ // add or view command aliases
	// ALIAS [name] [body]
	char *body, *cmd;
	s_alias *a;

	// data = name
	// cmd = first word in body
	// body = rest of body
	cmd = skipword(data);
	body = skipword(cmd);

	stringtolower(data);

	if(datalen == 0){
		// display all aliases
		a = alias_head;
		if(a == NULL){
			msgGotoServer(NULL, MF_NO_ALIASES, 0);
		} else for(; a != NULL; a = a->next){
			msgGotoServer(NULL, MF_LIST_ALIASES, 2, a->body, a->name);
		}
	} else if(strlen(cmd) > 0){
		// check for recursive alias
		if(*cmd == '/'){
			if(stringcompare(cmd+1, data)){
				msgGotoServer(NULL, MF_RECURSIVE_ALIAS, 1, data);
				return;
			}
		}

		// scan aliases, plugins and built-in commands for name conflict
		if(cmdLookup(data)){
			msgGotoServer(NULL, MF_ALIAS_EXISTS, 1, data);
			return;
		}

		// create alias node
		a = (s_alias *)track_malloc(sizeof(s_alias), "userAlias -");
		a->name = (char *)track_malloc(strlen(data)+1, "userAlias name");
		strcpy(a->name, data);
		a->body = (char *)track_malloc(strlen(cmd)+strlen(body)+2, "userAlias body");
		sprintf(a->body, "%s %s", cmd, body);

		a->next = alias_head;
		alias_head = a;

		msgGotoServer(NULL, MF_NEW_ALIAS, 2, a->body, data);
	} else {
		// display alias
		for(a = alias_head; a != NULL; a = a->next){
			if(!strcmp(a->name, data)){
				msgGotoServer(NULL, MF_LIST_ALIASES, 2, a->body, a->name);
				return;
			}
		}
		if(strlen(data) > 0)
			msgGotoServer(NULL, MF_NO_SUCH_ALIAS, 1, data);
		else
			msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/ALIAS [name] [body]");
	}
}


void userAway( char *data, int datalen ){ // set self as away
	// AWAY [message]
	char *msg;
	s_network *n;

	if(datalen > 0) msg = data;
	else msg = var_away_msg;

	for(n = n_head; n != NULL; n = n->next)
		if(n->socket)
			netSend(n, "AWAY :%s\n", msg);

	msgGotoServer(n_current, MF_AWAY, 1, msg);
}


void userBack( char *data, int datalen ){ // set self as not away
	// BACK
	s_network *n;

	if(datalen > 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/BACK");
	} else {
		for(n = n_head; n != NULL; n = n->next)
			if(n->socket)
				netSend(n, "AWAY\n");

		msgServer(-1, n_current, MF_BACK, 0);
	}
}


void userBanList( char *data, int datalen ){ // display list of banned users
	// BANLIST

	if(datalen > 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/BANLIST [channel]");
	} else {
		display_ban_totals = 1;
		total_bans = total_quiet = 0;
		netSend(n_current, "MODE %s +b\n", p_current->name);
	}
}


void userBind( char *data, int datalen ){ // bind key to command
	// BIND
	// BIND key [body]
	// BIND ^key [body]
	char *body;
	s_bind *b;
	int key;

	if(datalen == 0){
		// display all binds
		b = bind_head;
		if(b == NULL){
			msgGotoServer(NULL, MF_NO_BINDS, 0);
		} else for(; b != NULL; b = b->next){
			msgGotoServer(NULL, MF_LIST_BINDS, 2, b->body, keynames[b->key].name);
		}
	} else {
		// data key
		body = skipword(data);
		stringtoupper(data);

		// find key
		for(key = 0; keynames[key].name != NULL; key++)
			if(!strcmp(keynames[key].name, data))
				break;
		if(keynames[key].name == NULL){
			msgGotoServer(NULL, MF_INVALID_BIND, 1, data);
			return;
		}

		// find bind
		for(b = bind_head; b != NULL; b = b->next)
			if(b->key == key)
				break;

		if(b != NULL){
			if(strlen(body) == 0){
				// display bind
				msgGotoServer(NULL, MF_LIST_BINDS, 2, b->body, data);
			} else {
				// bind exists
				msgGotoServer(NULL, MF_BIND_EXISTS, 1, data);
			}
		} else {
			if(strlen(body) > 0){
				// create bind node
				b = (s_bind *)track_malloc(sizeof(s_bind), "userBind -");
				b->key = key;
				b->body = (char *)track_malloc(strlen(body)+1, "userBind body");
				strcpy(b->body, body);

				b->next = bind_head;
				bind_head = b;

				msgGotoServer(NULL, MF_NEW_BIND, 2, body, data);
			} else {
				// not bound
				msgGotoServer(NULL, MF_NO_SUCH_BIND, 1, data);
			}
		}
	}
}


void userClear( char *data, int datalen ){ // clear page
	// CLEAR

	if(datalen > 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/CLEAR");
	} else {
		bufferClear(p_current);
		pageShow(p_current);
	}
}


void userClearTopic( char *data, int datalen ){ // clear channel topic
	// CLEARTOPIC

	if(datalen > 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/CLEARTOPIC");
	} else {
		netSend(n_current, "TOPIC %s :\n", p_current->name);
	}
}


void userConnect( char *data, int datalen ){ // connect to network
	// CONNECT
	char password[33];
	int recon_failed;

	if(datalen > 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/CONNECT");
	} else if(n_current->socket){
		msgGotoServer(n_current, MF_SERVER_CONNECTED, 1, n_current->server_name);
	} else {
		if(n_current->flags.x.password){
			promptSetup(C_IN_FG, C_IN_INFO, "Enter password: ", strlen("Enter password: "),
				password, 32, 1, 1, &password_callback);
			got_password = 0;
			recon_failed = 0;
			while(!got_password && recon_failed != -1)
				recon_failed = listenSockets(RECONNECT_DISABLED);
			if(strlen(password) > 0){
				setMakeStr(&n_current->password, password);
			} else {
				msgGotoServer(n_current, MF_DEBUG, 1, "Connection aborted");
				return;
			}
		}

		netConnect(n_current, NULL);
	}
}


void userCTCP( char *data, int datalen ){ // send ctcp messages
	// CTCP nick request

  // TODO:

}


void userCycle( char *data, int datalen ){ // leave and rejoin channel, dont close page
	// CYCLE

	if(datalen > 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/CYCLE");
	} else {
		p_current->flags.x.cycled = 1;
		if(p_current->flags.x.joined){
			p_current->flags.x.joined = 0;
			nicksFree(p_current);

			// part
			netSend(n_current, "PART %s :cycling\n", p_current->name);
		}

		p_current->flags.x.unjoined = 0;

		// join
		if(p_current->host_key)
			netSend(n_current, "JOIN %s %s\n", p_current->name, p_current->host_key);
		else
			netSend(n_current, "JOIN %s\n", p_current->name);

		// get channel mode
		netSend(n_current, "MODE %s\n", p_current->name);
	}
}


void userDetach( char *data, int datalen ){ // detach client from stdin/out and move to background
	// DETACH [message]

/*
	if(datalen > 0){

  // TODO: spam message to all channels on all networks

	}

	// disable ncurses
	endwin();

	// set stdin to blocking i/o
	fcntl(0, F_SETFL, stdin_settings & ~O_NONBLOCK);

	// disable output
	p_current = NULL;

  // TODO: fork and exit


// need function to handle resuming after being detached
	// restore settings for stdin
	tcsetattr(0, TCSANOW, &client_settings);
	// redraw screen
	pageShow(p_server);
*/
}


void userDisconnect( char *data, int datalen ){ // disconnect from network
	// DISCONNECT [message]

	if(n_current->socket == 0){
		if(n_current->current_timeout){
			n_current->current_timeout = 0;
			msgGotoServer(n_current, MF_DEBUG, 1, "Connection aborted");
		} else {
			msgGotoServer(n_current, MF_NOT_CONNECTED, 0);
		}
	} else {
		if(datalen)
			netSend(n_current, "QUIT :%s\n", data);
		else
			netSend(n_current, "QUIT :%s\n", var_quit_msg);

		// clear usermode on infobar
		n_current->umode = 0;

		close(n_current->socket);
		n_current->socket = 0;
		if(n_current->server_name){
			track_free(n_current->server_name, "userDisconnect:");
			n_current->server_name = NULL;
		}
		n_current->s_current = NULL;

		msgGotoServer(n_current, MF_DISCONNECT, 0);
	}
}


void userDump( char *data, int datalen ){ // dump file to current page
	// DUMP filename
	char c_buffer[LEN_BUFFER_MSG], path[LEN_PATHNAME];
	FILE *fp;

	if(datalen == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/DUMP filename");
	} else {
		if(*data == '~'){
			snprintf(path, LEN_PATHNAME, "%s%s", home, data+1);
		} else strncpy(path, data, LEN_PATHNAME);

		if((fp = fopen(path, "r")) == NULL){
			msgGotoServer(NULL, MF_LOAD_FAILED, 1, data);
		} else {
			if(p_current == p_server || *(p_current->name) == '!'){
				// server or plugin page (!dcc)
				msgServer(-1, n_current, MF_DEBUG, 1, "can't DUMP to this page");
			} else if(p_current->flags.x.joined){
				// dump to channel or user page

				if(n_current->socket == 0){
					msgGotoServer(n_current, MF_NETWORK_OFFLINE, 0);
				} else {
					while(fgets(c_buffer, LEN_BUFFER_MSG-1, fp) != NULL){
						if(keyGetCtrlC()){
							netSend(n_current, "PRIVMSG %s :--dump aborted--\n", p_current->name);
							msgCurrentPage(-1, NULL, MF_SENT_MESSAGE, 2, "--dump aborted--", n_current->nickname);
							goto abort;
						}
						strip_whitespace(c_buffer, SIDE_RIGHT);
						if(strlen(c_buffer) == 0 && !(p_current->mode & MODE_c))
							sprintf(c_buffer, " ");
						if(strlen(c_buffer) > 0){
							netSend(n_current, "PRIVMSG %s :%s\n",
								p_current->name, c_buffer);
							msgCurrentPage(-1, NULL, MF_SENT_MESSAGE, 2,
								c_buffer, n_current->nickname);
						}
  // TODO: this sleep should be handled inside netSend()

						sleep(1);		// 1 sec
					}
					if(*(p_current->name) != '#' && p_current->host_key == NULL)
						netSend(n_current, "USERHOST %s\n", p_current->name);
				}
			} else msgCurrentPage(-1, NULL, MF_NOT_IN_CHAN, 0);

			abort:
			fclose(fp);
		}
	}
}


void userFuncBackspace( char *data, int datalen ){ // backspace input
	keyBackspace();
}
void userFuncBeep( char *data, int datalen ){ // insert beep character in input
	if(!input_special){
		if(input_hidden) putchar('*');
		else {
			do_rev_on();
			putchar('G');
			do_rev_off();
		}
		keyInsert('\007');
	}
}
void userFuncBold( char *data, int datalen ){ // insert bold character in input
	if(!input_special){
		if(input_hidden) putchar('*');
		else {
			do_rev_on();
			putchar('B');
			do_rev_off();
		}
		keyInsert('\002');
	}
}
void userFuncClearLine( char *data, int datalen ){ // clear input line
	input_offset = 0;
	input_scrolled = 0;
	input_pos = 0;
	input_max = 0;

	keyDisplayInput();
}
void userFuncColor( char *data, int datalen ){ // insert color character in input
	if(!input_special){
		if(input_hidden) putchar('*');
		else {
			do_rev_on();
			putchar('C');
			do_rev_off();
		}
		keyInsert('\003');
	}
}
void userFuncRevvid( char *data, int datalen ){ // insert reverse video character in input
	if(!input_special){
		if(input_hidden) putchar('*');
		else {
			do_rev_on();
			putchar('V');
			do_rev_off();
		}
		keyInsert('\020');
	}
}
void userFuncUnderline( char *data, int datalen ){ // insert underline character in input
	if(!input_special){
		if(input_hidden) putchar('*');
		else {
			do_rev_on();
			putchar('U');
			do_rev_off();
		}
		keyInsert('\017');
	}
}


void userGotoPage( char *data, int datalen ){ // go to page
	// GOTOPAGE [network] name
	s_page *page;
	char *name;
	s_network *network;

	name = skipword(data);
	if(*name == '\0'){
		name = data;
		if(strlen(name) == 0){
			msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/GOTOPAGE [network] pagename");
			return;
		}
		network = n_current;
	} else {
		network = networkGet(data);
		if(network == NULL){
			msgGotoServer(NULL, MF_NO_SUCH_NETWORK, 1, data);
			return;
		}
	}

	page = pageGet(network, name);
	if(page == NULL){
		msgGotoServer(network, MF_NO_SUCH_PAGE, 1, name);
	} else {
		if(page->flags.x.hidden){
			page->flags.x.hidden = 0;	// unhide if hidden
			n_current->nr_hidden_pages--;
			nr_hidden_pages--;
		}
		pageShow(page);
	}
}


void userHelp( char *data, int datalen ){ // display command help
	// HELP [command]

msgCurrentPage(-1, NULL, MF_FINISH_ME, 1, "HELP");
msgCurrentPage(-1, NULL, MF_HELP_DOCS, 1, DOCDIR);

	if(datalen == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/HELP [command]");

  // TODO: display list of all commands, read from file

	} else {

  // TODO: display help for command, read from file

	}
}


void userHide( char *data, int datalen ){ // hide page
	// HIDE [network page]
	char *name;
	s_page *page;
	s_network *network;

	if(datalen == 0){
		if(*(p_current->name) == '!'){
			msgGotoServer(NULL, MF_CANT_HIDE_PAGE, 1, p_current->name);
		} else {
			p_current->flags.x.hidden = 1;
			n_current->nr_hidden_pages++;
			nr_hidden_pages++;

			pageShow(p_current->next);
		}
	} else {
		name = skipword(data);
		if(strlen(name) == 0){
			if((page = pageGet(n_current, data)) == NULL){
				msgGotoServer(n_current, MF_NO_SUCH_PAGE, 1, data);
			} else if(!page->flags.x.hidden){
				page->flags.x.hidden = 1;
				n_current->nr_hidden_pages++;
				nr_hidden_pages++;

				if(page == p_current) pageShow(p_current->next);
				else barsShow();
			}
		} else {
			if((network = networkGet(data)) == NULL){
				msgGotoServer(NULL, MF_NO_SUCH_NETWORK, 1, data);
			} else if((page = pageGet(network, name)) == NULL){
				msgGotoServer(network, MF_NO_SUCH_PAGE, 1, name);
			} else if(!page->flags.x.hidden){
				page->flags.x.hidden = 1;
				network->nr_hidden_pages++;
				nr_hidden_pages++;

				if(page == p_current) pageShow(p_current->next);
				else barsShow();
			}
		}
	}
}


void userIgnore( char *data, int datalen ){ // add ignore
	// IGNORE from_mask [types...]
	// IGNORE from_mask ["body_mask" [types...]]
	char *from, *host, *body, *type;
	s_ignore *i;
	uint t = 0;

	if(datalen == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/IGNORE from_mask [types...]");
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/IGNORE from_mask [\"body_mask\" [types...]]");
		return;
	}

	from = data;
	body = skipword(from);
	if(*body == '\"'){
		type = skipstring(body+1);

		if(type == NULL){
			msgCurrentPage(-1, NULL, MF_ERROR, 1, "IGNORE: invalid body mask");
			return;
		}

		// remove quotes around body_mask
		body++;
		body[strlen(body)-1] = '\0';
	} else {
		type = body;
		body = "";
	}

	host = skipnick(from);

	if(*from == '#' && *host){
		msgCurrentPage(-1, NULL, MF_ERROR, 1, "IGNORE: invalid from, channels can't have hostmasks");
		return;
	}

	while(*type != '\0'){
		char *next_type = skipword(type);

		if(stringcompare(type, "notice")) t |= IG_NOTICE;
		else if(stringcompare(type, "message")) t |= IG_MESSAGE;
		else if(stringcompare(type, "public")) t |= IG_PUBLIC;
		else if(stringcompare(type, "ctcp")) t |= IG_CTCP;
		else if(stringcompare(type, "join")) t |= IG_JOIN;
		else if(stringcompare(type, "part")) t |= IG_PART;
		else if(stringcompare(type, "quit")) t |= IG_QUIT;
		else if(stringcompare(type, "nick")) t |= IG_NICK;
		else {
			msgCurrentPage(-1, NULL, MF_ERROR, 1, "IGNORE: invalid type");
			return;
		}

		type = next_type;
	}
	if(t == 0) t = 0xffffffff;

	i = (s_ignore *)track_malloc(sizeof(s_ignore), "userIgnore");
	i->type = t;
	i->from = NULL;
	setMakeStr(&i->from, from);
	i->host = NULL;
	if(*host != '\0') setMakeStr(&i->host, host);
	i->body = NULL;
	if(*body != '\0') setMakeStr(&i->body, body);
	i->next = NULL;

	// add to list
	if(ignore_head == NULL){
		ignore_head = ignore_tail = i;
		i->prev = NULL;
	} else {
		ignore_tail->next = i;
		i->prev = ignore_tail;
		ignore_tail = i;
	}

	// display ignore message
	{
		char from[500], types[100];

		if(i->host) snprintf(from, 500, "%s!%s", i->from, i->host);
		else snprintf(from, 500, "%s", i->from);
		ignoreMakeTypes(types, i->type);

		msgCurrentPage(-1, NULL, MF_IGNORE, 3, (i->body ? i->body : "*"), from, types);
	}
}


void userIgnores( char *data, int datalen ){ // show ignore list
	s_ignore *i;

	if(datalen){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/IGNORES");
		return;
	}

	if(ignore_head == NULL){
		msgCurrentPage(-1, NULL, MF_NO_IGNORES, 0);
		return;
	}

	for(i = ignore_head; i != NULL; i = i->next){
		char from[500], types[100];

		if(i->host) snprintf(from, 500, "%s!%s", i->from, i->host);
		else snprintf(from, 500, "%s", i->from);
		ignoreMakeTypes(types, i->type);

		msgCurrentPage(-1, NULL, MF_LIST_IGNORES, 3, (i->body ? i->body : "*"), from, types);
	}
}


void userInfo( char *data, int datalen ){ // get server info
	// INFO [server]

	netSend(n_current, "INFO %s\n", data);
}


void userInvite( char *data, int datalen ){ // invite nick to channel
	// INVITE nick

	if(datalen == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/INVITE nick");
	} else {

  // TODO: do all irc networks require channel ops to use invite?

  // TODO: check for multiple nicks and send invite per nick

		netSend(n_current, "INVITE %s %s\n", data, p_current->name);
	}
}


void userIsOn( char *data, int datalen ){ // check if users are online
	// ISON nick [nick nick ...]

	if(datalen == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/ISON nick [nick nick ...]");
	} else {
		netSend(n_current, "ISON %s\n", data);
	}
}


void help_user_join( ){
	msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/JOIN channel [key]");
	msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/JOIN irc://servername[:port]/[channel] [key]");
}
void userJoin( char *data, int datalen ){ // join channel
	// JOIN channel [key]
	// JOIN irc://servername[:port]/[channel] [key]
	char channel[LEN_CHANNEL], *key;
	s_page *page;
	int x;

	if(datalen == 0){
		if(*(p_current->name) == '!'){
			help_user_join();
		} else if(p_current->invite_channel != NULL){
			// create page before sending commands
			page = pageCreate(n_current, p_current->invite_channel, NULL);
			pageShow(page);

			page->flags.x.unjoined = 0;

			// join
			x = netSend(n_current, "JOIN %s\n", p_current->invite_channel);

			// get channel mode
			if(x != -1) netSend(n_current, "MODE %s\n", p_current->invite_channel);

			track_free(p_current->invite_channel, "userJoin:");
			p_current->invite_channel = NULL;
		} else if(*(p_current->name) != '#'){
			// message pages without an invitation
			msgCurrentPage(3, NULL, MF_NO_INVITE, 0);
		} else if(!p_current->flags.x.joined){
			// a parted channel

			p_current->flags.x.unjoined = 0;

			// join
			if(p_current->host_key)
				x = netSend(n_current, "JOIN %s %s\n", p_current->name, p_current->host_key);
			else
				x = netSend(n_current, "JOIN %s\n", p_current->name);

			// get channel mode
			if(x != -1) netSend(n_current, "MODE %s\n", p_current->name);
		} else {
			help_user_join();
		}

		return;
	}

	if(*data == '#'){
		data++;
		if(datalen == 1){
			// length of data would now be zero after stripping the '#'
			help_user_join();
			return;
		}
	}

	// data = channel
	// key = key
	key = skipword(data);

	if( findchar(key, ' ') != (key + strlen(key)) ){
		help_user_join();
		return;
	}

	if(stringNcompare(data, "irc://", 6)){
		char *server, *port;
		s_network *n;
		s_server *s;

		server = data += 6;

		data = findchar(data, '/');
		if(*data != '\0'){
			*data = '\0';
			data++;
		}

		port = findchar(server, ':');
		if(*port != '\0'){
			*port = '\0';
			port++;
		}

		if(*server == '\0'){
			help_user_join();
			return;
		}

		if(strlen(port) == 0 || setMakeNum(&x, port) == -1) x = 6667;

		for(n = n_head; n != NULL; n = n->next){
			for(s = n->server_head; s != NULL; s = s->next)
				if(stringcompare(server, s->name))
					if((uint)x == s->port)
						goto found_server;
		}
		found_server:

		if(n == NULL){
			char msg[LEN_BUFFER_MSG];

			// add network
			snprintf(msg, LEN_BUFFER_MSG, "/network + %s", server);
			cmdFromUser(msg);

			// add server
			snprintf(msg, LEN_BUFFER_MSG, "/server + %s %s", server, port);
			cmdFromUser(msg);
		} else {
			// change current network, switch to server page
			n_current = n;
			pageShow(p_server);
		}

		// connect to server
		if(n_current->socket == 0)
			netConnect(n_current, NULL);

		if(*data == '#') data++;
		if(*data == '\0') return;
	}

  // TODO: "/j #1,0" parts all channels, get confirmation

	// create page before sending commands
	snprintf(channel, LEN_CHANNEL, "#%s", data);
	if((page = pageCreate(n_current, channel, key)) != NULL){
		page->flags.x.unjoined = 0;
		pageShow(page);
	}

	if(n_current->flags.x.registered){
		// join
		x = netSend(n_current, "JOIN #%s %s\n", data, key);

		// get channel mode
		if(x != -1) netSend(n_current, "MODE #%s\n", data);
	}
}


void userLeave( char *data, int datalen ){ // leave channel, close page
	// PART [message]

	if(*(p_current->name) == '!'){
		msgCurrentPage(3, NULL, MF_CANT_PART_PAGE, 0);
	} else {
		if(*(p_current->name) == '#' && p_current->flags.x.joined && n_current->socket)
			netSend(n_current, "PART %s :%s\n", p_current->name, data);

		pageFree(p_current);
	}
}


void userLeaveLeaveOpen( char *data, int datalen ){ // leave channel, leave page open
	// LL [message]

	if(p_current->flags.x.joined && n_current->socket)
		netSend(n_current, "PART %s :%s\n", p_current->name, data);

	p_current->flags.x.joined = 0;
	p_current->flags.x.chanop = 0;
	p_current->flags.x.halfchanop = 0;
	p_current->flags.x.voice = 0;
	p_current->flags.x.unjoined = 1;
	nicksFree(p_current);

	barsShow();
}


void userList( char *data, int datalen ){ // list all channels on network
	// LIST -ALL
	// LIST #channel

	if(stringcompare(data, "-ALL")){
		netSend(n_current, "LIST\n");
	} else if(findchar(data, ' ') == (data + strlen(data))){
		if(*data == '#') netSend(n_current, "LIST %s\n", data);
		else netSend(n_current, "LIST #%s\n", data);
	} else {
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/LIST -ALL");
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/LIST #channel");
	}
}


void userLoad( char *data, int datalen ){ // load and process file
	// LOAD [-q] filename
	char *file;

	if(datalen == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/LOAD [-q] filename");
	} else {
		file = skipword(data);
		if(strlen(file) == 0){
			configLoadFile(data, 0, LOAD_NOISY);
		} else {
			if(!strcmp(data, "-q"))
				configLoadFile(file, 0, LOAD_QUIET);
			else
				configLoadFile(file, 0, LOAD_NOISY);
		}
		usleep(100000);		// take a nap, 0.1 sec
		pageShow(p_current);	// refresh page just in case it was a color file
	}
}


void userMe( char *data, int datalen ){ // send action to page
	// ME message

	if(datalen == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/ME message");
	} else if(p_current == p_server){

  // TODO: test oper status and send wallops action

		msgServer(3, n_current, MF_CANT_TALK, 0);
	} else if(*(p_current->name) != '!'){
		if(p_current->flags.x.joined){
			netSend(n_current, "PRIVMSG %s :\001ACTION %s\001\n", p_current->name, data);
			msgCurrentPage(-1, NULL, MF_SENT_ACTION, 2, data, n_current->nickname);
		} else msgCurrentPage(-1, NULL, MF_NOT_IN_CHAN, 0);
	}
	// ignore when sent to ! pages such as !dcc
}


void userMode( char *data, int datalen ){ // get channel mode
	// MODE

	netSend(n_current, "MODE %s\n", p_current->name);
}


void userMotd( char *data, int datalen ){ // get message of the day
	// MOTD

	if(datalen > 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/MOTD");
	} else {
		display_motd = 1;
		netSend(n_current, "MOTD\n");
	}
}


void userMovePageLeft( char *data, int datalen ){ // move current page to the left
	s_page *p, *pp;

	if(p_head->next == NULL) return;

	if(p_current == p_head){
		// move to end of list

		for(p = p_head; p->next != NULL; p = p->next);

		p->next = p_current;				// new tail
		p_head = p_current->next;			// new head
		p_current->next = NULL;				// close tail
	} else {
		// swap p_current with previous page

		for(pp = p = p_head; p != NULL; pp = p, p = p->next)
			if(p == p_current)
				break;

		pp->next = p_current->next;			// move prev page
		p_current->next = pp;				// after current

		if(pp == p_head){
			p_head = p_current;			// new head
		} else {
			for(p = p_head; p->next != pp; p = p->next);

			p->next = p_current;			// move current
		}
	}

	barsShow();
}


void userMovePageRight( char *data, int datalen ){ // move current page to the right
	s_page *p, *pp;

	if(p_head->next == NULL) return;

	if(p_current->next == NULL){
		// move to head

		for(p = p_head; p->next != p_current; p = p->next);

		p->next = NULL;					// new tail
		p_current->next = p_head;			// new second page
		p_head = p_current;				// new head
	} else {
		// swap p_current with next page

		if(p_current == left_page)
			left_page = p_current->next;

		if(p_current == p_head){
			pp = p_current->next;			// swap
			p_current->next = pp->next;		// p_current
			pp->next = p_current;			// with
			p_head = pp;				// next page
		} else {
			for(pp = p = p_head; p != NULL; pp = p, p = p->next)
				if(p == p_current)
					break;			// find page before p_current

			pp->next = p_current->next;		// swap
			pp = p_current->next;			// p_current
			p_current->next = pp->next;		// with
			pp->next = p_current;			// next page
		}
	}

	barsShow();
}


void userMsay( char *data, int datalen ){ // send to all channels
	// MSAY message
	s_page *p;
	s_network *n;

	if(datalen == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/MSAY message");
	} else {
		for(p = p_head; p != NULL; p = p->next){
			n = p->network;
			if(*(p->name) == '#' && p->flags.x.joined && n->socket){
				netSend(n, "PRIVMSG %s :%s\n", p->name, data);
				msgPage(p, -1, NULL, MF_SENT_MESSAGE, 2, data, n->nickname);
			}
		}
	}
}


void userMsg( char *data, int datalen ){ // send message
	// MSG #channel_or_nick message
	char *msg;

	// data = #channel or nick
	// msg = message
	msg = skipword(data);

	if(strlen(msg) == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/MSG #channel_or_nick message");
	} else {
		netSend(n_current, "PRIVMSG %s :%s\n", data, msg);
		msgUnkPage(n_current, data, -1, NULL, MF_SENT_MESSAGE, 2, msg, n_current->nickname);
	}
}


void userMsgMe( char *data, int datalen ){ // send action message
	// MME #channel_or_nick message
	char *msg;

	// data = #channel or nick
	// msg = message
	msg = skipword(data);

	if(strlen(msg) == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/MME #channel_or_nick message");
	} else {
		netSend(n_current, "PRIVMSG %s :\001ACTION %s\001\n", data, msg);
		msgUnkPage(n_current, data, -1, NULL, MF_SENT_ACTION, 2, msg, n_current->nickname);
	}
}


void userNames( char *data, int datalen ){ // display all names on channel
	// NAMES

	if(datalen > 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/NAMES");
	} else {
		nicksFree(p_current);
		netSend(n_current, "NAMES %s\n", p_current->name);
	}
}


void help_user_network( ){
	msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/NETWORK *");
	msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/NETWORK + network_name");
	msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/NETWORK - network_name");
	msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/NETWORK NICK [your_nickname]");
	msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/NETWORK MODE [your_mode]");
	msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/NETWORK AUTO yes/no");
}
void userNetwork( char *data, int datalen ){ // add/remove network
	// NETWORK network_name
	// NETWORK *
	// NETWORK +/- network_name
	// NETWORK NICK [your_nickname]
	// NETWORK MODE [your_usermode]
	// NETWORK AUTO yes/no
	// NETWORK PASS yes/no
	char server_id[16], *value;
	s_network *n;
	int x;

	if(datalen == 0){
		help_user_network();
	} else {
		value = skipword(data);

		if(p_current != p_server) pageShow(p_server);

		if(!strcmp(data, "*")){
			for(n = n_head, x = 0; n != NULL; n = n->next, x++){
				msgGotoServer(NULL, MF_LIST_NETWORKS, 1, n->name);

  // TODO: display all info about the network

			}
			snprintf(server_id, 16, "%d", x);
			msgGotoServer(NULL, MF_TOTAL_NETWORKS, 1, server_id);
		} else if(!strcmp(data, "+")){
			if(strlen(value) > 0){
				n = networkCreate(value, "", umodeToInt(0, "+iw"), 0);
				n_current = n;
				msgGotoServer(NULL, MF_ADD_NETWORK, 1, value);
				_UPDATE_TITLEBAR();
			} else {
				msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/NETWORK + network_name");
			}
		} else if(!strcmp(data, "-")){
			if(strlen(value) > 0){

  // TODO: remove server from current network
msgCurrentPage(-1, NULL, MF_FINISH_ME, 1, "NETWORK - network_name");

			} else {
				msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/NETWORK - network_name");
			}
		} else if(stringcompare(data, "nick")){
			setMakeStr(&n_current->requested_nickname, value);
			msgGotoServer(n_current, MF_NETSET_NICK, 1, value);
		} else if(stringcompare(data, "mode")){
			n_current->requested_umode = umodeToInt(0, value);
			if(n_current->socket == 0) n_current->umode = n_current->requested_umode;
			msgGotoServer(n_current, MF_NETSET_UMODE, 1, value);
		} else if(stringcompare(data, "auto")){
			if(strlen(value) == 0 || setMakeNum(&x, value) == -1){
				msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/NETWORK AUTO yes/no");
			} else {
				if(x) n_current->flags.x.autoconn = 1;
				else n_current->flags.x.autoconn = 0;
				msgGotoServer(n_current, MF_NETSET_AUTO, 1, value);
			}
		} else if(stringcompare(data, "pass")){
			if(strlen(value) == 0 || setMakeNum(&x, value) == -1){
				msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/NETWORK PASS yes/no");
			} else {
				if(x) n_current->flags.x.password = 1;
				else n_current->flags.x.password = 0;
				msgGotoServer(n_current, MF_NETSET_PASS, 1, value);
			}
		} else if((n = networkGet(data)) != NULL){
			n_current = n;
			pageShow(p_server);
		} else {
			help_user_network();
		}
	}
}


void userNextNetwork( char *data, int datalen ){ // change to next network
	// NEXTNETWORK

	if(!input_special){
		n_current = n_current->next;
		if(n_current == NULL) n_current = n_head;

		if(p_current != p_server) pageShow(p_server);
		else _UPDATE_TITLEBAR()

		barsShow();
	}
}


void userNick( char *data, int datalen ){ // change nick
	// NICK newnick

	if(datalen == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/NICK newnick");
	} else {
		netSend(n_current, "NICK %s\n", data);
	}
}


void userNotice( char *data, int datalen ){ // send notice to nick or channel
	// NOTICE #channel_or_nick message
	char *msg;

	// data = #channel/nick
	// msg = message
	msg = skipword(data);

	if(strlen(msg) == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/NOTICE #channel_or_nick message");
	} else {
		netSend(n_current, "NOTICE %s :%s\n", data, msg);
		msgUnkPage(n_current, data, -1, NULL, MF_SENT_NOTICE, 2, msg, n_current->nickname);
	}
}


void userPageLeft( char *data, int datalen ){ // switch to left page
	// PAGELEFT
	s_page *p, *pc;

	if(!input_special){

  // TODO: why does this test input_special???

		pc = p_current;
		do{
			p = p_head;
			if(pc == p_head){
				while(p->next != NULL) p = p->next;
			} else {
				while(p->next != pc) p = p->next;
			}
			pc = p;
		} while(pc->flags.x.hidden);

		pageShow(pc);
		keyDisplayInput();
	}
}


void userPageRight( char *data, int datalen ){ // switch to right page
	// PAGERIGHT

	if(!input_special){

  // TODO: why does this test input_special???

		pageShow(p_current->next);

		keyDisplayInput();
	}
}


void userPageTop( char *data, int datalen ){ // switch to left-most page
	// PAGETOP

	if(!input_special) pageShow(p_head);

  // TODO: why does this test input_special???

}


void userPing( char *data, int datalen ){ // send ping request to nick or channel
	// PING [nick]
	struct timeval tv;

	if(datalen > 0){
		// ping nick or nicks

		skipword(data); // allow one nick

  // TODO: check for multiple nicks, send ping to each

		msgUnkPage(n_current, data, -1, NULL, MF_SENT_CTCP_USER, 2, "PING", data);

		gettimeofday(&tv, NULL);
		netSend(n_current, "PRIVMSG %s :\001PING %d %d\001\n", data,
			(int)tv.tv_sec, (int)tv.tv_usec);
	} else if(*(p_current->name) == '#'){
		// ping every nick in channel

		if(p_current->flags.x.joined){
			msgCurrentPage(-1, NULL, MF_SENT_CTCP_CHANNEL, 2, "PING", p_current->name);

			gettimeofday(&tv, NULL);
			netSend(n_current, "PRIVMSG %s :\001PING %d %d %s\001\n", p_current->name,
				(int)tv.tv_sec, (int)tv.tv_usec, p_current->name);
		} else msgCurrentPage(-1, NULL, MF_NOT_IN_CHAN, 0);
	} else if(*(p_current->name) != '!'){
		// ping nick in message page

		msgCurrentPage(-1, NULL, MF_SENT_CTCP_USER, 2, "PING", p_current->name);

		gettimeofday(&tv, NULL);
		netSend(n_current, "PRIVMSG %s :\001PING %d %d\001\n", p_current->name,
			(int)tv.tv_sec, (int)tv.tv_usec);
	} else if(p_current == p_server){
		// ping server

		if(n_current->server_name != NULL)
			msgServer(-1, n_current, MF_SENT_PING_SERVER, 1, n_current->server_name);

		gettimeofday(&tv, NULL);
		netSend(n_current, "PING %s %d.%d\n", n_current->server_name, (int)tv.tv_sec, (int)tv.tv_usec);
	}
}


void userPlug( char *data, int datalen ){ // load plugin
	// PLUG plugin_name

	if(datalen == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/PLUG plugin_name");
	} else {
		plugLoad(data);
	}
}


void userPlugList( char *data, int datalen ){ // list plugins
	// PLUGS
	s_plug *plug;
	int lp, nlp;
	char lp_str[16], nlp_str[16];

	if(datalen > 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/PLUGS");
	} else {
		lp = nlp = 0;

		// loaded plugins
		for(plug = plug_head; plug != NULL; plug = plug->next){
			msgGotoServer(NULL, MF_LIST_LPLUGS, 1, plug->name);
			lp++;
		}

		// unloaded plugins
		#ifdef COREPLUGS
		{
			int x;

			// display each plug from coreplugs list that is not in plug_head list
			for(x = 0; coreplugs[x].name != NULL; x++){
				for(plug = plug_head; plug != NULL; plug = plug->next){
					if(stringcompare(plug->name, (char *)coreplugs[x].name))
						break;
				}
				if(plug == NULL){
					msgGotoServer(NULL, MF_LIST_NLPLUGS, 1, coreplugs[x].name);
					nlp++;
				}
			}
		}
		#else
		{
			DIR *d = opendir(PLUGDIR);

			if(d != NULL){
				struct dirent *f;

				while((f = readdir(d)) != NULL){
					if(f->d_name[0] != '.'){
						char p[100];
						int len = strlen(f->d_name)-2;

						snprintf(p, len < 100 ? len : 100, "%s", f->d_name);
						for(plug = plug_head; plug != NULL; plug = plug->next){
							if(stringcompare(plug->name, p))
								break;
						}
						if(plug == NULL){
							msgGotoServer(NULL, MF_LIST_NLPLUGS, 1, p);
							nlp++;
						}
					}
				}
			}
		}
		#endif

		snprintf(lp_str, 16, "%d", lp);
		snprintf(nlp_str, 16, "%d", nlp);
		msgGotoServer(NULL, MF_TOTAL_PLUGS, 2, nlp_str, lp_str);
	}
}


void userProxy( char *data, int datalen ){ // setup socks proxy
	// PROXY [ip [port]]
	// PROXY
	char *ip, *port;
	struct in_addr inp;

	if(datalen == 0){
		proxy = 0;
		msgServer(3, NULL, MF_PROXY_DISABLED, 0);
	} else {
		ip = data;
		port = skipword(ip);
		data = skipword(port);

		if(*port == '\0') port = "1080";
		proxy_port = cria_atoui(port);

		if(!inet_aton(ip, &inp) || proxy_port == 0 || *data != '\0'){
			msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/PROXY [ip [port]]");
		} else {
			proxy = 4;
			proxy_ip = inp.s_addr;
			msgServer(3, NULL, MF_PROXY_ENABLED, 2, port, ip);
		}
	}
}


void userQuit( char *data, int datalen ){ // shutdown client
	// QUIT [message]

	if(datalen == 0){
		shutdownClient(var_quit_msg);
	} else {
		shutdownClient(data);
	}
}


void userRaw( char *data, int datalen ){ // send raw IRC command
	// RAW command

	if(datalen == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/RAW raw_irc_command");
	} else {
		// <micro$oft>i dont feel like fixing this strcat() :)</micro$oft>
		strcat(data, "\n");

		if(p_current == p_server){
			msgCurrentPage(-1, n_current, MF_SENT_RAW, 1, data);
			netSend(n_current, data);
		} else if(*(p_current->name) != '!'){
			msgCurrentPage(-1, NULL, MF_SENT_RAW, 1, data);
			netSend(n_current, data);
		}
	}
}


void userReconnect( char *data, int datalen ){ // reconnect to server
	// RECONNECT

	if(n_current->socket > 0){
		msgGotoServer(n_current, MF_SERVER_RECONNECT, 0);
		netSend(n_current, "QUIT :reconnecting\n");
	}
	netConnect(n_current, n_current->s_current);
}


void userRefresh( char *data, int datalen ){ // refresh current page
	// REFRESH

	pageShow(p_current);
}


void userRename( char *data, int datalen ){ // rename message page
	// RENAME new_nick
	s_page *page;

	if(datalen == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/RENAME new_nick");
	} else if(*(p_current->name) != '#' && *(p_current->name) != '!'){
		if(!stringcompare(p_current->name, data)){
			page = pageGet(n_current, data);
			if(page == NULL){
				// new_nick page does not exist, rename current to new_nick
				setMakeStr(&p_current->name, data);

				if(n_current->socket)
					netSend(n_current, "USERHOST %s\n", data);

				barsShow();
			} else {
				// new_nick page exists, merge pages
				msgCurrentPage(3, NULL, MF_MERGE_PAGES, 1, p_current->name);

				// prepend current buffer onto page->buffer
				page->nr_lines += p_current->nr_lines;
				p_current->buffer_tail->next = page->buffer_head;
				page->buffer_head->prev = p_current->buffer_tail;
				page->buffer_head = p_current->buffer_head;

				// recalculate top_line
				msgScrollPage(page, SCROLL_PAGE_UP);

				// resize buffer, if needed
				bufferResizePage(page);

				// free old_nick page
				p_current->buffer_head = NULL;
				pageFree(p_current);

				// switch to new_nick page
				pageShow(page);
			}
		} else msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/RENAME new_nick");
	} else {
		msgGotoServer(NULL, MF_TEMP_HELP, 1, "/RENAME only works on message pages.");
	}
}


void userSave( char *data, int datalen ){ // save settings
	// SAVE

  // TODO: save info, colors, formats or servers. type will be in data

}


void userSaveAll( char *data, int datalen ){ // save all settings
	// SAVEALL

  // TODO: save all info, colors, formats and servers

}


void userSay( char *data, int datalen ){ // send to current page
	// SAY message

	if(datalen == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/SAY message");
	} else if(p_current == p_server){

  // TODO: check oper status and send wallops

		msgServer(3, n_current, MF_CANT_TALK, 0);
	} else {
		if(*(p_current->name) == '#' && !p_current->flags.x.joined){
			msgCurrentPage(-1, NULL, MF_NOT_IN_CHAN, 0);
		} else {
			netSend(n_current, "PRIVMSG %s :%s\n", p_current->name, data);
			msgCurrentPage(-1, NULL, MF_SENT_MESSAGE, 2, data, n_current->nickname);
		}
	}
}


void userSendTime( char *data, int datalen ){ // send time to page
	// ST [nick_or_channel]
	time_t ct;
	struct tm *tp;
	char buffer[150], *who;

	if(datalen == 0 && *(p_current->name) == '!'){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/ST [nick_or_channel]");
		msgGotoServer(NULL, MF_TEMP_HELP, 1, "/ST only works on channel and user pages.");
	} else {
		if(datalen) who = data;
		else who = p_current->name;

		ct = time(NULL);
		tp = localtime(&ct);
		strftime(buffer, 150, var_time_format, tp);

		netSend(n_current, "PRIVMSG %s :%s\n", who, buffer);
		msgUnkPage(n_current, who, -1, NULL, MF_SENT_MESSAGE, 2, buffer, n_current->nickname);
	}
}


void userSendVersion( char *data, int datalen ){ // send version to page
	// SV [nick_or_channel]
	struct utsname uts;
	char buffer[150], *who;

	if(datalen == 0 && *(p_current->name) == '!'){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/SV [nick_or_channel]");
		msgGotoServer(NULL, MF_TEMP_HELP, 1, "/SV only works on channel and user pages.");
	} else {
		if(datalen) who = data;
		else who = p_current->name;

		uname(&uts);
		snprintf(buffer, 150, "%s %s [%s %s %s] -- %s", PROG_NAME, version_string,
			uts.sysname, uts.release, uts.machine, PROG_URL);

		netSend(n_current, "PRIVMSG %s :%s\n", who, buffer);
		msgUnkPage(n_current, who, -1, NULL, MF_SENT_MESSAGE, 2, buffer, n_current->nickname);
	}
}


void userServer( char *data, int datalen ){ // add/remove servers
	// SERVER
	// SERVER *
	// SERVER CHANGE [server_number]
	// SERVER +/- server_name [port]
	char server_id[16], *value, *port;
	s_server *s;
	int x;

	if(datalen == 0){
		if(n_current->server_name != NULL)
			msgCurrentPage(3, n_current, MF_CURRENT_SERVER, 1, n_current->server_name);
		else
			msgCurrentPage(3, n_current, MF_SERVER_OFFLINE, 0);

		return;
	}

	value = skipword(data);

	if(!strcmp(data, "*")){
		for(s = n_current->server_head, x = 0; s != NULL; s = s->next, x++){
			snprintf(server_id, 16, "%d", x);
			msgGotoServer(n_current, MF_LIST_SERVERS, 2, s->name, server_id);

  // TODO: display all info about servers for current network

		}
		snprintf(server_id, 16, "%d", x);
		msgGotoServer(n_current, MF_TOTAL_SERVERS, 1, server_id);
	} else if(stringcompare(data, "change")){
		if(strlen(value) > 0){
			if(setMakeNum(&x, value)){
				for(s = n_current->server_head; x && s != NULL; s = s->next)
					x--;

				if(s == NULL){
					msgGotoServer(n_current, MF_NO_SUCH_SERVER, 1, value);
				} else {
					if(n_current->socket){
						msgGotoServer(n_current, MF_SERVER_CHANGE, 0);
						netSend(n_current, "QUIT :changing servers\n");
					}
					netConnect(n_current, s);
				}
			} else {
				msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/SERVER CHANGE [server_number]");
			}
		} else {
			s = n_current->s_current->next;
			if(s == NULL) s = n_current->server_head;
			if(s != n_current->s_current){
				if(n_current->socket){
					msgGotoServer(n_current, MF_SERVER_CHANGE, 0);
					netSend(n_current, "QUIT :changing servers\n");
				}
				netConnect(n_current, s);
			} else msgGotoServer(n_current, MF_SERVER_CONNECTED, 1, n_current->server_name);
		}
	} else if(!strcmp(data, "+")){
		if(strlen(value) > 0){
			port = skipword(value);
			if(strlen(port) == 0 || setMakeNum(&x, port) == -1) x = 6667;
			serverCreate(n_current, value, x);
			snprintf(server_id, 16, "%d", x);
			msgGotoServer(n_current, MF_ADD_SERVER, 2, server_id, value);
		} else {
			msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/SERVER + server_name [port]");
		}
	} else if(!strcmp(data, "-")){
		if(strlen(value) > 0){
			port = skipword(value);

  // TODO: remove server
msgCurrentPage(-1, NULL, MF_FINISH_ME, 1, "SERVER - server_name [port]");

		} else {
			msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/SERVER - server_name [port]");
		}
	} else {
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/SERVER");
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/SERVER *");
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/SERVER CHANGE [server_number]");
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/SERVER + server_name [port]");
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/SERVER - server_name [port]");
	}
}


void userSet( char *data, int datalen ){ // set or get item value
	// SET item value
	// SET item
	// SET partial_item*
	char *value;
	int vlen;
	void *var_ptr;

	if(datalen == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/SET name [value]");
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/SET partial_name*");
	} else {
		// data = item
		// value = value
		value = skipword(data);

		vlen = strlen(value);
		if(vlen > 0){
			vlen--;
			if(value[vlen] == '\"') value[vlen] = '\0';
			if(*value == '\"') value++;
			var_ptr = setItem(data, value);

			if(var_ptr == (void *)&var_buffer_max){
				bufferResizeAll();
			} else if(var_ptr == (void *)&var_history_max){

  // TODO: resize history

			} else if(var_ptr == (void *)&var_color){
				if(var_color) configEnableColor();
				else configDisableColor();
			} else if(var_ptr == (void *)&var_timestamps){
				pageShow(p_current);
			} else if(!strncmp(data, "color_", 6) && !suppress_messages){
				pageShow(p_current);
			}
		} else {
			getItem(data);
		}
	}
}


void userShowHidden( char *data, int datalen ){ // show hidden pages
	// SHOWHIDDEN
	s_page *page;
	s_network *network;
	char page_list[200];
	int list_len, page_len, n;

	if(datalen > 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/SHOWHIDDEN");
	} else {
		n = 0;
		for(network = n_head; network != NULL; network = network->next){
			list_len = 0;
			for(page = p_head; page != NULL; page = page->next){
				if(page->network == network && page->flags.x.hidden){
					page_len = strlen(page->name) + 2;
					if(list_len == 0){
						strncpy(page_list, page->name, 200);
						list_len = page_len - 2;
					} else if(list_len + page_len < 200){
						strncat(page_list, ", ", 200-strlen(page_list));
						strncat(page_list, page->name, 200-strlen(page_list));
						list_len += page_len;
					} else {
						msgServer(3, network, MF_LIST_HIDDEN_PAGES, 1, page_list);
						strncpy(page_list, page->name, 200);
						list_len = page_len - 2;
						n++;
					}
				}
			}
			if(list_len){
				msgServer(3, network, MF_LIST_HIDDEN_PAGES, 1, page_list);
				n++;
			}
		}
		if(n == 0)
			msgServer(3, NULL, MF_NO_HIDDEN_PAGES, 0);
	}
}


void userTime( char *data, int datalen ){ // get server time or user's time
	// TIME [server]
	// TIME [nick]

	if(p_current == p_server){
		// server
		netSend(n_current, "TIME %s\n", data);
	} else if(datalen){
		// selected user
		skipword(data);
		netSend(n_current, "PRIVMSG %s :\001TIME\001\n", data);
		msgUnkPage(n_current, data, -1, NULL, MF_SENT_CTCP_USER, 2, "TIME", data);
	} else if(*(p_current->name) == '#'){
		// channel
		if(p_current->flags.x.joined){
			netSend(n_current, "PRIVMSG %s :\001TIME\001\n", p_current->name);
			msgCurrentPage(-1, NULL, MF_SENT_CTCP_CHANNEL, 2, "TIME", p_current->name);
		} else msgCurrentPage(-1, NULL, MF_NOT_IN_CHAN, 0);
	} else if(*(p_current->name) != '!'){
		// user
		netSend(n_current, "PRIVMSG %s :\001TIME\001\n", p_current->name);
		msgCurrentPage(-1, NULL, MF_SENT_CTCP_USER, 2, "TIME", p_current->name);
	}
}


void userTopic( char *data, int datalen ){ // set or display channel topic
	// TOPIC [topic]

	if(datalen == 0)
		netSend(n_current, "TOPIC %s\n", p_current->name);
	else
		netSend(n_current, "TOPIC %s :%s\n", p_current->name, data);
}


void userUnalias( char *data, int datalen ){ // remove alias
	// UNALIAS name
	s_alias *a, *pa;

	if(datalen == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/UNALIAS name");
	} else {
		stringtolower(data);

		for(a = pa = alias_head; a != NULL; pa = a, a = a->next)
			if(!strcmp(a->name, data))
				break;
		if(a == NULL){
			msgGotoServer(NULL, MF_NO_SUCH_ALIAS, 1, data);
		} else {
			// remove alias
			if(a == alias_head) alias_head = a->next;
			else pa->next = a->next;
			track_free(a->name, "userUnalias: name");
			track_free(a->body, "userUnalias: body");
			track_free(a, "userUnalias:");
			msgGotoServer(NULL, MF_CLEAR_ALIAS, 1, data);
		}
	}
}


void userUnbind( char *data, int datalen ){ // remove bind
	// UNBIND key
	s_bind *b, *pb;
	int key;

	if(datalen == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/UNBIND name");
	} else {
		stringtoupper(data);

		// lookup key
		for(key = 0; keynames[key].name != NULL; key++)
			if(!strcmp(keynames[key].name, data))
				break;
		if(keynames[key].name == NULL){
			msgGotoServer(NULL, MF_INVALID_BIND, 1, data);
		} else {
			// lookup bind
			for(b = pb = bind_head; b != NULL; pb = b, b = b->next)
				if(b->key == key)
					break;
			if(b == NULL){
				msgGotoServer(NULL, MF_NO_SUCH_BIND, 1, data);
			} else {
				// remove bind
				if(b == bind_head) bind_head = b->next;
				else pb->next = b->next;
				track_free(b->body, "userUnbind: body");
				track_free(b, "userUnbind:");
				msgGotoServer(NULL, MF_CLEAR_BIND, 1, data);
			}
		}
	}
}


void userUnhide( char *data, int datalen ){ // unhide page
	// UNHIDE [network] page
	char *name;
	s_page *page;
	s_network *network;

	if(datalen == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/UNHIDE [network] page");
	} else {
		name = skipword(data);
		if(strlen(name) == 0){
			if((page = pageGet(n_current, data)) == NULL){
				msgGotoServer(n_current, MF_NO_SUCH_PAGE, 1, data);
			} else if(page->flags.x.hidden){
				page->flags.x.hidden = 0;
				n_current->nr_hidden_pages--;
				nr_hidden_pages--;

				barsShow();
			}
		} else {
			if((network = networkGet(data)) == NULL){
				msgGotoServer(NULL, MF_NO_SUCH_NETWORK, 1, data);
			} else if((page = pageGet(network, name)) == NULL){
				msgGotoServer(network, MF_NO_SUCH_PAGE, 1, name);
			} else if(page->flags.x.hidden){
				page->flags.x.hidden = 0;
				network->nr_hidden_pages--;
				nr_hidden_pages--;

				barsShow();
			}
		}
	}
}


void userUnignore( char *data, int datalen ){ // remove ignore
	// UNIGNORE *
	// UNIGNORE from_mask [types...]
	// UNIGNORE from_mask ["body_mask" [types...]]
	char *from, *host, *body, *type;
	s_ignore *i;
	uint t = 0;

	if(datalen == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/UNIGNORE *");
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/UNIGNORE from_mask [types...]");
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/UNIGNORE from_mask [\"body_mask\" [types...]]");
		return;
	}

	if(stringcompare(data, "*")){
		s_ignore *i, *next;

		msgCurrentPage(-1, NULL, MF_CLEAR_IGNORES, 0);

		for(i = ignore_head; i != NULL;){
			next = i->next;
			track_free(i->from, "userUnignore");
			if(i->host) track_free(i->host, "userUnignore");
			if(i->body) track_free(i->body, "userUnignore");
			track_free(i, "userUnignore");
			i = next;
		}
		ignore_head = ignore_tail = NULL;
		return;
	}

	from = data;
	body = skipword(from);
	if(*body == '\"'){
		type = skipstring(body+1);

		if(type == NULL){
			msgCurrentPage(-1, NULL, MF_ERROR, 1, "UNIGNORE: invalid body mask");
			return;
		}

		// remove quotes around body_mask
		body++;
		body[strlen(body)-1] = '\0';
	} else {
		type = body;
		body = "";
	}

	host = skipnick(from);

	if(*from == '#' && *host){
		msgCurrentPage(-1, NULL, MF_ERROR, 1, "UNIGNORE: invalid from, channels can't have hostmasks");
		return;
	}

	while(*type != '\0'){
		char *next_type = skipword(type);

		if(stringcompare(type, "notice")) t |= IG_NOTICE;
		else if(stringcompare(type, "message")) t |= IG_MESSAGE;
		else if(stringcompare(type, "public")) t |= IG_PUBLIC;
		else if(stringcompare(type, "ctcp")) t |= IG_CTCP;
		else if(stringcompare(type, "join")) t |= IG_JOIN;
		else if(stringcompare(type, "part")) t |= IG_PART;
		else if(stringcompare(type, "quit")) t |= IG_QUIT;
		else if(stringcompare(type, "nick")) t |= IG_NICK;
		else {
			msgCurrentPage(-1, NULL, MF_ERROR, 1, "UNIGNORE: invalid type");
			return;
		}

		type = next_type;
	}
	if(t == 0) t = 0xffffffff;

	// find ignore
	for(i = ignore_head; i != NULL; i = i->next){
		if(i->type != t)
			continue;
		if(!stringcompare(i->from, from))
			continue;
		if(i->host && !stringcompare(i->host, host))
			continue;
		if(i->host == NULL && !stringcompare("", host))
			continue;
		if(i->body && !stringcompare(i->body, body))
			continue;
		if(i->body == NULL && !stringcompare("", body))
			continue;
		break;
	}
	if(i == NULL){
		msgCurrentPage(-1, NULL, MF_ERROR, 1, "UNIGNORE: no such ignore, can't remove");
	} else {
		char from[500], types[100];

		if(i->host) snprintf(from, 500, "%s!%s", i->from, i->host);
		else snprintf(from, 500, "%s", i->from);
		ignoreMakeTypes(types, i->type);

		msgCurrentPage(-1, NULL, MF_UNIGNORE, 3, (i->body ? i->body : "*"), from, types);

		// remove from list
		if(ignore_head == i)
			ignore_head = i->next;
		if(ignore_tail == i)
			ignore_tail = i->prev;
		if(i->prev)
			i->prev->next = i->next;
		if(i->next)
			i->next->prev = i->prev;

		// free ignore
		track_free(i->from, "userUnignore");
		if(i->host) track_free(i->host, "userUnignore");
		if(i->body) track_free(i->body, "userUnignore");
		track_free(i, "userUnignore");
	}
}


void userUnplug( char *data, int datalen ){ // unload plugin
	// UNPLUG plugin_name

	if(datalen == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/UNPLUG plugin_name");
	} else if(plugUnload(plugGet(data)) == -1){
		msgGotoServer(NULL, MF_PLUG_NONE, 1, data);
	}
}


void userUserMode( char *data, int datalen ){ // set user mode
	// USERMODE mode

	if(datalen == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/USERMODE mode");
	} else {
		netSend(n_current, "MODE %s %s\n", n_current->nickname, data);
	}
}


void userVersion( char *data, int datalen ){ // send version request to nick or channel
	// VERSION [nick]

	if(datalen > 0){
		// request version for nick or nicks

		skipword(data); // allow only one nick

  // TODO: check for multiple nicks, send version to each

		msgUnkPage(n_current, data, -1, NULL, MF_SENT_CTCP_USER, 2, "VERSION", data);
		netSend(n_current, "PRIVMSG %s :\001VERSION\001\n", data);
	} else if(*(p_current->name) == '#'){
		// request version for every nick in channel

		if(p_current->flags.x.joined){
			msgCurrentPage(-1, NULL, MF_SENT_CTCP_CHANNEL, 2, "VERSION", p_current->name);

			netSend(n_current, "PRIVMSG %s :\001VERSION\001\n", p_current->name);
		} else msgCurrentPage(-1, NULL, MF_NOT_IN_CHAN, 0);
	} else if(*(p_current->name) != '!'){
		// request version for nick page

		msgCurrentPage(-1, NULL, MF_SENT_CTCP_USER, 2, "VERSION", p_current->name);

		netSend(n_current, "PRIVMSG %s :\001VERSION\001\n", p_current->name);
	} else if(p_current == p_server){
		// request server version

		if(n_current->server_name != NULL)
			msgServer(-1, n_current, MF_SENT_VERSION_SERVER, 1, n_current->server_name);

		netSend(n_current, "VERSION\n");
	}
}


void userWho( char *data, int datalen ){ // get info on nicks in channel
	// WHO [nick]

	if(datalen > 0){
		netSend(n_current, "WHO %s\n", data);
	} else if(*(p_current->name) == '#'){
		if(p_current->flags.x.joined){
			netSend(n_current, "WHO %s\n", p_current->name);
		} else msgCurrentPage(-1, NULL, MF_NOT_IN_CHAN, 0);
	} else msgCurrentPage(-1, NULL, MF_CHANNEL_ONLY, 1, "WHO");
}


void userWhoami( char *data, int datalen ){ // get whois info for self
	// WHOAMI

	if(datalen > 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/WHOAMI");
	} else {
		netSend(n_current, "WHOIS %s\n", n_current->nickname);
	}
}


void userWhois( char *data, int datalen ){ // get whois info for nick
	// WHOIS [nick]

	if(datalen > 0){
		netSend(n_current, "WHOIS %s\n", data);
	} else if(*(p_current->name) != '#' && *(p_current->name) != '!'){
		netSend(n_current, "WHOIS %s\n", p_current->name);
	} else {
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/WHOIS [nick]");
	}
}


void userWhowas( char *data, int datalen ){ // get whowas info for nick
	// WHOWAS nick

	if(datalen == 0){
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/WHOWAS nick");
	} else {
		netSend(n_current, "WHOWAS %s\n", data);
	}
}
